// AssemblyHelper.cpp : Implementation of DLL Exports.


#include "stdafx.h"
#include "resource.h"
#include <cor.h>
#include "..\AssemblyHelper.h"
#include "dlldatax.h"

#include "XmlLogHelper.h"				// LogXmlError()

#ifndef INCLUDED_MSXML_TYPE_LIBRARY
#define INCLUDED_MSXML_TYPE_LIBRARY
#import "msxml.dll" rename_namespace("MSXML"), named_guids, raw_interfaces_only, raw_native_types
#endif // INCLUDED_MSXML_TYPE_LIBRARY

static const TCHAR* szASSEMBLY_HELPER_LOG_FILE			= _T("AssemblyHelper.log");
static const TCHAR* szASSEMBLY_HELPER_MODULE			= _T("AssemblyHelper");

HRESULT GetLogSettingsNode(HINSTANCE hInstance, MSXML::IXMLDOMNode** ppiXmlLogSettings);

class CAssemblyHelperModule : public CAtlDllModuleT< CAssemblyHelperModule >
{
public :
	DECLARE_LIBID(LIBID_AssemblyHelperLib)
	DECLARE_REGISTRY_APPID_RESOURCEID(IDR_ASSEMBLYHELPER, "{2BFA754E-9724-4268-BC6C-FA8B5D76856F}")
};

CAssemblyHelperModule _AtlModule;

#ifdef _MANAGED
#pragma managed(push, off)
#endif

// DLL Entry Point
extern "C" BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	HRESULT hr = S_OK;

#ifdef _MERGE_PROXYSTUB
    if (!PrxDllMain(hInstance, dwReason, lpReserved))
        return FALSE;
#endif
	hInstance;

	// Step 1:	Initializes logging

	if (!GetLog().isInitialized())
	{
		hr = GetLog().initialize( szASSEMBLY_HELPER_MODULE );
		if (SUCCEEDED(hr))
		{
			CComPtr<MSXML::IXMLDOMNode> spiLogSettingsNode;
			hr = GetLogSettingsNode(hInstance, &spiLogSettingsNode);
			if (hr == S_OK && spiLogSettingsNode)
			{
				hr = ::GetLog().loadConfig(spiLogSettingsNode);
				_ASSERTE(SUCCEEDED(hr));
			}
			else
			{
				hr = ::GetLog().useLogFile(szASSEMBLY_HELPER_LOG_FILE);
				_ASSERTE(SUCCEEDED(hr));
			}
		}
	}

	return _AtlModule.DllMain(dwReason, lpReserved); 
}

#ifdef _MANAGED
#pragma managed(pop)
#endif




// Used to determine whether the DLL can be unloaded by OLE
STDAPI DllCanUnloadNow(void)
{
#ifdef _MERGE_PROXYSTUB
    HRESULT hr = PrxDllCanUnloadNow();
    if (hr != S_OK)
        return hr;
#endif
    return _AtlModule.DllCanUnloadNow();
}


// Returns a class factory to create an object of the requested type
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
#ifdef _MERGE_PROXYSTUB
    if (PrxDllGetClassObject(rclsid, riid, ppv) == S_OK)
        return S_OK;
#endif
    return _AtlModule.DllGetClassObject(rclsid, riid, ppv);
}


// DllRegisterServer - Adds entries to the system registry
STDAPI DllRegisterServer(void)
{
    // registers object, typelib and all interfaces in typelib
    HRESULT hr = _AtlModule.DllRegisterServer();
#ifdef _MERGE_PROXYSTUB
    if (FAILED(hr))
        return hr;
    hr = PrxDllRegisterServer();
#endif
	return hr;
}


// DllUnregisterServer - Removes entries from the system registry
STDAPI DllUnregisterServer(void)
{
	HRESULT hr = _AtlModule.DllUnregisterServer();
#ifdef _MERGE_PROXYSTUB
    if (FAILED(hr))
        return hr;
    hr = PrxDllRegisterServer();
    if (FAILED(hr))
        return hr;
    hr = PrxDllUnregisterServer();
#endif
	return hr;
}

HRESULT GetLogSettingsNode(HINSTANCE hInstance, MSXML::IXMLDOMNode** ppiXmlLogSettings)
{
	HRESULT hr = S_OK;

	_ASSERTE(ppiXmlLogSettings);
	_ASSERTE(!*ppiXmlLogSettings);

	// Step 1:	Find out the path to ourselves (Patcher.dll)

	wchar_t szModulePath[MAX_PATH+1];
	DWORD dwRet = GetModuleFileNameW(	hInstance,
										szModulePath,
										MAX_PATH);
	if (!dwRet)
	{
		hr = HRESULT_FROM_WIN32(GetLastError());
		return ILXLOG(hr, _T("Failed to obtain module file name."), ILXLOG_ERROR);
	}

	// Step 2:	Extract our module directory

	wchar_t szDrive[_MAX_DRIVE];
	wchar_t szDir[_MAX_DIR];
	_wsplitpath_s(szModulePath, szDrive, _countof(szDrive), szDir, _countof(szDir), 
		0, 0, 0, 0);

	// Step 3:	Create the path to ILXLogSettings.xml

	wchar_t szProfilerLogFilePath[MAX_PATH];
	_wmakepath_s(szProfilerLogFilePath, _countof(szProfilerLogFilePath),
		szDrive, szDir, L"ILXLogSettings", L"xml");

	// We may need to remove "Debug" from the directory name
	if (!::PathFileExistsW(szProfilerLogFilePath) && *szDir)
	{
		wchar_t* pLastSlash = wcsrchr(szDir, L'\\');
		if (pLastSlash)
		{
			*pLastSlash = L'\0';
			pLastSlash = wcsrchr(szDir, L'\\');
			if (pLastSlash)
			{
				*pLastSlash = L'\0';
				_wmakepath_s(szProfilerLogFilePath, _countof(szProfilerLogFilePath),
					szDrive, szDir, L"ILXLogSettings", L"xml");
			}
		}
	}

	// Step 4:	Load the log file

	_variant_t varUrlString = szProfilerLogFilePath;

	VARIANT_BOOL bIsSuccessful = VARIANT_FALSE;
	CComPtr<MSXML::IXMLDOMDocument> spiXmlDocument;

	hr = spiXmlDocument.CoCreateInstance(MSXML::CLSID_DOMDocument);
	VERIFY_HR(hr, _T("Failed to create XML parser."), ILXLOG_ERROR);

	hr = spiXmlDocument->load(varUrlString, &bIsSuccessful);
	VERIFY_HR(hr, _T("Failed to load log settings file."), ILXLOG_ERROR);
	
	if (!bIsSuccessful)
	{
		hr = ::LogXmlError(spiXmlDocument);
		return ILXLOG(hr, _T("Failed to load log settings file."), ILXLOG_ERROR);
	}

	CComPtr<MSXML::IXMLDOMElement> spiDocumentElement;
	hr = spiXmlDocument->get_documentElement(&spiDocumentElement);
	VERIFY_HR_COND(hr, spiDocumentElement, 
		_T("Failed to retrieve the document element for the log settings file."), ILXLOG_ERROR);

	hr = spiDocumentElement->QueryInterface(ppiXmlLogSettings);
	VERIFY_HR_COND(hr, *ppiXmlLogSettings, 
		_T("Failed to QI for IXMLDOMNode."), ILXLOG_ERROR);

	return S_OK;
}
