///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Common/AssemblyHelper/EnCDeltaHelper.cpp 17    5/10/06 8:02p Eaddy $
//
//    File name:	EnCDeltaHelper.cpp
//   Created By:	Marc Eaddy
//   Created on:	04/23/06
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Common/AssemblyHelper/EnCDeltaHelper.cpp $
 * 
 * 17    5/10/06 8:02p Eaddy
 * o Closer to multiple weaving
 * 
 * 16    5/05/06 4:11p Eaddy
 * o Now keeps the buffer around while we have a
 * metadata scope open (just in case)
 * 
 * 15    5/05/06 11:25a Eaddy
 * o Now supports weaving multiple times
 * 
 * 14    5/03/06 8:15p Eaddy
 * o Now allows IMetaDataImport to come from Wicca
 * 
 * 13    4/28/06 8:01p Eaddy
 * o Figured out how to marshal iCorDebugModule from Wicca
 * 
 * 12    4/23/06 12:39p Eaddy
 * o Attempting to dump the iface passed by Coordinator
 */

#include "stdafx.h"

#include "EnCDeltaHelper.h"     // Me

#include "MetadataHelpers.h"

#include "MetadataIterators.h"

#include "AssemblyDiff.h"

//#include "AxDump.h"

// CEnCDeltaHelper

CEnCDeltaHelper::CEnCDeltaHelper()
{ }

CEnCDeltaHelper::~CEnCDeltaHelper()
{ }

HRESULT CEnCDeltaHelper::FinalConstruct()
{
	return S_OK;
}

void CEnCDeltaHelper::FinalRelease()
{ }

STDMETHODIMP CEnCDeltaHelper::CreateDeltas(	/*[in]*/ LONGLONG _piunkCorModuleOrig,
                                            /*[in]*/ BSTR bstrOriginalModulePath, 
											/*[in]*/ BSTR bstrNewModulePath,
											/*[in]*/ BSTR bstrDeltaDir,
											/*[out]*/ BSTR* pbstrMetadataDeltaPath,
											/*[out]*/ BSTR* pbstrILDeltaPath,
											/*[out]*/ BSTR* pbstrDebugMsg)
{
	HRESULT hr = S_OK;

	// Verify args
	if (!bstrOriginalModulePath || !bstrNewModulePath)
		return E_POINTER;

	if (!*bstrOriginalModulePath || !*bstrNewModulePath)
		return E_INVALIDARG;

	// Must specify at least one
	if (!pbstrMetadataDeltaPath && !pbstrILDeltaPath)
		return E_INVALIDARG;

	// Initialize out params
	*pbstrMetadataDeltaPath = 0;
	*pbstrILDeltaPath = 0;

	if (pbstrDebugMsg)
		*pbstrDebugMsg = 0;

	USES_CONVERSION;

	ILXLOGMSG2(ILXLOG_INFO, _T("Comparing '%ls' and '%ls'."), 
		bstrOriginalModulePath, bstrNewModulePath);

    CComPtr<ICorDebugModule> spiCorDebugModuleOrig;
	CComPtr<IMetaDataImport> spiImportOrig;
    LPBYTE pbMapViewOfFileOrig = 0;
    CQuickBytes metadataBuffer;

    // Step 1:  Obtain ICorDebugModule and IMetaDataImport ifaces for the old file and the

    if (_piunkCorModuleOrig)
    {
        hr = ((IUnknown*) _piunkCorModuleOrig)->QueryInterface(__uuidof(spiCorDebugModuleOrig.p),
            (void**) &spiCorDebugModuleOrig.p);
        VERIFY_HR_COND(hr, spiCorDebugModuleOrig.p, _T("Failed to QI for ICorDebugModule."), ILXLOG_ERROR);

        hr = spiCorDebugModuleOrig->GetMetaDataInterface(IID_IMetaDataImport, 
            (IUnknown**) &spiImportOrig.p);
        VERIFY_HR_COND(hr, spiImportOrig.p, _T("GetMetaDataInterface() failed."), ILXLOG_ERROR);

        // We have to make a copy of the metadata since we aren't allowed
        // to modify the original metadata directly (only indirectly via
        // ICorDebugModule::ApplyEdits).
	    CComPtr<IMetaDataImport> spiImportOrigCopy;
        hr = CopyMetaData(spiImportOrig, true, &metadataBuffer, &spiImportOrigCopy);
        VERIFY_HR_COND(hr, spiImportOrigCopy.p, _T("Failed to copy metadata."), ILXLOG_ERROR);

        // Use the copy
        spiImportOrig = spiImportOrigCopy;
    }
    else
    {
	    hr = GetMetaDataImport(	OLE2CA(bstrOriginalModulePath), 
							    true, // Track deltas
							    &spiImportOrig.p, 
							    &pbMapViewOfFileOrig);
	    if (FAILED(hr)) return E_FAIL;
    }

	CComPtr<IMetaDataEmit2> spiMetaDataEmitOrig;
	hr = spiImportOrig->QueryInterface(	IID_IMetaDataEmit2, 
										(void**) &spiMetaDataEmitOrig.p );
	VERIFY_HR_COND(hr, spiMetaDataEmitOrig, _T("QI for IMetaDataEmit2 failed."), ILXLOG_ERROR);

    // Step 2:  Obtain IMetaDataImport iface for the new file

    CComPtr<IMetaDataImport> spiImportNew;
	LPBYTE pbMapViewOfFileNew = 0;

	hr = GetMetaDataImport(	OLE2CA(bstrNewModulePath), 
							false,
							&spiImportNew.p, 
							&pbMapViewOfFileNew);
	if (FAILED(hr)) return E_FAIL;

	// Step 3:  Create the output delta file paths

	CComBSTR bstrMetadataDeltaPath;
	CComBSTR bstrILDeltaPath;
	if (bstrDeltaDir)
	{
		const TCHAR* pNewPath = OLE2CT(bstrNewModulePath);
		const TCHAR* pExt = _tcsrchr(pNewPath, _T('.'));
		const TCHAR* pDirSeparator = _tcsrchr(pNewPath, _T('\\'));

		_ASSERTE(pExt);
		if (!pExt) 
			return E_UNEXPECTED;

		if (pDirSeparator)
		{
			bstrMetadataDeltaPath.AppendBytes((char*) pNewPath, 
				(int) (pDirSeparator-pNewPath+1) * sizeof(TCHAR));
			++pDirSeparator;
		}
		else
		{
			pDirSeparator = pNewPath;
		}

		bstrMetadataDeltaPath.AppendBytes((char*) pDirSeparator, 
			(int) (pExt-pDirSeparator) * sizeof(TCHAR));

		bstrILDeltaPath = bstrMetadataDeltaPath;
	}
	else
	{
		bstrMetadataDeltaPath = bstrNewModulePath;
		bstrILDeltaPath = bstrNewModulePath;
	}

	bstrMetadataDeltaPath.Append(L".dmeta");
	bstrILDeltaPath.Append(L".dil");

	FILE* pF = NULL;
	errno_t err = _wfopen_s(&pF, bstrILDeltaPath, L"wb");
	VERIFY_HR1(HRESULT_FROM_WIN32(err), ILXLOG_ERROR,
		_T("Failed to create delta IL file '%s'."), bstrILDeltaPath);

	// Save room for the delta file size

	ULONG blah = 0xBAADF00D;
	fwrite(&blah, sizeof(blah), 1, pF);

    // Step 4:  Diff the files

	TypeMapMethodInfo mapChangedFunctions;
    TypeListTokenMappings listNewOrReorderedTokens;

	hr = ::CreateUpdatePatches(	spiCorDebugModuleOrig,
                                spiImportOrig,
								pbMapViewOfFileOrig,
								true,
								0, // ModuleID
								spiImportNew,
								pbMapViewOfFileNew,
								NULL,
								spiMetaDataEmitOrig,
								NULL,
								true,
								pF,
								&mapChangedFunctions,
                                &listNewOrReorderedTokens);
	if (FAILED(hr)) 
	{
		fclose(pF);
		pF = 0;
		return E_FAIL;
	}

	// Update the delta file size

	ULONG size = ftell(pF) - sizeof(DWORD);

    if (size == 0)
    {
        // No methods were changed so remove the empty
        // dil file

    	fclose(pF);
	    pF = 0;
        _wremove(bstrILDeltaPath);
        bstrILDeltaPath.Empty();
    }
    else
    {
	    rewind(pF);
	    fwrite(&size, sizeof(size), 1, pF);
	    fclose(pF);
	    pF = 0;
    }

	if (FAILED(hr)) return E_FAIL;

	// Step 5:  Output the metadata delta file

    bool somethingGotEmitted = false;

    for(TypeListTokenMappings::const_iterator it = listNewOrReorderedTokens.begin();
        it != listNewOrReorderedTokens.end();
        ++it)
    {
        const STokenMapping& tokMapping = *it;

        if (tokMapping.isNew)
        {
            somethingGotEmitted = true;
            break;
        }
    }

    if (!somethingGotEmitted)
    {
	    *pbstrMetadataDeltaPath = 0;
        return S_OK;
    }

	DWORD metaDataSize = 0;
	hr = spiMetaDataEmitOrig->GetDeltaSaveSize(cssAccurate, &metaDataSize);
	VERIFY_HR(hr, _T("Failed to obtain delta metadata size."), ILXLOG_ERROR);

	hr = spiMetaDataEmitOrig->SaveDelta(bstrMetadataDeltaPath, 0);
	VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to save metadata delta to file '%ls'."), 
		bstrMetadataDeltaPath);

	CComPtr<IMetaDataDispenserEx> spiMetaDataDispenser;

	hr = ::CoCreateInstance(	CLSID_CorMetaDataDispenser, 
								NULL, 
								CLSCTX_SERVER,
								IID_IMetaDataDispenserEx,
								(void**)(&spiMetaDataDispenser.p));

    if (FAILED(hr) || !spiMetaDataDispenser)
	{
		hr = FAILED(hr) ? hr : E_FAIL;
		return ILXLOG(hr, _T("Unable to create metadata dispenser."), ILXLOG_ERROR);
	}

	*pbstrMetadataDeltaPath = bstrMetadataDeltaPath.Detach();

	// step 6:  Update the IL delta file path

	*pbstrILDeltaPath = bstrILDeltaPath.Detach();

    // Step 7:  Update the debug msg

    if (!pbstrDebugMsg)
        return S_OK;

    CComBSTR bstrDebugMsg;

    ULONG32 lastType = 0;
    bool appendedHeader = false;
    bool needsComma = false;

    for(TypeListTokenMappings::const_iterator it = listNewOrReorderedTokens.begin();
        it != listNewOrReorderedTokens.end();
        ++it)
    {
        const STokenMapping& tokMapping = *it;

        if (!tokMapping.isNew)
            continue;

        ULONG32 type = TypeFromToken(tokMapping.from);

        if (!appendedHeader)
        {
            bstrDebugMsg.Append(L"Emitted tokens:");
            appendedHeader = true;
        }

        if (type != lastType)
        {
            bstrDebugMsg.Append(L"\n  ");
            bstrDebugMsg.Append(Token2TypeName(tokMapping.to));
            bstrDebugMsg.Append(L"s: ");
            lastType = type;
            needsComma = false;
        }

        if (needsComma)
        {
            bstrDebugMsg.Append(L", ");
        }

        WCHAR szTok[100];
        swprintf_s(szTok, (sizeof(szTok)-1) / sizeof(*szTok), 
            L" (0x%X)", RidFromToken(tokMapping.to));

        bstrDebugMsg.Append(tokMapping.szWhat);
        bstrDebugMsg.Append(szTok);

        needsComma = true;
    }

    if (appendedHeader)
    {
        *pbstrDebugMsg = bstrDebugMsg.Detach();
    }

	return S_OK;
}
