///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Common/AssemblyHelper/ILFetcher.cpp 5     5/11/06 10:17a Eaddy $
//
//    File name:	ILFetcher.cpp
//   Created By:	Marc Eaddy
//   Created on:	04/21/06
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Common/AssemblyHelper/ILFetcher.cpp $
 * 
 * 5     5/11/06 10:17a Eaddy
 * o Added more checks to obtain the correct
 * method header
 * 
 * 4     5/10/06 8:02p Eaddy
 * o Closer to multiple weaving
 * 
 * 3     5/03/06 8:15p Eaddy
 * o Now allows IMetaDataImport to come from Wicca
 * 
 * 2     4/23/06 12:50p Eaddy
 * 
 * 1     4/22/06 12:35p Eaddy
 * o Adding support for using the Debugger API to query for the current
 * metadata and IL
 */

#include "stdafx.h"

#include "ILFetcher.h"			// Me

#include "corhlpr.h"			// IMetaDataImport, mdMethodDef

#include <dbghelp.h>			// ImageNtHeader()

// HELPER FUNCTIONS

 HRESULT GetMethodHeader(	LPCBYTE pbMapViewOfFile,
							ULONG rva, 
							bool bIsImage,
							unsigned __int8** ppbMethodHeader);

 HRESULT GetMethodPtrInfo(	LPCBYTE pbMapViewOfFile,
							bool bIsImage,
							ULONG rva, 
							ULONG* pulMethodSize,
							unsigned __int8** ppbMethodHeader,
							ULONG* pulMethodHeaderSize,
							LPCBYTE* ppbMethodIL,
							ULONG* pulCodeSize,
							mdSignature* ptokLocalVarSig);

 bool IsBadMethodHeaderPtr(unsigned __int8* pbMethodHeader);

//////////////////////////////
// IL FETCHER IMPLEMENTATION
//////////////////////////////

CILFetcher::CILFetcher()
    : spiCorModule(0)
    , pbMapViewOfFile(NULL)
    , bIsImage(false)
{ }

HRESULT CILFetcher::init(ICorDebugModule* piCorDebugModule)
{
    _ASSERTE(piCorDebugModule);
    if (!piCorDebugModule) return E_POINTER;

    // Only call once!
    _ASSERTE(this->pbMapViewOfFile == NULL);
    _ASSERTE(this->bIsImage == false);
    _ASSERTE(!this->spiCorModule);
    
    spiCorModule = piCorDebugModule;

    return S_OK;
}

HRESULT CILFetcher::init(LPCBYTE const pbMapViewOfFile, bool bIsImage)
{
    _ASSERTE(pbMapViewOfFile);
    if (!pbMapViewOfFile) return E_POINTER;

    // Only call once!
    _ASSERTE(this->pbMapViewOfFile == NULL);
    _ASSERTE(this->bIsImage == false);
    _ASSERTE(!this->spiCorModule);

    this->pbMapViewOfFile = pbMapViewOfFile;
    this->bIsImage = bIsImage;

    return S_OK;
}

HRESULT CILFetcher::GetMethodPtrInfo(   /*[in]*/ ULONG methodRVA,
                                        /*[in]*/ mdMethodDef methodDef, 
							            /*[out]*/ ULONG* pulMethodSize,
							            /*[out]*/ unsigned __int8** ppbMethodHeader,
							            /*[out]*/ ULONG* pulMethodHeaderSize,
							            /*[out]*/ LPCBYTE* ppbMethodIL,
							            /*[out]*/ ULONG* pulCodeSize,
							            /*[out]*/ mdSignature* ptokLocalVarSig,
                                        /*[out]*/ bool* pbMustDeleteHeader) const
{
    HRESULT hr = S_OK;

	if (!ppbMethodHeader && !ppbMethodIL && !pulCodeSize) return E_POINTER;
    if (!pbMustDeleteHeader) return E_POINTER;

    if (spiCorModule)
    {
        _ASSERTE(ppbMethodHeader); // Because we will need to delete it

        CComPtr<ICorDebugFunction> spiCorFunction;

        // Step 1:  Get the ICorDebugCode

        hr = spiCorModule->GetFunctionFromToken(methodDef, &spiCorFunction.p);
        VERIFY_HR_COND1(hr, spiCorFunction.p, ILXLOG_ERROR, _T("Failed to retrieve ICorDebugFunction for method token 0x%08X."), 
            methodDef);

        CComPtr<ICorDebugCode> spiCode;
        hr = spiCorFunction->GetILCode(&spiCode.p);
        VERIFY_HR_COND1(hr, spiCode.p, ILXLOG_ERROR, _T("Failed to retrieve ICorDebugCode for method token 0x%08X."), 
            methodDef);

        #ifdef _DEBUG
        BOOL bIsIL = FALSE;
        hr = spiCode->IsIL(&bIsIL);
        CHECK_HR_COND1(hr, bIsIL, ILXLOG_ERROR, _T("Expected IL Code for method token 0x%08X."), 
            methodDef);
        #endif // _DEBUG
        
        //ULONG32 version = 0;
        //hr = spiCode->GetVersionNumber(&version);
        //VERIFY_HR_COND1(hr, version > 0, ILXLOG_ERROR, _T("Unexpected version for method token 0x%08X."), 
        //    methodDef);

        // Step 2:  Get the address of the method IL relative to
        //          client process

        CORDB_ADDRESS pbMethodILRemote = 0;
        hr = spiCode->GetAddress(&pbMethodILRemote);
        VERIFY_HR_COND1(hr, pbMethodILRemote != 0, ILXLOG_ERROR, 
            _T("Failed to retrieve address for method token 0x%08X."), 
            methodDef);

        // Step 3:  Get the ICorDebugProcess

        CComPtr<ICorDebugProcess> spiProcess;
        hr = spiCorModule->GetProcess(&spiProcess);
        VERIFY_HR_COND1(hr, spiProcess.p, ILXLOG_ERROR, 
            _T("Failed to obtain process for method token 0x%08X."), methodDef);

        // Step 4:  Get the total method size
        //
        //          We do this by probing the memory immediately before the
        //          method IL address to obtain the method header.  This is
        //          very kludgy since a fat method may have extra sections
        //          between the IL and the header (true?).

        // Step 5:  Allocate memory for the method

        ULONG32 codeSize = 0;
        spiCode->GetSize(&codeSize);
        VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to get size for method token 0x%08X."), 
            methodDef);

        DWORD methodSizeAllocated = codeSize + sizeof(COR_ILMETHOD_FAT) + 
            (sizeof(COR_ILMETHOD_SECT_FAT) * 100);

        *ppbMethodHeader = new BYTE[methodSizeAllocated];

        #ifdef _DEBUG
        memset(*ppbMethodHeader, 0xCC, methodSizeAllocated);
        #endif // _DEBUG

        // So CMethodInfo deletes the memory later
        *pbMustDeleteHeader = true;

        DWORD headerSize = sizeof(COR_ILMETHOD_FAT);

        SIZE_T numBytesCopied = 0;
        hr = spiProcess->ReadMemory(pbMethodILRemote - headerSize, 
                                    headerSize + codeSize, 
                                    *ppbMethodHeader, 
                                    &numBytesCopied);

        _ASSERTE(numBytesCopied <= headerSize + codeSize);

        COR_ILMETHOD* pMethodHeader = (COR_ILMETHOD*) *ppbMethodHeader;
        DWORD methodSize = headerSize + codeSize;

        if (FAILED(hr) || 
            !pMethodHeader->Fat.IsFat() ||
            pMethodHeader->Fat.Size != 3 ||
            pMethodHeader->Fat.GetCodeSize() != codeSize)
        {
            headerSize = sizeof(COR_ILMETHOD_TINY);
            hr = spiProcess->ReadMemory(pbMethodILRemote - headerSize, 
                                        headerSize + codeSize, 
                                        *ppbMethodHeader, 
                                        &numBytesCopied);
            VERIFY_HR_COND1(hr, numBytesCopied <= headerSize + codeSize, ILXLOG_ERROR, 
                _T("Failed to read data from remote process for method token 0x%08X."), methodDef);

            pMethodHeader = (COR_ILMETHOD*) *ppbMethodHeader;

            methodSize = headerSize + codeSize;

            // Sanity check
            _ASSERTE(pMethodHeader->Tiny.IsTiny());
            _ASSERTE(pMethodHeader->Tiny.GetSect() == NULL);
            _ASSERTE(pMethodHeader->Tiny.GetCodeSize() == codeSize);
        }
        else if (pMethodHeader->Fat.Flags & CorILMethod_MoreSects)
        {
            // Deal with fatso
	        COR_ILMETHOD_DECODER methodDecoder(pMethodHeader);

            _ASSERTE(pMethodHeader->Fat.Size == 3);

            _ASSERTE(methodDecoder.CodeSize == codeSize);

            CORDB_ADDRESS pCurRemote = pbMethodILRemote + codeSize;
            BYTE* pCurLocal = *ppbMethodHeader + methodSize;

            bool hasMoreSections = true;

            while(hasMoreSections)
            {
                COR_ILMETHOD_SECT_FAT sect;
                numBytesCopied = 0;
                hr = spiProcess->ReadMemory(pCurRemote, 
                                            sizeof(sect), 
                                            (BYTE*) &sect, 
                                            &numBytesCopied);
                VERIFY_HR_COND1(hr, numBytesCopied <= sizeof(sect), ILXLOG_ERROR, 
                    _T("Failed to read data from remote process for method token 0x%08X."), methodDef);

                COR_ILMETHOD_SECT* pSect = (COR_ILMETHOD_SECT*) &sect;

                DWORD sectSize = 0;

                if (pSect->IsFat())
                {
                    sectSize = sizeof(COR_ILMETHOD_SECT_FAT);
                }
                else
                {
                    sectSize = sizeof(COR_ILMETHOD_SECT_SMALL);
                }

                DWORD align = (DWORD) ((((UINT_PTR) pCurLocal) + 3) & ~3) - (UINT_PTR) pCurLocal;

                sectSize += pSect->DataSize();

                methodSize += sectSize + align;

                if (methodSize > methodSizeAllocated)
                {
                    BYTE* tmp = new BYTE[methodSize];
                    memcpy(tmp, *ppbMethodHeader, methodSize);
                    delete[] *ppbMethodHeader;
                    *ppbMethodHeader = tmp;
                    methodSizeAllocated = methodSize;
                }

                if (align > 0)
                {
                    memset(pCurLocal, 0, align);
                    pCurLocal += align;
                }

                // Copy the section
                numBytesCopied = 0;
                hr = spiProcess->ReadMemory(pCurRemote, 
                                            sectSize, 
                                            pCurLocal, 
                                            &numBytesCopied);
                VERIFY_HR_COND1(hr, numBytesCopied <= sectSize, ILXLOG_ERROR, 
                    _T("Failed to read data from remote process for method token 0x%08X."), methodDef);

                hasMoreSections = pSect->More();
                pCurRemote += sectSize;
                pCurLocal += sectSize;
            }
        }
        else
        {
            _ASSERTE(pMethodHeader->Fat.IsFat());
            _ASSERTE(pMethodHeader->Fat.Size == 3);
            _ASSERTE(pMethodHeader->Fat.GetCodeSize() == codeSize);
        }

        // Step 7:  Decode the header

	    hr = GetMethodIL(     *ppbMethodHeader, 
                              pulMethodSize,
                              pulMethodHeaderSize, 
                              ppbMethodIL,
		                      pulCodeSize, 
                              ptokLocalVarSig);
        
        _ASSERTE(pulMethodSize == NULL || *pulMethodSize == methodSize);
        _ASSERTE(pulCodeSize == NULL || *pulCodeSize == codeSize);

        return hr;
    }
    else if (pbMapViewOfFile)
    {
        *pbMustDeleteHeader = false;
        return ::GetMethodPtrInfo(pbMapViewOfFile, bIsImage, methodRVA, pulMethodSize, 
            ppbMethodHeader, pulMethodHeaderSize, ppbMethodIL, pulCodeSize, ptokLocalVarSig);
    }
    else
    {
        _ASSERTE(!"CILFetcher not initialized.");
        *pbMustDeleteHeader = false;
        return E_POINTER;
    }

    return S_OK;
}

/////////////////////
// HELPER FUNCTIONS
/////////////////////

HRESULT GetMethodPtrInfo(	LPCBYTE const pbMapViewOfFile,
							bool bIsImage,
							ULONG rva, 
							ULONG* pulMethodSize,
							unsigned __int8** ppbMethodHeader,
							ULONG* pulMethodHeaderSize,
							LPCBYTE* ppbMethodIL,
							ULONG* pulCodeSize,
							mdSignature* ptokLocalVarSig)
{
	if (!pbMapViewOfFile) return E_POINTER;
	if (!ppbMethodHeader && !ppbMethodIL && !pulCodeSize) return E_POINTER;

	HRESULT hr = S_OK;

	unsigned __int8* pbMethodHeader = 0;
	hr = GetMethodHeader(pbMapViewOfFile, rva, bIsImage, &pbMethodHeader);
	if (FAILED(hr)) return hr;
	if (!pbMethodHeader) return E_UNEXPECTED;

	if (ppbMethodHeader)
		*ppbMethodHeader = pbMethodHeader;

	return GetMethodIL(pbMethodHeader, pulMethodSize, pulMethodHeaderSize, ppbMethodIL,
		pulCodeSize, ptokLocalVarSig);
}

// Adapted from a presentation by John Lam
HRESULT GetMethodHeader(	LPCBYTE pbMapViewOfFile,
							ULONG rva, 
							bool bIsImage,
							unsigned __int8** ppbMethodHeader)
{
	HRESULT hr = S_OK;

	_ASSERTE(ppbMethodHeader);
	if (!ppbMethodHeader) return E_POINTER;

	if (bIsImage)
	{
		PIMAGE_NT_HEADERS pNtHeader = ::ImageNtHeader((PVOID) pbMapViewOfFile);
		if (!pNtHeader)
		{
			hr = HRESULT_FROM_WIN32(::GetLastError());
			return ILXLOG(hr, _T("ImageNtHeader failed."), ILXLOG_ERROR);
		}

		*ppbMethodHeader = 
			(unsigned __int8*) ImageRvaToVa(pNtHeader, (PVOID) pbMapViewOfFile, rva, 0);
	}

	// KLUDGE: An RVA can either be a direct offset from the base address or
	// a virtual address (VA) relative to a section of a PE file.  It appears
	// to always be an offset if the CLR is loading the file, except when
	// running under Rotor where it always appears to be a VA.  It always
	// appears to be a VA when the base address comes from mapping a PE file
	// in memory.  So it appears that Rotor uses memory mapped files and the
	// CLR doesn't.  Possibly the CLR is obtaining the memory via IMethodAlloc
	// or something.  If ImageRvaToVa() doesn't appear to work we try treating
	// it as an offset.  This is very kludgy but I don't know another way to do this.

	if (!bIsImage) /* || IsBadMethodHeaderPtr(*ppbMethodHeader))*/
	{
		*ppbMethodHeader = (unsigned __int8*) pbMapViewOfFile + rva;
		//if (IsBadMethodHeaderPtr(*ppbMethodHeader))
		//{
		//	return ILXLOG(E_UNEXPECTED, _T("ImageRvaToVa() failed."), ILXLOG_ERROR);
		//}
	}

	return S_OK;
}

bool IsBadMethodHeaderPtr(unsigned __int8* pbMethodHeader)
{
	if (!pbMethodHeader)
	{
		return true;
	}

	COR_ILMETHOD* pMethodHeader = (COR_ILMETHOD*) pbMethodHeader;

	if (IsBadReadPtr(pMethodHeader, sizeof(COR_ILMETHOD)))
	{
		return true;
	}

	// Check for sane/legal values

	if (!pMethodHeader->Tiny.IsTiny() && !pMethodHeader->Fat.IsFat())
	{
		return true;
	}

	const DWORD MAX_CODE_SIZE_KLUDGE = 10000;
	if (pMethodHeader->Fat.IsFat() && pMethodHeader->Fat.CodeSize > MAX_CODE_SIZE_KLUDGE)
	{
		return true;
	}

	return false;
}

HRESULT GetMethodIL(	/*[in]*/  unsigned __int8* const pbMethodHeader,
						/*[out]*/ ULONG* pulMethodSize,
						/*[out]*/ ULONG* pulHeaderSize,
						/*[out]*/ LPCBYTE* ppbMethodIL,
						/*[out]*/ ULONG* pulCodeSize,
						/*[out]*/ mdSignature* ptokLocalVarSig)
{
	HRESULT hr = S_OK;

	_ASSERTE(pbMethodHeader);
	if (!pbMethodHeader) return E_POINTER;
	if (!ppbMethodIL && !pulHeaderSize && !pulCodeSize) return E_INVALIDARG;

	COR_ILMETHOD* pMethodHeader = 
		reinterpret_cast<COR_ILMETHOD*>(pbMethodHeader);

	// For some reason this bombs sometimes
	COR_ILMETHOD_DECODER methodDecoder(pMethodHeader);

    _ASSERTE(pMethodHeader->Tiny.IsTiny() || pMethodHeader->Fat.IsFat());

	if (ptokLocalVarSig)
	{
		*ptokLocalVarSig = methodDecoder.GetLocalVarSigTok() ? 
			methodDecoder.GetLocalVarSigTok() : mdSignatureNil;
	}

	if (ppbMethodIL)
		*ppbMethodIL = methodDecoder.Code;

	if (pulCodeSize)
		*pulCodeSize = methodDecoder.CodeSize;

	if (pulMethodSize)
		*pulMethodSize = methodDecoder.GetOnDiskSize(pMethodHeader);

	if (pulHeaderSize)
		*pulHeaderSize = (ULONG) (methodDecoder.Code - (BYTE*) pMethodHeader);

	return S_OK;
}
