///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Common/Columbia.Utilities/MyTrace.cs 2     5/02/06 7:41p Eaddy $
//
//    File name:	MyTrace.cs
//   Created By:	Marc Eaddy
//   Created on:	10/10/04
//
//      Purpose:	
//
//  Description:	You must include "using System.Diagnostics;" to use
//					this class.
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Common/Columbia.Utilities/MyTrace.cs $
 * 
 * 2     5/02/06 7:41p Eaddy
 * 
 * 1     5/02/06 11:02a Eaddy
 * o Moved MyTrace into separate library
 * 
 * 13    5/30/05 9:51p Eaddy
 * o Fixed bug where we logged filename for Info msgs
 * 
 * 12    5/28/05 9:12p Eaddy
 * o Added some error checking
 * 
 * 11    5/28/05 9:10p Eaddy
 * o Now provides filename and line number for warnings and errors
 * 
 * 10    5/28/05 8:58p Eaddy
 * o Minor changes
 * 
 * 9     5/28/05 8:25p Eaddy
 * o Added newline after header
 * 
 * 8     5/28/05 8:06p Eaddy
 * o Now all trace statements go thru MyTrace for consistency
 * 
 * 7     5/28/05 1:21p Eaddy
 * Changed copyright
 * 
 * 6     4/06/05 6:59p Eaddy
 * o Added method overload for logging info messages
 * 
 * 5     2/25/05 5:09p Eaddy
 * o No longer outputs debug string twice. Now relies on the trace config
 * file to specify debug output.
 * 
 * 4     2/04/05 5:49p Eaddy
 * o Fixed name in the log messages
 * 
 * 3     2/04/05 3:04p Eaddy
 * o Added comment
 * 
 * 2     2/04/05 2:39p Eaddy
 * 
 * 1     2/04/05 2:38p Eaddy
 * o Added to Goblin project
 * 
 * Revision 1.5  2005/01/14 23:34:19  me133
 * o Now displays milliseconds
 *
 * Revision 1.4  2004/10/24 19:31:03  me133
 * o Implemented heartbeat message
 * o Now handles more network error conditions
 *
 * Revision 1.3  2004/10/17 08:43:38  me133
 * o Now logs messages to the debug console
 *
 * Revision 1.2  2004/10/11 05:57:54  me133
 * no message
 *
 * Revision 1.1  2004/10/10 07:40:07  me133
 * o Now outputs trace messages to log file usnig System.Diagnostics.Trace
 */

using System;
using System.Diagnostics;
using System.Threading;

using System.Runtime.InteropServices;
using System.Security.Permissions;
using System.Collections;
using System.IO;
using System.Text;

namespace Columbia.Utilities
{
	/// <summary>
	/// Summary description for MyTrace.
	/// </summary>
	public class MyTrace
	{
		static bool bIsFirstMessage = true;
		//static DateTime date = new DateTime();
		public static string source = "Watcher";

		static TraceSwitch level = new TraceSwitch("level", "Entire Application");

		/// Overloads for Write()

		public static void Write(object obj)
		{
			LogImpl(TraceLevel.Info, obj.ToString(), false);
		}

		public static void Write(string msg)
		{
			LogImpl(TraceLevel.Info, msg, false);
		}

		public static void Write(string format, params object[] args)
		{
			LogImpl(TraceLevel.Info, String.Format(format, args), false);
		}

		/// Overloads for WriteLine

		public static void WriteLine(object obj)
		{
			LogImpl(TraceLevel.Info, obj.ToString(), true);
		}

		public static void WriteLine(string msg)
		{
			LogImpl(TraceLevel.Info, msg, true);
		}

		public static void WriteLine(string format, params object[] args)
		{
			Log(TraceLevel.Info, format, args);
		}

		/// Overloads for Log()

		public static void Log(string format, params object[] args)
		{
			Log(TraceLevel.Info, format, args);
		}

		public static void Log(TraceLevel levelMessage, string msg)
		{
			Log(levelMessage, msg, true);
		}

		public static void Log(TraceLevel levelMessage, string format, params object[] args)
		{
			if (levelMessage > level.Level)
				return;

			LogImpl(levelMessage, String.Format(format, args), true);
		}

		/// <summary>
		/// This guys does all the work
		/// </summary>
		/// <param name="levelMessage"></param>
		/// <param name="msg"></param>
		/// <param name="newline"></param>
		public static void LogImpl(TraceLevel levelMessage, string msg, bool newline)
		{
			//if (levelMessage > level.Level)
			//	return;

			if (bIsFirstMessage)
			{
				Trace.WriteLine("");
				Trace.WriteLine(String.Format("**  Started {0}.log on {1} **", source, GetDate()) );
				Trace.WriteLine("**  Log File Format Version: 2.2 **");
				Trace.WriteLine("");
				bIsFirstMessage = false;
			}

			// 10/10 20:56:01  3028 - INFO - Watcher: blah blah blah.

			string strMsg;
			if (levelMessage <= TraceLevel.Warning)
			{
				StackTrace stack = new StackTrace(1, true);
				StackFrame frame = null;
				if (stack != null)
					frame = stack.GetFrame(1);

				strMsg = String.Format("{0}  {1} - {2} - {3}: {4} [File: {5}, Line: {6}]",
					GetDate(), 
					Thread.CurrentThread.GetHashCode(),
					GetSeverity(levelMessage),
					source,
					msg,
					frame != null ? frame.GetFileName() : "[UNK]",
					frame != null ? frame.GetFileLineNumber() : -1);
			}
			else
			{
				strMsg = String.Format("{0}  {1} - {2} - {3}: {4}",
					GetDate(), 
					Thread.CurrentThread.GetHashCode(),
					GetSeverity(levelMessage),
					source,
					msg);
			}

			if (newline)
				Trace.WriteLine(strMsg);
			else
				Trace.Write(strMsg);
			//System.Diagnostics.Debugger.Log((int) levelMessage, string.Empty, strMsg + "\r\n");
		}

		/// Private helper functions

		static string GetDate()
		{
			return DateTime.Now.ToString("MM/dd HH:mm:ss:fff");
		}

		static string GetSeverity(TraceLevel levelMessage)
		{
			if (levelMessage == TraceLevel.Verbose)
				return "SUCCESS";
			else if (levelMessage == TraceLevel.Info)
				return "INFO";
			else if (levelMessage == TraceLevel.Warning)
				return "WARNING";
			else if (levelMessage == TraceLevel.Error)
				return "ERROR";
			else
			{
				Debug.Assert(false, "Not supposed to happen");
				return levelMessage.ToString();
			}
		}
	}
}