/**
 * Copyright (C) 1999 ILX Systems.  All Rights Reserved.
 * File name: ILXQuoteCP.h CP proxy class for _IILXQuoteCtrlEvents interface
 * Created By: Andrey S. Koshelev
 * Created on 3/6/99 10:13:44 PM
 * Comments: ...
 *
 * $Log: ILXCommonEventProxy.cpp,v $
 * Revision 1.1  2004/09/07 06:10:49  me133
 * o Initial checkin to CVS
 *
 * 
 * 18    4/24/02 7:28p Marce
 * o Fixed bug where parameters were supposed to be in reverse order
 * 
 * 17    4/11/02 5:53p Marce
 * o Rolled back recent changes because it would cause some builds to fail
 * 
 * 14    10/25/01 4:27p Marce
 * o Added OnContextChanged
 * 
 * 12    10/16/00 5:19p Youngh
 * Added OnSymListChange event.
 * 
 * 10    2/10/00 3:50p Alexanderk
 * 
 * 9     2/09/00 4:16p Alexanderk
 * 
 * 8     2/09/00 3:33p Alexanderk
 * 
 * 6     10/06/99 11:41a Rafaelc
 * Moved OnQuickQuote() from Monitor into ILX common events
 * 
 * 5     10/05/99 6:28p Youngh
 * Added OnNavigateURL() in ILX common events.
 * 
 * 4     9/28/99 11:15a Marce
 * o Billy removed the double click event
 * 
 * 3     9/20/99 3:40p Marce
 * o Added OnAlert()
 * 
 * 2     9/20/99 1:09p Marce
 * 
 * 1     9/16/99 4:36p Marce
 * 
 * 12    8/24/99 11:20a Marce
 * o Now uses long instead of enum
 * 
 * 11    7/23/99 6:22p Marce
 * o Should not fire OnAlert()
 * 
 * 10    7/23/99 4:22p Marce
 * o Fire_OnOnAlert() was not initializing the returned HRESULT
 * 
 * 9     7/23/99 3:16p Alexanderk
 * 
 * 8     7/21/99 3:21p Marce
 * o IlxCommon Connection status signatures changed
 * 
 * 7     6/24/99 5:07p Marce
 * o Changed ProgId and interface names from ILXQuoteCtrl to ILXQuote
 * 
 * 6     5/06/99 1:33p Marce
 * o Added AppFramework common properties: DisplayVScrollBar,
 * DisplayHScrollBar, InstanceName, and DataDirectory.  The properties are
 * programmatically accessable but they do nothing and there is no
 * property page for them.  This should be added soon.
 * o Changed OnSymbolChanged() signature and implementation to support
 * passing the symbol change source as an event parameter
 * 
 * 5     4/29/99 5:36p Marce
 * o Methods were returning an uninitialized scode value (0xcdcdcdcd)
 * 
 * 4     4/22/99 1:05p Marce
 * o Moved some stuff around to minimize compile-time dependencies and to
 * improve the organization of the code
 * 
 * 3     4/15/99 10:46a Marce
 * o Now uses ILX DISPIDs
 * 
 * 2     3/26/99 5:50p Andreyk
 * Added stuff for Right mouse button
 * 
 * 1     3/06/99 10:15p Andreyk
 * Created
 */

#include "IlxCommonEventProxy.h"	// Me

#include "IlxCommon.h"				// DISPID_ILX_*

CILXCommonEventProxy::CILXCommonEventProxy(	CComObjectRootEx<CComSingleThreadModel>* pIlxControl,
											CComDynamicUnkArray* pUnkArray)
	: m_pIlxControl(pIlxControl)
	, m_pvec(pUnkArray)
{ 
	_ASSERTE(m_pIlxControl);
	_ASSERTE(m_pvec);
}

void CILXCommonEventProxy::lock()
{
	m_pIlxControl->Lock();
}

void CILXCommonEventProxy::unlock()
{
	m_pIlxControl->Unlock();
}

long CILXCommonEventProxy::getTotalConnections() const
{
	return m_pvec->GetSize();
}

IUnknown* CILXCommonEventProxy::getConnection(long nIndex)
{
	return m_pvec->GetAt(nIndex);
}

HRESULT CILXCommonEventProxy::Fire_OnContextMenu(IDispatch * pTransferableProperties, IDispatch * pSourceMenuInfo, LONG x, LONG y)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[4];
	int nConnections = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			pvars[3] = pTransferableProperties;
			pvars[2] = pSourceMenuInfo;
			pvars[1] = x;
			pvars[0] = y;
			DISPPARAMS disp = { pvars, NULL, 4, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_CONTEXT_MENU, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnSymbolChange(BSTR bstrSymbol, LONG ePropertyChangeSource)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[2];
	int nConnections = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			pvars[1] = bstrSymbol;
			pvars[0] = ePropertyChangeSource;
			DISPPARAMS disp = { pvars, NULL, 2, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_SYMBOL_CHANGE, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnILXError(VARIANT& varErrorObject)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	int nConnections = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			DISPPARAMS disp = { &varErrorObject, NULL, 1, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_ILX_ERROR, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnAlert()
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	int nConnections = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			VariantClear(&varResult);
			DISPPARAMS disp = { NULL, NULL, 0, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_ALERT, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}

	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnNavigateURL(BSTR bstrURL)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[1];
	int nConnections = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			VariantClear(&varResult);
			pvars[0] = bstrURL;
			DISPPARAMS disp = { pvars, NULL, 1, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_NAVIGATE_URL, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}

	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnQuickQuote(BSTR pVal)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[1];
	int nConnections = getTotalConnections();
		
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			VariantClear(&varResult);
			pvars[0] = pVal;
			DISPPARAMS disp = { pvars, NULL, 1, 0 };
			pDispatch->Invoke(0xbc0, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnDrillDown(
  long lDrillDownID,   
  IDispatch* pTransferableProperties,
  BSTR bstrSourceName,
  BSTR bstrTargetName)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[4];
	int nConnections   = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			pvars[3] = lDrillDownID;
			pvars[2] = pTransferableProperties;
			pvars[1] = bstrSourceName;
			pvars[0] = bstrTargetName;
			DISPPARAMS disp = { pvars, NULL, 4, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_DRILL_DOWN, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnValidateLinks(IDispatch* pDrillDownLinks)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[1];
	int nConnections   = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			pvars[0] = pDrillDownLinks;
			DISPPARAMS disp = { pvars, NULL, 1, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_VALIDATE_LINKS, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnSymListChange(BSTR bstrSymList)
{
	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	CComVariant* pvars = new CComVariant[1];
	int nConnections = getTotalConnections();
	
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			pvars[0] = bstrSymList;
			DISPPARAMS disp = { pvars, NULL, 1, 0 };
			pDispatch->Invoke(DISPID_ILX_ON_SYMLIST_CHANGE, IID_NULL, LOCALE_USER_DEFAULT, DISPATCH_METHOD, &disp, &varResult, NULL, NULL);
		}
	}
	delete[] pvars;
	pvars = 0;
	return varResult.scode;
}

HRESULT CILXCommonEventProxy::Fire_OnContextChanged(VARIANT varProps, VARIANT varReason)
{
	CComVariant aryArgs[] = 
	{ 
		varReason,
		varProps
	};
     
	DISPPARAMS dispParams = { aryArgs, 0, sizeof aryArgs / sizeof *aryArgs, 0};

	CComVariant varResult;
	varResult.Clear();
	varResult.scode = S_OK;

	int nConnections = getTotalConnections();
	for (int nConnectionIndex = 0; nConnectionIndex < nConnections; nConnectionIndex++)
	{
		lock();
		CComPtr<IUnknown> sp = getConnection(nConnectionIndex);
		unlock();
		IDispatch* pDispatch = reinterpret_cast<IDispatch*>(sp.p);
		if (pDispatch != NULL)
		{
			varResult.Clear();
			varResult.scode = S_OK;

			pDispatch->Invoke(	DISPID_ILX_ON_CONTEXT_CHANGED, 
								IID_NULL, 
								LOCALE_USER_DEFAULT, 
								DISPATCH_METHOD, 
								&dispParams,
								&varResult, 
								NULL, 
								NULL);
		}
	}

	return varResult.scode;
}
