//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Attributes/AdviceAttribute.cs 3     2/24/06 8:17p Eaddy $
//
// Description:
//
//    Handles weaving advice into a method
//
// Remarks:
//
//    Instead of having the WeavingPhase call AdviceAttribute.Morph(),
//    we first gather all the attributes and pass them into
//    MethodEditor.Weave() which is more efficient.
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Attributes/AdviceAttribute.cs $
 * 
 * 3     2/24/06 8:17p Eaddy
 * o Initial checkin
 */

using System;

using Phx.Aop.Pointcuts;

namespace Phx.Aop.Attributes
{
   /// <summary>
   /// Handles weaving advice into a method
   /// </summary>

   [AttributeUsage(AttributeTargets.Method, AllowMultiple = true)]
   public class AdviceAttribute : Attribute
   {
      AdviceType adviceType;
      string pointcutStr;

      public AdviceAttribute
      (
         AdviceType adviceType, 
         string pointcutStr
      )
      {
         this.adviceType = adviceType;
         this.pointcutStr = pointcutStr;
      }

      /// <summary>
      /// The pointcut associated with this advice
      /// </summary>

      public string Pointcut
      {
         get { return this.pointcutStr; }
         set { this.pointcutStr = value; }
      }

      public AdviceType AdviceType
      {
         get { return this.adviceType; }
         set { this.adviceType = value; }
      }
   }
}
