//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Joinpoints/SimpleJoinpointSignature.cs 11    5/24/06 1:34a Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Joinpoints/SimpleJoinpointSignature.cs $
 * 
 * 11    5/24/06 1:34a Eaddy
 * o Added ArgNames to IJoinpointSignature so that we can
 * pass context args to the advice method
 * 
 * 10    5/04/06 4:59p Eaddy
 * o Changed from IEnumerable<object> to IEnumerator
 * 
 * 9     5/04/06 4:39p Eaddy
 * o Property names now make more sense
 * o Added properties for accessing attributes and args
 * as an object collection
 * 
 * 8     5/04/06 3:04p Eaddy
 * o Changed return type of Attributes property so that we aren't
 * locked in to how we implement it
 * 
 * 7     4/28/06 5:54p Asv2105
 * Fixed so data passed to contructor from other IJoinpoint is actually
 * copied.
 * 
 * 6     4/21/06 1:05p Asv2105
 * Added constructors to allow for cloning any IJoinpoint into a
 * SimpleJoinpoint.
 * 
 * 5     4/03/06 12:22a Eaddy
 * o Changed jp.WithinSignature to jp.Within and made it an
 * IJoinpointSignature.
 * This is needed because we need access to return type, arg types, etc.
 * just like we do for jp.Signature.
 */

using System.Text;
using System.Collections;
using System.Collections.Generic;

namespace Phx.Aop.Joinpoints
{
   public class SimpleJoinpointSignature : IJoinpointSignature
   {
      public List<string> attributes = null;
      public StringBuilder access = null;
      public StringBuilder returnType = null;
      public StringBuilder declaringType = null;
      public StringBuilder memberName = null;
      public StringBuilder[] argTypes = null;
      public StringBuilder[] argNames = null;
      public SimpleSourceLocation location = null;

      public SimpleJoinpointSignature()
      {
         location = new SimpleSourceLocation();
      }

      public SimpleJoinpointSignature(IJoinpointSignature sig)
      {
         AttributeTypes = sig.AttributeTypes;
         access = new StringBuilder(sig.Access.ToString());
         returnType = new StringBuilder(sig.ReturnType.ToString());
         declaringType = new StringBuilder(sig.DeclaringType.ToString());
         memberName = new StringBuilder(sig.MemberName.ToString());
         ArgTypes = sig.ArgTypes;
         ArgNames = sig.ArgNames;
         location = new SimpleSourceLocation(sig.SourceLocation);
      }

      public IEnumerable<string> AttributeTypes
      {
         get { return attributes; }
         set
         {
            if (value == null)
            {
               attributes = null;
            }
            else
            {
               attributes = new List<string>(value);
            }
         }
      }

      public IEnumerator Attributes
      {
         get { return null; } // Not implemented
      }

      public StringBuilder Access
      {
         get { return access; }
         set { access = value; }
      }

      public StringBuilder ReturnType
      {
         get { return returnType; }
         set { returnType = value; }
      }

      public StringBuilder DeclaringType
      {
         get { return declaringType; }
         set { declaringType = value; }
      }

      public StringBuilder MemberName
      {
         get { return memberName; }
         set { memberName = value; }
      }

      public IEnumerable<StringBuilder> ArgTypes
      {
         get { return argTypes; }
         set
         {
            if (value == null)
            {
               argTypes = null;
            }
            else
            {
               SetStringBuilderArray(ref argTypes, value);
            }
         }
      }

      public IEnumerable<StringBuilder> ArgNames
      {
         get { return argNames; }
         set
         {
            if (value == null)
            {
               argNames = null;
            }
            else
            {
               SetStringBuilderArray(ref argNames, value);
            }
         }
      }

      public IEnumerator Args
      {
         get { return null; } // Not implemented
      }

      public ISourceLocation SourceLocation
      {
         get { return location; }
      }

      public static void ClearStringBuilder(ref StringBuilder sb)
      {
         if (sb != null)
            sb.Length = 0;
      }

      public static void SetStringBuilder(ref StringBuilder sb, string val)
      {
         if (sb == null)
         {
            sb = new StringBuilder(val);
         }
         else
         {
            sb.Length = 0;
            sb.Append(val);
         }
      }

      void SetStringBuilderArray(   ref StringBuilder[] arrayStringBuilders,  
                                    IEnumerable<StringBuilder> enumStringBuilders)
      {
         // First get the count
         IEnumerator<StringBuilder> e = enumStringBuilders.GetEnumerator();
         int count = 0;
         while (e.MoveNext())
            ++count;

         e.Reset();

         if (arrayStringBuilders == null || arrayStringBuilders.Length != count)
         {
            // Only create a new array if necessary
            arrayStringBuilders = new StringBuilder[count];
         }

         int i = 0;
         while (e.MoveNext())
         {
            SetStringBuilder(ref arrayStringBuilders[i++], e.Current.ToString());
         }
      }

      public void SetAccess(string val)
      {
         SetStringBuilder(ref access, val);
      }

      public void SetReturnType(string val)
      {
         SetStringBuilder(ref returnType, val);
      }

      public void SetDeclaringType(string val)
      {
         SetStringBuilder(ref declaringType, val);
      }

      public void SetMemberName(string val)
      {
         SetStringBuilder(ref memberName, val);
      }

      public void Reset()
      {
         attributes = null;

         if (access != null)
            access.Length = 0;

         if (returnType != null)
            returnType.Length = 0;

         if (declaringType != null)
            declaringType.Length = 0;

         if (memberName != null)
            memberName.Length = 0;

         argTypes = null;
         argNames = null;

         location.Reset();
      }

      public override string ToString()
      {
         return DeclaringType + "." + MemberName;
      }
   }
}
