//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Joinpoints/SimpleSourceLocation.cs 6     5/04/06 3:32p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Joinpoints/SimpleSourceLocation.cs $
 * 
 * 6     5/04/06 3:32p Eaddy
 * o Joinpoints now have source column so that we can weave
 * the same advice multiple times at the same joinpoint line 
 * 
 * 5     4/28/06 5:54p Asv2105
 * Fixed so data passed to contructor from other IJoinpoint is actually
 * copied.
 * 
 * 4     4/21/06 1:05p Asv2105
 * Added constructors to allow for cloning any IJoinpoint into a
 * SimpleJoinpoint.
 * 
 * 3     4/03/06 12:22a Eaddy
 * o Changed jp.WithinSignature to jp.Within and made it an
 * IJoinpointSignature.
 * This is needed because we need access to return type, arg types, etc.
 * just like we do for jp.Signature.
 * 
 * 2     3/13/06 8:57p Eaddy
 * o Fixed bug where the method name was incorrect
 * 
 * 1     3/03/06 2:52p Eaddy
 */

using System.Text;

namespace Phx.Aop.Joinpoints
{
   public class SimpleSourceLocation : ISourceLocation
   {
      public StringBuilder filePath;
      public uint line;
      public uint column;

      public SimpleSourceLocation() {
          // do nothing
      }

      public SimpleSourceLocation(ISourceLocation sl) {
          filePath = new StringBuilder(sl.FilePath.ToString());
          line = sl.Line;
      }

      public StringBuilder FilePath
      {
         get { return filePath; }
         set { filePath = value; }
      }

      public string FileName
      {
         get { return System.IO.Path.GetFileName(filePath.ToString()); }
      }

      public uint Line
      {
         get { return line; }
         set { line = value; }
      }

      public uint Column
      {
         get { return column; }
         set { column = value; }
      }

      public void Reset()
      {
         if (filePath != null)
            filePath.Length = 0;

         line = 0;
         column = 0;
      }

      public override string ToString()
      {
         StringBuilder sb = new StringBuilder();

         sb.Append("[File: ");
         sb.Append(FileName);
         sb.Append(", Line: ");
         sb.Append(Line);
         sb.Append(']');

         return sb.ToString();
      }
   }
}
