//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/PatternPointcutExpression.cs 6     5/11/06 12:07p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/PatternPointcutExpression.cs $
 * 
 * 6     5/11/06 12:07p Eaddy
 * o Now considers 'int' and 'System.Int32' the same type
 * 
 * 5     3/01/06 2:17a Eaddy
 * o Now uses Phx.Aop library for pointcuts, advice and joinpoint matching
 * 
 * 4     2/26/06 9:12p Eaddy
 * o Now supports joinpoint matching
 * o Changed tabs to 3 spaces to conform with MS standard
 * 
 * 3     2/26/06 7:48p Eaddy
 * 
 * 2     2/24/06 8:17p Eaddy
 * o Initial checkin
 */

using Dbg = System.Diagnostics.Debug;

using Phx.Aop.Joinpoints;

namespace Phx.Aop.Pointcuts
{
   /// <summary>
   /// A string-based expression
   /// </summary>
   public class PatternPointcutExpression : PointcutExpression
   {
      protected WildcardMatch matcher = null;

      public PatternPointcutExpression()
      { }

      public PatternPointcutExpression(string val)
      {
         parse(val);
      }

      public bool matchesAnything
      {
         get 
         {
            return matcher == null || matcher.MatchesAnything();
         }
      }

      public override PointcutExpression parse(string expr)
      {
         if (expr == null || expr.Length == 0)
         {
            // We will match anything
            matcher = null;
            return this;
         }
         else
         {
            matcher = new WildcardMatch(expr);
            return this;
         }
      }

      public override bool innerMatches(System.Text.StringBuilder rhs)
      {
         if (rhs == null)
            return false;
         else if (matcher == null)
            return true;
         else
            return matcher.IsAMatch(rhs);
      }

      public override bool innerMatches(string rhs)
      {
         if (rhs == null)
            return false;
         else if (matcher == null)
            return true;
         else
            return matcher.IsAMatch(rhs);
      }

      public override bool innerMatches(IJoinpoint rhs)
      {
         throw new System.Exception("The method or operation is not implemented.");
      }

      public override bool innerMatches(PointcutExpression rhs)
      {
         if (rhs == null)
            return false;
         else if (matcher == null)
            return true;

         PatternPointcutExpression rhs2 = rhs as PatternPointcutExpression;
         return rhs2 == null ? false : rhs2.matcher.IsAMatch(rhs.ToString());
      }

      public override string innerToString()
      {
         if (matcher == null)
            return "";
         else
            return matcher.ToString();
      }
   }
}
