//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/SignatureExpression.cs 21    5/11/06 11:19p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/SignatureExpression.cs $
 * 
 * 21    5/11/06 11:19p Eaddy
 * o Minor changes to diagnostic messages
 * 
 * 20    5/11/06 5:04p Eaddy
 * o Can now leave out access specifier
 * 
 * 19    5/11/06 3:38p Eaddy
 * o Now handles the case when the user
 * specifies "()" (no args)  and "(,)" (invalid
 * empty args)
 * 
 * 18    5/11/06 1:56p Eaddy
 * o Set pointcuts now require parameters
 * o We now require the access to be specified
 * 
 * 17    5/11/06 12:43p Eaddy
 * o Now verifies that the access specifier
 * is valid
 * o Put Phx.Aop before output strings
 * 
 * 16    5/11/06 12:09p Eaddy
 * o Minor
 * 
 * 15    5/11/06 12:07p Eaddy
 * o Now considers 'int' and 'System.Int32' the same type
 * 
 * 14    5/04/06 4:39p Eaddy
 * o Property names now make more sense
 * o Added properties for accessing attributes and args
 * as an object collection
 * 
 * 13    5/04/06 3:04p Eaddy
 * o Changed return type of Attributes property so that we aren't
 * locked in to how we implement it
 * 
 * 12    4/03/06 12:23a Eaddy
 * o Now supports WithinCode and Within
 * 
 * 11    4/02/06 7:58p Eaddy
 * o Reversed last change. Treating ".ctor" like "new" is now done by
 * PhxJoinpointSignature
 * 
 * 10    4/02/06 4:58p Rr2318
 * 
 * 9     3/01/06 2:17a Eaddy
 * o Now uses Phx.Aop library for pointcuts, advice and joinpoint matching
 * 
 * 8     2/26/06 10:01p Eaddy
 * o Renamed
 * o Can now handle call, exec, get, and set joinpoints
 * 
 * 7     2/26/06 9:39p Eaddy
 * o Now uses containment instead of inheritance (makes more sense)
 * 
 * 6     2/26/06 9:26p Eaddy
 * o Renamed
 * 
 * 5     2/26/06 9:20p Eaddy
 * o Trivial
 * 
 * 4     2/26/06 9:12p Eaddy
 * o Now supports joinpoint matching
 * o Changed tabs to 3 spaces to conform with MS standard
 * 
 * 3     2/26/06 7:48p Eaddy
 * 
 * 2     2/24/06 8:17p Eaddy
 * o Initial checkin
 */

using System.Text.RegularExpressions;

using Dbg = System.Diagnostics.Debug;

using Phx.Aop.Joinpoints;

namespace Phx.Aop.Pointcuts
{
   /// <summary>
   /// call(* *.*(int))
   /// </summary>
   public abstract class SignatureExpression : PointcutExpression
   {
      static Regex regexSigWParams = 
         new Regex(@"^((\[\w+\] )*)((public|private|protected|internal|\*) )?(([\w\.\*]+) )?(.*)\.(.+)\((.*)\)$", 
            RegexOptions.Compiled);
      static Regex regexSigWOParams = 
         new Regex(@"^((\[\w+\] )*)((public|private|protected|internal|\*) )?(([\w\.\*]+) )?(.*)\.(.+)$", 
            RegexOptions.Compiled);

      PatternPointcutExpression[] attrPatternExpressions = null;
      PatternPointcutExpression sigAccess = null;
      TypeExpression sigReturnType = null;
      TypeExpression sigDeclaringType = null;
      PatternPointcutExpression sigMemberName = null;
      TypeExpression[] argTypes = null;
      JoinpointType joinpointType;

      public SignatureExpression(JoinpointType joinpointType)
      {
         this.joinpointType = joinpointType;
      }

      public bool IgnoreParams()
      {
         return joinpointType == JoinpointType.GET;
      }

      public abstract string Name { get; }

      public override PointcutExpression parse(string expr)
      {
         Dbg.Assert(expr != null && expr.Length != 0);
         if (expr == null || expr.Length == 0)
         {
            System.Console.WriteLine("Phx.Aop: Pointcut declaration is empty or null.");
            return null;
         }

         // e.g. "* *.*(int)"

         //match   0: '[Hey] [Jack] public void A.B.C(int, ..)'
         //  group 0: '[Hey] [Jack] public void A.B.C(int, ..)'
         //    cap 0: '[Hey] [Jack] public void A.B.C(int, ..)'
         //  group 1: '[Hey] [Jack] '
         //    cap 0: '[Hey] [Jack] '
         //  group 2: '[Jack] '
         //    cap 0: '[Hey] '
         //    cap 1: '[Jack] '
         //  group 3: 'public '
         //    cap 0: 'public '
         //  group 4: 'public'
         //    cap 0: 'public'
         //  group 5: 'void '
         //    cap 0: 'void '
         //  group 6: 'void'
         //    cap 0: 'void'
         //  group 7: 'A.B'
         //    cap 0: 'A.B'
         //  group 8: 'C'
         //    cap 0: 'C'
         //  group 9: 'int, ..'
         //    cap 0: 'int, ..'
         Regex regex = IgnoreParams() ? regexSigWOParams : regexSigWParams;
         MatchCollection matches = regex.Matches(expr);
         //dumpMatches(matches);

         if (matches.Count != 1)
         {
            System.Console.WriteLine("Phx.Aop: Invalid pointcut expression: '{0}'.", expr);
            return null; // Invalid expr
         }

         GroupCollection groups = matches[0].Groups;
         Dbg.Assert(groups != null);

         int expected = IgnoreParams() ? 9 : 10;

         // A valid match will have exactly 9-10 groups
         if (groups.Count != expected)
         {
            System.Console.WriteLine("Phx.Aop: Unexpected pointcut parsing error: '{0}'. " +
                "Expected {1} groups but got {2}.",
                expr, expected, groups.Count);
            return null; // Invalid expr
         }

         int totalAttrs = groups[2].Captures.Count;
         attrPatternExpressions = new PatternPointcutExpression[totalAttrs];
         for (int i = 0; i < totalAttrs; ++i)
         {
            // Remove brackets and trailing space
            string attrWithoutBrackets = groups[2].Captures[i].Value;
            attrWithoutBrackets = attrWithoutBrackets.Substring(1, attrWithoutBrackets.Length - 3);
            attrPatternExpressions[i] = new PatternPointcutExpression(attrWithoutBrackets);
         }
         
         sigAccess = new PatternPointcutExpression(groups[4].Value.Trim());

         if (!VerifyAccessSpecifer(expr, sigAccess))
         {
            return null;
         }

         sigReturnType = new TypeExpression(groups[5].Value.Trim());
         sigDeclaringType = new TypeExpression(groups[7].Value.Trim());
         sigMemberName = new PatternPointcutExpression(groups[8].Value.Trim());

         if (!IgnoreParams())
         {
            string[] args = groups[9].Value.Trim().Split(',');
            
            for(int i = 0; i < args.Length; ++i)
            {
               args[i] = args[i].Trim();

               if (args[i] == string.Empty)
               {
                  if (args.Length > 1)
                  {
                     System.Console.WriteLine("Phx.Aop: Pointcut parsing error: '{0}'. " +
                         "Parameter argument {1} is empty.",
                         expr, i+1);
                     return null;
                  }
                  else
                  {
                     // If there is only one argument and its empty then
                     // don't complain since this is the default when
                     // the parameter string is empty
                     args = null;
                     break;
                  }
               }
            }

            if (args != null && args.Length > 0)
            {
               argTypes = new TypeExpression[args.Length];
               for(int i = 0; i < argTypes.Length; ++i)
               {
                  argTypes[i] = new TypeExpression(args[i]);
               }
            }
         }

         return this;
      }

      public override string innerToString()
      {
         System.Text.StringBuilder sb = new System.Text.StringBuilder();

         sb.Append(Name);
         sb.Append('(');

         if (attrPatternExpressions != null)
         {
            foreach (PatternPointcutExpression pat in attrPatternExpressions)
            {
               sb.Append('[');
               sb.Append(pat.ToString());
               sb.Append("] ");
            }
         }

         string access = sigAccess == null ? "(null)" : sigAccess.ToString();
         if (access.Length != 0)
         {
            sb.Append(access);           // Access
            sb.Append(' ');
         }

         sb.Append(sigReturnType == null ? "(null)" : sigReturnType.ToString()); // Return Type
         sb.Append(' ');
         sb.Append(sigDeclaringType == null ? "(null)" : sigDeclaringType.ToString()); // Type
         sb.Append('.');
         sb.Append(sigMemberName == null ? "(null)" : sigMemberName.ToString()); // MemberName

         if (!IgnoreParams() && argTypes != null)
         {
            sb.Append('(');
            for(int i = 0; i < argTypes.Length; ++i)
            {
               if (i > 0)
                  sb.Append(",");

               sb.Append(argTypes[i].ToString()); // Params
            }
            sb.Append(')');
         }

         sb.Append(')');

         return sb.ToString();
      }

      public override bool innerMatches(string rhs)
      {
         throw new System.Exception("Phx.Aop: The method or operation is not implemented.");
      }

      public override bool innerMatches(IJoinpoint rhs)
      {
         if (rhs.JoinpointType != joinpointType && joinpointType != JoinpointType.STATEMENT)
             return false;

         IJoinpointSignature sigToUse;
         if (joinpointType == JoinpointType.STATEMENT)
         {
             sigToUse = rhs.Within;
         }
         else
         {
             sigToUse = rhs.Signature;
         }

         if (attrPatternExpressions.Length > 0)
         {
            foreach (PatternPointcutExpression patLhs in attrPatternExpressions)
            {
               if (patLhs.ToString().Equals(".."))
                  continue;

               // Method has no attributes but we expected some
               if (sigToUse.AttributeTypes == null)
                  return false;

               bool foundMatch = false;
               foreach (string rhsAttr in sigToUse.AttributeTypes)
               {
                  if (patLhs.matches(rhsAttr))
                  {
                     foundMatch = true;
                     break;
                  }
               }
               if (!foundMatch)
               {
                  return false;
               }
            }
         }

         if (  sigAccess != null &&
               !sigAccess.matchesAnything &&
               !sigAccess.matches(sigToUse.Access))
            return false;

         if (  sigReturnType != null &&
               !sigReturnType.matchesAnything &&
               !sigReturnType.matches(sigToUse.ReturnType))
            return false;

         if (  sigMemberName != null &&
               !sigMemberName.matchesAnything &&
               !sigMemberName.matches(sigToUse.MemberName))
            return false;

         if (  sigDeclaringType != null &&
               !sigDeclaringType.matchesAnything &&
               !sigDeclaringType.matches(sigToUse.DeclaringType))
            return false;

         if (IgnoreParams() || argTypes == null)
            return true;

         if (  argTypes != null &&
               argTypes.Length == 1 &&
               argTypes[0].ToString().Equals(".."))
            return true;

         int i = 0;
         foreach(System.Text.StringBuilder rhsArgType in sigToUse.ArgTypes)
         {
            if (i >= argTypes.Length)
               break;

            if (!argTypes[i++].matches(rhsArgType))
               return false;
         }

         if (i != argTypes.Length)
            return false;

         return true;
      }

      public override bool innerMatches(PointcutExpression rhs)
      {
         if (rhs == null)
            return false;

         SignatureExpression expRhs = rhs as SignatureExpression;
         return expRhs == null ? false : innerMatches(expRhs);
      }

      public bool innerMatches(SignatureExpression rhs)
      {
         throw new System.Exception("Phx.Aop: The method or operation is not implemented.");
      }

      public static void dumpMatches(string pattern, string expr)
      {
         System.Console.WriteLine("Pattern  : " + pattern);
         System.Console.WriteLine("Expr     : " + expr);
         dumpMatches(new Regex(pattern).Matches(expr));
      }

      public static void dumpMatches(MatchCollection matches)
      {
         for (int mat = 0; mat < matches.Count; ++mat)
         {
            System.Console.WriteLine("match   {0}: '{1}'",
                mat, matches[mat].Value);

            for (int grp = 0; grp < matches[mat].Groups.Count; ++grp)
            {
               System.Console.WriteLine("  group {0}: '{1}'", grp,
                   matches[mat].Groups[grp].Value);

               for (int cap = 0; cap < matches[mat].Groups[grp].Captures.Count; ++cap)
               {
                  System.Console.WriteLine("    cap {0}: '{1}'", cap,
                      matches[mat].Groups[grp].Captures[cap].Value);
               }
            }
         }
      }

      static bool VerifyAccessSpecifer(string origExpr, PatternPointcutExpression access)
      {
         string[] validAccess = { "public", "protected", "private", "internal" };

         if (access == null)
         {
            return false;
         }
         else if (access.matchesAnything)
         {
            return true;
         }

         foreach(string valid in validAccess)
         {
            if (access.matches(valid))
               return true;
         }

         System.Console.Write("Phx.Aop: Pointcut parsing error: '{0}'. " +
             "Invalid access '{1}'.  Expected one of: ",
             origExpr, access);
         
         for(int i = 0; i < validAccess.Length; ++i)
         {
            if (i == validAccess.Length - 1)
               System.Console.Write(", or " + validAccess[i]);
            else if (i > 0)
               System.Console.Write(", " + validAccess[i]);
            else 
               System.Console.Write(validAccess[i]);
         }

         System.Console.WriteLine('.');

         return false;
      }
   }
}
