//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/TypeExpression.cs 1     5/11/06 12:07p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Aop/Pointcuts/Expressions/TypeExpression.cs $
 * 
 * 1     5/11/06 12:07p Eaddy
 * o Now considers 'int' and 'System.Int32' the same type
 */

using System.Collections.Generic;

using Dbg = System.Diagnostics.Debug;

using Phx.Aop.Joinpoints;

namespace Phx.Aop.Pointcuts
{
   /// <summary>
   /// A string-based expression
   /// </summary>
   public class TypeExpression : PatternPointcutExpression
   {
      public TypeExpression()
      { }

      public TypeExpression(string val)
      {
         parse(val);
      }

      public override bool innerMatches(System.Text.StringBuilder rhs)
      {
         if (rhs == null)
         {
            return false;
         }
         if (base.innerMatches(rhs))
         {
            return true;
         }
         else
         {
            string newName;
            bool found = mapFQN2keyword.TryGetValue(rhs.ToString(), out newName);
            if (!found)
               return false;

            return base.innerMatches(newName);
         }
      }

      public override bool innerMatches(string rhs)
      {
         if (rhs == null)
         {
            return false;
         }
         if (base.innerMatches(rhs))
         {
            return true;
         }
         else
         {
            string newName;
            bool found = mapFQN2keyword.TryGetValue(rhs, out newName);
            if (!found)
               return false;

            return base.innerMatches(newName);
         }
      }

      public override bool innerMatches(IJoinpoint rhs)
      {
         throw new System.Exception("The method or operation is not implemented.");
      }

      public override bool innerMatches(PointcutExpression rhs)
      {
         if (rhs == null)
            return false;
         else if (matcher == null)
            return true;

         TypeExpression rhs2 = rhs as TypeExpression;
         return rhs2 == null ? false : rhs2.innerMatches(rhs.ToString());
      }

      static Dictionary<string, string> mapFQN2keyword;

      static TypeExpression()
      {
         mapFQN2keyword = new Dictionary<string, string>();
         mapFQN2keyword[typeof(string).FullName]   = "string";
         mapFQN2keyword[typeof(object).FullName]   = "object";
         mapFQN2keyword[typeof(short).FullName]    = "short";
         mapFQN2keyword[typeof(ushort).FullName]   = "ushort";
         mapFQN2keyword[typeof(int).FullName]      = "int";
         mapFQN2keyword[typeof(uint).FullName]     = "uint";
         mapFQN2keyword[typeof(long).FullName]     = "long";
         mapFQN2keyword[typeof(ulong).FullName]    = "ulong";
         mapFQN2keyword[typeof(char).FullName]     = "char";
         mapFQN2keyword[typeof(bool).FullName]     = "bool";
         mapFQN2keyword[typeof(float).FullName]    = "float";
         mapFQN2keyword[typeof(double).FullName]   = "double";
         mapFQN2keyword[typeof(byte).FullName]     = "byte";
         mapFQN2keyword[typeof(sbyte).FullName]    = "sbyte";
         mapFQN2keyword[typeof(decimal).FullName]  = "decimal";
      }
   }
}
