//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Aop/PhxSourceLocation.cs 7     5/04/06 3:32p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Aop/PhxSourceLocation.cs $
 * 
 * 7     5/04/06 3:32p Eaddy
 * o Joinpoints now have source column so that we can weave
 * the same advice multiple times at the same joinpoint line 
 * 
 * 6     4/03/06 12:23a Eaddy
 * o Changed jp.WithinSignature to jp.Within and made it an
 * IJoinpointSignature.
 * This is needed because we need access to return type, arg types, etc.
 * just like we do for jp.Signature.
 * 
 * 5     4/02/06 10:53p Eaddy
 * o Now displays new instead of .ctor
 * 
 * 4     4/02/06 10:30p Eaddy
 * o Now shows "new" instead of ".ctor"
 * 
 * 3     3/03/06 5:05p Eaddy
 * o Fixed bug where sometimes the file name was empty
 * 
 * 2     3/03/06 4:19p Eaddy
 * o Removed unused references
 * 
 * 1     3/03/06 3:00p Eaddy
 */

using System.Text;

using Phx.Morph;

namespace Phx.Aop.Joinpoints
{
   public class PhxSourceLocation : ISourceLocation
   {
      Phx.IR.Instr instr = null;

      SimpleSourceLocation location = new SimpleSourceLocation();

      public Phx.IR.Instr Instr
      {
         get { return instr; }
         set
         {
            this.instr = value;

            // Reset state variables
            if (location != null)
               location.Reset();
         }
      }

      public StringBuilder FilePath
      {
         get 
         {
            if (location.filePath != null && location.filePath.Length > 0)
               return location.filePath;

            Helpers.SafeGetSourceFilePath(instr, ref location.filePath);

            return location.filePath; 
         }
      }

      public string FileName
      {
         get { return System.IO.Path.GetFileName(FilePath.ToString()); }
      }

      public uint Line
      {
         get 
         {
            if (location.line != 0)
               return location.line;

            location.line = Helpers.SafeGetSourceLine(instr);

            return location.line; 
         }
      }

      public uint Column
      {
         get
         {
            if (location.column != 0)
               return location.column;

            location.column = Helpers.SafeGetSourceColumn(instr);

            return location.column;
         }
      }

      public override string ToString()
      {
         StringBuilder sb = new StringBuilder();

         sb.Append("[File: ");
         sb.Append(FileName);
         sb.Append(", Line: ");
         sb.Append(Line);
         sb.Append(']');

         return sb.ToString();
      }
   }
}
