//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/MorphExecAttributeHandler.cs 4     5/24/06 9:53p Eaddy $
//
// Description:
//
//    Uses Reflection to execute arbitrary code at morph time
//
// Remarks:
//
// ISSUE-TODO-t-marcea-2005/07/23-Phx#? Remove Reflection
// for unmanaged implementation
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/MorphExecAttributeHandler.cs $
 * 
 * 4     5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 3     3/03/06 3:39p Eaddy
 * o Name change
 * 
 * 2     2/24/06 8:23p Eaddy
 * o Now contains Attribute instead of deriving from it
 * o Uses Phx.Aop.dll which has generic AOP attributes
 * o Added checkin comment
 */

using System;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;
using Phx.Aop.Attributes;

namespace Phx.Morph.Attributes
{
   public class MorphExecAttributeHandler : MorphTypeAttributeHandlerBase, IAttributeHandler
   {
      public MorphExecAttributeHandler()
      { }

      public string AttributeName { get { return "Phx.Aop.Attributes.MorphExecAttribute"; } }

      public IExtensionAttribute Create(Phx.Syms.AttrSym attrSym, Type attrType, AttributeHelper helper)
      {
         MorphExecAttributeHandler handler = new MorphExecAttributeHandler();
         return handler;
      }

      public override bool
      Morph
      (
         ModuleEditor moduleEditor,
         TypeEditor typeEditor
      )
      {
         // Verify args

         Dbg.Assert(moduleEditor != null);
         Dbg.Assert(typeEditor != null);

         // Sanity checks

         Dbg.Assert(AttachedSym != null);
         Dbg.Assert(AttachedSym.IsFuncSym);
         Dbg.Assert(AttachedSym.AsFuncSym.EnclosingAggrType != null);
         Dbg.Assert(AttachedType != null);

         // Attribute must be attached to a method

         Dbg.Assert(AttachedType.IsFuncType);

         // We only support calling static methods

         Dbg.Assert(!AttachedType.AsFuncType.IsInstanceMethod);

         // Step 1:  Convert the method's class type from Phx to CLR

         System.Type ownerType =
            ReflectionHelpers.ConvertPhxTypeToClrType(AttachedSym.AsFuncSym.EnclosingAggrType);

         // Shouldn't be null since we load the assemblies for all the
         // module references

         Dbg.Assert(ownerType != null);
         if (ownerType == null)
         {
            return false;
         }

         // Step 2:  Use Reflection to bind to the method

         System.Reflection.MethodInfo methodToExec =
            ownerType.GetMethod(Name,
            System.Reflection.BindingFlags.NonPublic |
            System.Reflection.BindingFlags.Public |
            System.Reflection.BindingFlags.Static);

         // We expect to find the method since that's how we got
         // the attribute in the first place

         Dbg.Assert(methodToExec != null);
         if (methodToExec == null)
         {
            return false;
         }

         // Step 3:  Execute the method on the type editor object

         try
         {
            methodToExec.Invoke(null, new object[] { typeEditor });
         }
         catch (System.Reflection.TargetInvocationException ex)
         {
            Helpers.WriteError("Failed to execute method '{0}': {1}",
               methodToExec.ToString(),
               ex.Message);
         }

         // We ignore the return value

         return true;
      }
   }
}
