//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/ReplaceAttributeHandler.cs 8     5/24/06 9:53p Eaddy $
//
// Description:
//
//    Custom attribute for replacing method bodies
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/Handlers/ReplaceAttributeHandler.cs $
 * 
 * 8     5/24/06 9:53p Eaddy
 * o Better diagnostics when the ctor param list for the inner
 * attribute doesn't match any ctor
 * 
 * 7     5/24/06 8:56a Eaddy
 * o Changed file names but forgot to change class
 * names
 * 
 * 6     5/24/06 1:35a Eaddy
 * o Switched names for MethodWeaver and MethodEditor
 * so now they make sense
 * o Fixed bug where AppendFuncArg didn't work when
 * weaving static methods
 * 
 * 5     3/15/06 1:20p Eaddy
 * o Change message
 * 
 * 4     3/03/06 3:39p Eaddy
 * o Name change
 * 
 * 3     3/03/06 2:57p Asv2105
 * Refactored so all classes refer to base class of MethodWeaver rather
 * than MethodEditor, and create new instances through
 * MethodWeaverFactory.
 * 
 * 2     2/24/06 8:23p Eaddy
 * o Now contains Attribute instead of deriving from it
 * o Uses Phx.Aop.dll which has generic AOP attributes
 * o Added checkin comment
 */

using System;
using System.Diagnostics;

using Dbg = System.Diagnostics.Debug;

using Phx.Morph.Editors;
using Phx.Aop.Attributes;

namespace Phx.Morph.Attributes
{
   public class ReplaceAttributeHandler : MorphTypeAttributeHandlerBase, IAttributeHandler
   {
      public ReplaceAttributeHandler()
      { }

      public string AttributeName { get { return "Phx.Aop.Attributes.ReplaceAttribute"; } }

      public IExtensionAttribute Create(Phx.Syms.AttrSym attrSym, Type attrType, AttributeHelper helper)
      {
         return new MorphExecAttributeHandler();
      }

      /// <summary>
      /// Replaces the method
      /// </summary>
      /// <param name="moduleEditor">Module where the type to be morphed lives</param>
      /// <param name="typeEditor">The Type to be morphed</param>
      /// <returns>true on success</returns>

      public override bool
      Morph
      (
         ModuleEditor moduleEditor,
         TypeEditor typeEditor
      )
      {
         Dbg.Assert(this.AttachedSym != null);
         Dbg.Assert(this.AttachedSym.IsFuncSym);

         MethodEditor methodToReplace =
            typeEditor.FindMethod(this.AttachedSym.AsFuncSym);
         if (methodToReplace == null)
         {
            Helpers.WriteError(
               "Failed to replace method. Method '{0}.{1}' not found or signature is differnt.",
               typeEditor.ToString(),
               this.Name);

            return false;
         }

         MethodEditor methodToCopy = MethodEditorFactory.CreateEditor(AttachedSym.AsFuncSym);
         if (methodToCopy == null)
         {
            // Already reported error

            return false;
         }

         return methodToReplace.ReplaceWith(methodToCopy);
      }
   }
}
