//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Attributes/IExtensionAttribute.cs 2     2/24/06 8:24p Eaddy $
//
// Description:
//
//    Base interface for all extension custom attributes including:
//
//       Morph extensions: AddAttribute, ReplaceAttribute,
//          ThisAttribute, and MorphExecAttribute
//
//       AOP extensions: AdviceAttribute, PointcutAttribute,
//          ThisAttribute, and the JoinpointAttributes
//
// Remarks:
//
// ISSUE-HACK-t-marcea-2005/08/03-Subclasses of IExtensionAttribute
// must not have two constructors that have the same number of args
// but different types.  AttributeHelper.InstantiateAttribute()
// uses the constructor arg count to uniquely identify the
// constructor (see the issue description in that file).
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Attributes/IExtensionAttribute.cs $
 * 
 * 2     2/24/06 8:24p Eaddy
 * o Added checkin comment
 */

namespace Phx.Morph.Attributes
{
   /// <summary>
   /// Base interface for all morph extension attributes
   /// </summary>
   /// <remarks>
   /// All custom morph attributes must derive from
   /// IExtensionAttribute in order for the Morph tool to
   /// recognize them.  The interface provides the
   /// Init(Phx.Syms.Sym ownerSym) method so that we can
   /// associate the symbol of the thing that the attribute
   /// is attached to with the attribute.  For example, if
   /// the attribute is attached to a field, the ownerSym
   /// will be the FieldSym for that field.
   /// </remarks>

   public interface IExtensionAttribute
   {
      /// <summary>
      /// Called after attribute creation to specify the
      /// 'thing' (ownerSym) that this attribute is attached
      /// to and to initialize the attribute
      /// </summary>
      /// <param name="ownerSym">
      /// The Phx symbol that this attribute is attached to
      /// or null if the attribute was created via the
      /// command-line</param>
      /// <param name="globalTypeTable">
      /// Type table used to resolve the type
      /// associated with the attribute
      /// </param>
      /// <returns>true on success</returns>

      bool
      Init
      (
         Phx.Syms.Sym ownerSym,
         Phx.Types.Table globalTypeTable
      );

      /// <returns>
      /// true if Init() completed successfully
      /// </returns>
      bool
      IsInited
      (
      );
   }
}
