//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/Morpher/Phx.Morph/Editors/BreakpointWeaver.cs 14    5/24/06 9:56p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Morpher/Phx.Morph/Editors/BreakpointWeaver.cs $
 * 
 * 14    5/24/06 9:56p Eaddy
 * o Now tries to resolve assemblies that can't be loaded
 * o Improved instantiation of custom attributes and finding
 * the correct ctor for a given blob
 * 
 * 13    5/24/06 2:05p Eaddy
 * o When setting a breakpoint for after advice, we
 * need to set it on the line after the jp's line
 * 
 * 12    5/24/06 8:56a Eaddy
 * o Changed file names but forgot to change class
 * names
 * 
 * 11    5/05/06 6:18p Eaddy
 * o Did some renaming to improve clarity
 * o Can now list current advice and enable/disable (although
 * enable/disable doesn't actually work yet)
 * 
 * 10    4/21/06 12:33p Eaddy
 * o Now addBreakpoint only supplies the jp and adviceMethod
 * 
 * 9     4/11/06 3:39p Asv2105
 * Added ability to call code in debugged process
 * 
 * 8     4/03/06 12:52a Asv2105
 * Added a bunch of code for executing code in the client process when a
 * breakpoint is hit that doesn't entirely work
 * 
 * 7     4/03/06 12:23a Eaddy
 * o Changed jp.WithinSignature to jp.Within and made it an
 * IJoinpointSignature.
 * This is needed because we need access to return type, arg types, etc.
 * just like we do for jp.Signature.
 * 
 * 6     4/02/06 7:57p Eaddy
 * o Changed diagnostics
 * 
 * 5     4/02/06 4:34p Asv2105
 * Added ability for BreakpointWeaver to pass MethodInfo to the
 * coordinator
 * 
 * Adjusted ICoordinator, IEventListener interfaces to match reality
 * 
 * Added ability for Coordinator to call static methods when a breakpoint
 * fires and pass off to EventMgr for non-static methods
 * 
 * 4     3/15/06 2:40p Eaddy
 * o Fixed stack overflow
 * 
 * 3     3/15/06 2:37p Eaddy
 * o Can now get breakpoint callback value
 * 
 * 2     3/15/06 2:31p Eaddy
 * o Now uses a set breakpoint callback
 */

using System;
using Phx.Aop;

namespace Phx.Morph.Editors
{
   public class BreakpointWeaver : MethodEditor
   {
      static Phx.Aop.BreakpointWeaverCallback setBreakpointCallback = null;

      public static Phx.Aop.BreakpointWeaverCallback BreakpointWeaverCallback
      {
         get { return setBreakpointCallback; }
         set { setBreakpointCallback = value; }
      }

      public
      BreakpointWeaver
      (
         TypeEditor typeEditor,
         Phx.Syms.FuncSym funcSym
      )
         : base(typeEditor, funcSym)
      { }

      /// <summary>
      /// Create a method editor using the FuncSym of the method
      /// we want to edit
      /// </summary>

      public
      BreakpointWeaver
      (
         Phx.Syms.FuncSym funcSym
      )
         : base(funcSym)
      { }

      public override bool
      Weave
      (
           Phx.IR.Instr insertAfterInstr,
           MethodEditor methodToCall,
           Phx.Aop.AdviceType adviceType,
           Phx.Aop.Joinpoints.IJoinpoint jp
      )
      {
         uint lineToSetBreakpoint = jp.Within.SourceLocation.Line;
         if (adviceType == AdviceType.after &&
            jp.JoinpointType != Phx.Aop.Joinpoints.JoinpointType.EXECUTION)
         {
            ++lineToSetBreakpoint;
         }

         if (setBreakpointCallback != null)
         {
            System.Reflection.MethodInfo methodInfoToCall =
               ReflectionHelpers.ConvertPhxFuncSymToMethodInfo(methodToCall.FuncSym);
            if (methodInfoToCall == null)
            {
               Helpers.WriteError(
                  "Failed to add breakpoint - {0}, {1}: Advice method '{2}' could not be resolved.",
                  jp.Within.SourceLocation.FileName.ToString(),
                  lineToSetBreakpoint,
                  methodToCall.ToString());

               return false;
            }

            Helpers.WriteOutput("Adding breakpoint - {0}, {1}.",
               jp.Within.SourceLocation.FileName.ToString(),
               lineToSetBreakpoint);

            // Set the breakpoint
            setBreakpointCallback(jp, lineToSetBreakpoint, methodInfoToCall);
         }
         else
         {
            Helpers.WriteError(
               "Failed to add breakpoint - {0}, {1}: Breakpoint callback is null.",
               jp.Within.SourceLocation.FileName.ToString(),
               lineToSetBreakpoint);
         }

         return true;
      }
   }
}
