using System.IO;
using System.Xml.Serialization;

using Phx.Aop;
using Phx.Aop.Attributes;

[Extends("Person")]
public class Invariants
{
   [Add]
   public bool IsConst;

   [Advice(AdviceType.before, "set(* * Person.*(..))")]
   void CheckIsConst([SourceLocationAttribute.WithinSignature] string withinSignature,
                     [SourceLocationAttribute.FileName] string fileName,
                     [SourceLocationAttribute.Line] uint line,
		     [This] object This)
   {
      if (IsConst)
      {
         // ISSUE-BUG-t-marcea-2005/08/10-For some reason if I pass 'line'
         // to WriteLine() I get an invalid ImmOpnd.

         //System.Console.WriteLine(
         //   "'{0}' violated IsConst constraint inside {1}() [File: {2}, Line: {3}]",
         //   This, withinSignature, fileName, line);

         System.Console.WriteLine(
            "'{0}' violated IsConst constraint inside {1}() [File: {2}]",
            This,
            withinSignature,
            fileName);
      }
   }
}

[Extends("Person")]
public class DependenceAspect
{
   // Dependent.<Which> = this.<Which> <Operator> <Factor>;
   // Ex: Dependent.Age = this.Age * 1.1;

   [Add]
   public Person Dependent;

   [Add]
   public int Which;       // 0=none, 1=age, 2=salary

   [Add]
   public int Operator;    // 0=none, 1=add, 2=mult

   [Add]
   public double Factor;

   [Advice(AdviceType.after, "set(* * Person.*(..))")]
   public void UpdateDependent([This] Person This)
   {
      if (Which == 0)
         return;

      switch(Operator)
      {
         case 0:
            switch(Which)
            {
               case 1: Dependent.Age = This.Age; break;
               case 2: Dependent.Salary = This.Salary; break;
            }
            break;

         case 1: 
            switch(Which)
            {
               case 1: Dependent.Age = (int) (This.Age + Factor); break;
               case 2: Dependent.Salary = (int) (This.Salary + Factor); break;
            }
            break;

         case 2: 
            switch(Which)
            {
               case 1: Dependent.Age = (int) (This.Age * Factor); break;
               case 2: Dependent.Salary = (int) (This.Salary * Factor); break;
            }
            break;
      }
   }
}
