@REM Use 'set _echo=on' to echo all the commands
@if not "%_ECHO%"=="on" @if not "%_ECHO%"=="ON" @echo off

REM
REM Tries to mimic Visual Studio build for a C# project
REM
REM Remarks:
REM
REM     This script should not be executed directly.  Instead, it
REM	is called by a build script from a project directory.
REM
REM Environment required:
REM
REM	TARGET - Name (w/ extension) of the test assembly to build
REM	   (MorphPlugin.dll)
REM
REM	TARGET_DIR - Directory where target and references will be
REM	   output
REM
REM	SRCS - Space separated list of source files
REM
REM	PROJECT_DIR - Directory where the source files are located.
REM
REM	FLAGS - [Optional] Compile flags for csc.exe
REM
REM	REFS - [Optional] Space separated list of references. These
REM	   will be copied to the target directory. If Phx.Morph.dll or
REM	   MorphPlugin.dll are in the list, they will be built if
REM	   the DLL isn't found.
REM
REM Options:
REM
REM     Passes all command-line options to the called batch files.
REM
REM	-demo  - Pauses between significant steps and supresses debug msgs
REM
REM Returns:
REM
REM     0 if build succeeds, 1 otherwise
REM

setlocal

REM *****
REM ***** Verify that Phx environment has been setup
REM ***** 

if "%RDKROOT%"=="" echo Must setup Phx environment first && exit /B 1

REM *****
REM ***** Local vars
REM ***** 

set MORPH_DIR=%~dp0..

REM *****
REM ***** Sets BUILD, BUILD_EXT, PHX_REF, MSIL_REF, X86_REF, MSIL_RT_REF, X86_RT_REF
REM ***** 

call "%MORPH_DIR%\scripts\SetBuildAndPhxRefs.bat"

REM *****
REM ***** Verify required environment vars
REM ***** 

call "%MORPH_DIR%\scripts\requireenv.bat" TARGET %TARGET%
if not "%errorlevel%"=="0" exit /B 1

call "%MORPH_DIR%\scripts\requireenv.bat" TARGET_DIR "%TARGET_DIR%"
if not "%errorlevel%"=="0" exit /B 1

call "%MORPH_DIR%\scripts\requireenv.bat" PROJECT_DIR "%PROJECT_DIR%"
if not "%errorlevel%"=="0" exit /B 1

call "%MORPH_DIR%\scripts\requireenv.bat" SRCS %SRCS%
if not "%errorlevel%"=="0" exit /B 1

REM *****
REM ***** Parse command-line options
REM ***** 

call "%MORPH_DIR%\scripts\checkswitch.bat" echo %*
if "%errorlevel%"=="1" (
	set _ECHO=on
	echo on
)

call "%MORPH_DIR%\scripts\checkswitch.bat" demo %*
set DEMO=%errorlevel%

REM The INDENT variable looks empty but actually contains spaces
if "%DEMO%"=="1" set INDENT=      

set FLAGS=%FLAGS% /nologo
if "%BUILD%"=="debug" set FLAGS=%FLAGS% /debug /D:DEBUG /D:TRACE

REM *****
REM ***** Check for Phx.Morph.dll and MorphPlugin.dll in REFS
REM ***** 

set PHX_DLL=
for %%i in (%REFS%) do if "%%~nxi"=="phx%BUILD_EXT%.dll" set PHX_DLL=%%i

set PHX_AOP_DLL=
for %%i in (%REFS%) do if "%%~nxi"=="Phx.Aop.dll" set PHX_AOP_DLL=%%i

set PHX_MORPH_DLL=
for %%i in (%REFS%) do if "%%~nxi"=="Phx.Morph.dll" set PHX_MORPH_DLL=%%i

set PHX_PLUGIN_DLL=
for %%i in (%REFS%) do if "%%~nxi"=="MorphPlugin.dll" set PHX_PLUGIN_DLL=%%i

REM *****
REM ***** Build Phx.Aop.dll if necessary
REM ***** 

if '%PHX_AOP_DLL%'=='' goto :SkipBuildPhxAopDll

if exist %PHX_AOP_DLL% goto :SkipBuildPhxAopDll

REM Prevent recursion
if '%PHX_AOP_DLL%'=='%TARGET%' goto :SkipBuildPhxAopDll

call "%MORPH_DIR%\Phx.Aop\buildall.bat" %*

:SkipBuildPhxAopDll

REM *****
REM ***** Build Phx.Morph.dll if necessary
REM ***** 

if '%PHX_MORPH_DLL%'=='' goto :SkipBuildPhxMorphDll

if '%PHX_DLL%'=='' set REFS=%REFS% %PHX_REF% %MSIL_REF% %MSIL_RT_REF% %X86_REF% %X86_RT_REF%

if exist %PHX_MORPH_DLL% goto :SkipBuildPhxMorphDll

REM Prevent recursion
if '%PHX_MORPH_DLL%'=='%TARGET%' goto :SkipBuildPhxMorphDll

call "%MORPH_DIR%\Phx.Morph\buildall.bat" %*

:SkipBuildPhxMorphDll

REM *****
REM ***** Build MorphPlugin.dll if necessary
REM ***** 

if "%PHX_PLUGIN_DLL%"=="" goto :SkipBuildMorphPluginDll
if exist %PHX_PLUGIN_DLL% goto :SkipBuildMorphPluginDll

REM Prevent recursion
if "%PHX_PLUGIN_DLL%"=="%TARGET%" goto :SkipBuildMorphPluginDll

call "%MORPH_DIR%\MorphPlugin\buildall.bat" %*

:SkipBuildMorphPluginDll

REM *****
REM ***** Copy the REFS to target dir and initialize REFS_OPTION
REM ***** 

call "%MORPH_DIR%\scripts\CopyAndAssignRefs.bat" %* %REFS%
if not "%errorlevel%"=="0" goto :eof

REM *****
REM ***** Parse command-line options
REM ***** 

REM Parse command-line options
call "%MORPH_DIR%\scripts\checkswitch.bat" quiet %*
set QUIET=%errorlevel%

REM *****
REM ***** Build the assembly
REM ***** 

call :displaymsg %TARGET%

call :filenamewithoutext %TARGET%
if "%DEMO%"=="0" set XML_DOC_OPTION=/doc:"%RESULT%.xml" /nowarn:1591 /nowarn:1587 /nowarn:1574 /nowarn:1570 /nowarn:1572 /nowarn:1573

if not exist "%TARGET_DIR%" mkdir "%TARGET_DIR%"

if "%KEYFILE%"=="" set KEYFILE=%RDKROOT%rdk\setup\build\testkey.snk
if exist "%KEYFILE%" set DELAYSIGN=-delaysign+ -keyfile:"%KEYFILE%"

if "%DEMO%"=="1" (
	echo.
	echo csc.exe %FLAGS% %XML_DOC_OPTION% %TARGET_FLAGS% %REFS_OPTION% %SRCS% %DELAYSIGN%
)

if not exist "%PROJECT_DIR%" (
	echo Invalid project directory: Directory "%PROJECT_DIR%" does not exist.
	goto :EOF
)

cd "%PROJECT_DIR%"

csc.exe %FLAGS% %XML_DOC_OPTION% %TARGET_FLAGS% %REFS_OPTION% %SRCS% %DELAYSIGN%
if not "%errorlevel%"=="0" goto :EOF

if "%DEMO%"=="1" (
    if exist "%KEYFILE%" echo sn.exe -Raq %TARGET% "%KEYFILE%"
	if exist "%KEYFILE%" sn.exe -Raq %TARGET% "%KEYFILE%"
)

goto :EOF

:displaymsg
if "%QUIET%"=="0" (
	if "%DEMO%"=="1" (
		echo %INDENT%Building %~xn1
	) else (
		echo Building - %1
	)
)
goto :eof

:filenamewithoutext
set RESULT=%~dpn1
goto :eof
