@REM Use 'set _echo=on' to echo all the commands
@if not "%_ECHO%"=="on" @if not "%_ECHO%"=="ON" @echo off

REM
REM Builds the example and all its dependents and runs the morph
REM
REM Remarks:
REM
REM	Buildall executes a set of build scripts to build an
REM	example from the examples directory (e.g.
REM	rdk/samples/Morpher/examples/hello).  This script should
REM	not be executed directly.  Instead, it is called by a
REM	corresponding buildall.bat script in the examples directory.
REM	It just serves as a convenient place to put the sequence
REM	of steps required by all examples.
REM
REM Environment required:
REM
REM	Must be called from a buildall.bat script from the examples
REM	directory (e.g. 
REM	rdk/samples/Morpher/examples/hello/buildall.bat)
REM
REM Options:
REM
REM     Passes all command-line options to the called batch files.
REM
REM	-skip_build_morph - Doesn't build Phx.Morph.dll
REM
REM Returns:
REM
REM     0 if build succeeds, 1 otherwise
REM

setlocal

REM *****
REM ***** Verify that Phx environment has been setup
REM ***** 

if "%RDKROOT%"=="" echo Must setup Phx environment first && exit /B 1

REM *****
REM ***** Local vars
REM ***** 

set MORPH_DIR=%~dp0..

REM *****
REM ***** Parse command-line options
REM ***** 

call "%MORPH_DIR%\scripts\checkswitch.bat" demo %*
set DEMO=%errorlevel%

call "%MORPH_DIR%\scripts\checkswitch.bat" skip_build_morph %*
set SKIP_BUILD_MORPH=%errorlevel%

if "%DEMO%"=="1" (
	echo.
	echo ===== Build the Morpher =====
	echo.
	set /p ANS=      Press enter to continue...
)

REM I hate DOS!
call :clearerror

if "%SKIP_BUILD_MORPH%"=="0" (

	REM *****
	REM ***** Build Phx.Aop.dll
	REM ***** 

	if "%DEMO%"=="1" (
		call "%MORPH_DIR%\Phx.Aop\buildall.bat" %* -quiet
	) else (
		call "%MORPH_DIR%\Phx.Aop\buildall.bat" %*
	)
) 

if not "%errorlevel%"=="0" goto :EOF

if "%SKIP_BUILD_MORPH%"=="0" (

	REM *****
	REM ***** Build Phx.Morph.dll
	REM ***** 

	if "%DEMO%"=="1" (
		call "%MORPH_DIR%\Phx.Morph\buildall.bat" %* -quiet
	) else (
		call "%MORPH_DIR%\Phx.Morph\buildall.bat" %*
	)
) 

if not "%errorlevel%"=="0" goto :EOF

if "%SKIP_BUILD_MORPH%"=="0" (

	REM *****
	REM ***** Build MorphPlugin.dll
	REM ***** 

	if "%DEMO%"=="1" (
		call "%MORPH_DIR%\MorphPlugin\buildall.bat" %* -quiet
	) else (
		call "%MORPH_DIR%\MorphPlugin\buildall.bat" %*
		if not "%QUIET%"=="1" echo.
	)
)

if not "%errorlevel%"=="0" goto :EOF

if not "%SKIP_BUILD_MORPH%"=="0" (
	if not "%QUIET%"=="1" echo Skipping building Phx.Morph.dll and MorphPlugin.dll
)

REM *****
REM ***** Build the original (pre-morph) assembly
REM ***** 

if "%DEMO%"=="1" (
	echo.
	echo ===== Build the original assembly =====
	echo.
	set /p ANS=      Press enter to continue...
	echo.
)

call build-orig.bat %*
if not "%errorlevel%"=="0" goto :EOF

REM *****
REM ***** Build the extension assembly
REM ***** 

if "%DEMO%"=="1" (
	echo.
	echo ===== Build the aspect =====
	echo.
	set /p ANS=      Press enter to continue...
	echo.
)

call build-ext.bat %*
if not "%errorlevel%"=="0" goto :EOF

REM *****
REM ***** Morph the original assembly
REM ***** 

call morph.bat %*
if not "%errorlevel%"=="0" goto :EOF

REM *****
REM ***** Build the client that uses the morphed assembly
REM ***** 

if "%DEMO%"=="1" (
	echo.
	echo ===== Build the client =====
	echo.
	set /p ANS=      Press enter to continue...

	call build-client.bat %* -quiet
) else (
	call build-client.bat %*
)

if not "%errorlevel%"=="0" goto :EOF
exit /B 0

:clearerror
exit /B 0
