header
{
	using System.IO;
	using System.Text;
	using System.Collections.Generic;		// for the List of advices
	using System.Collections;				// For the hashtable used for a symbol table
}

options
{
	language 	= "CSharp";	
	namespace	= "Kunle.CSharpParser";
}

/*
[The "BSD licence"]
Copyright (c) 2002-2005 Kunle Odutola
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
3. The name of the author may not be used to endorse or promote products
derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


/// <summary>
/// An AST Printer (or un-parser) that prints the source code that a tree represents. 
/// </summary>
///
/// <remarks>
/// <para>
///	The default behaviour of this PrettyPrinter is to print out the AST and generate
/// source code that is as close to the original as is possible.
/// </para>
/// <para>
///	This behaviour can be overridden by supplying an <see cref="ICodeStyleScheme"/>
/// object that contains the settings for a custom formatting code style.
/// </para>
/// <para>
/// The TreeParser defined below is designed for the AST created by CSharpParser.
/// See the file "CSharpParser.g" for the details of that Parser.
/// </para>
///
/// <para>
/// History
/// </para>
///
/// <para>
/// 05-Jun-2003 kunle	  Created file.
/// </para>
///
/// </remarks>
*/



class CSharpPrettyPrinter extends TreeParser;

options
{
	importVocab						= CSharp;
	buildAST						= false;
	ASTLabelType					= "ASTNode";
	defaultErrorHandler				= false;
}

//=============================================================================
// Start of CODE
//=============================================================================

{
	//---------------------------------------------------------------------
	// PRIVATE DATA MEMBERS
	//---------------------------------------------------------------------
	
	// Variables to maintain line count consistent with original 
	// text (with the advice methods and attributes injected appropriately)
	private int 					currLine = 1;				// Added by Boriana
	private int 					nodeLinePretend = -1;
	private int						currLinePretend = -1;
	private bool					usedPretendCounter = false;
	private int						lineCountDiff = 0;
	private int						mode;						// Specifies what sort of pass you're currently doing.
																// Three passes of the code all in all.
	const int						POPULATE_TABLE_AND_FIND_ADVICE = 0;
	const int						CALCULATE_SPACING = 1;
	const int						CREATE_JOINPOINTS = 2;
	const int						OUTPUT_FILE = 3;

	
	// To maintain list of advices
	List<Wcs.Aop.WcsAdvice> advices = new List<Wcs.Aop.WcsAdvice>();

	// The symbol table I maintain to extract variable types for the joinpoints
	private Wicca.wcs.TablesList tablesList = null;					// hashtable containing symbol tables for all files you're parsing
	private Hashtable currTable = new Hashtable();
	private string className = "";			// stores the name of the current class we're tree-walking



	//---------------------------------------------------------------------
	// HELPER FUNCTIONS
	//---------------------------------------------------------------------

	public void PopulateSymbolTable_GetAdvice(TextWriter w, ASTNode astRoot, Wicca.wcs.TablesList list, List<Wcs.Aop.WcsAdvice> adv)
	{
		tablesList = list;
		advices = adv;
		mode = POPULATE_TABLE_AND_FIND_ADVICE;
		if (astRoot.Type == COMPILATION_UNIT)
		{
			compilationUnit(astRoot, w);
		}
	}

	public void PrintFile(TextWriter w, ASTNode astRoot, string scheme, Wicca.wcs.TablesList tables, List<Wcs.Aop.WcsAdvice> adv, bool weave)
	{
		if (astRoot.Type == COMPILATION_UNIT)
		{
			 advices = adv;
			 tablesList = tables;

			 // First pass to set the line numbers of the statements attributes apply to correctly
			 mode = CALCULATE_SPACING;
			compilationUnit(astRoot, w);

			if (weave)
			{
				// Reset variables and do real tree traversal with printing
				mode = CREATE_JOINPOINTS;
				currLine = 1;
				nodeLinePretend = -1;
				currLinePretend = -1;
				usedPretendCounter = false;
				lineCountDiff = 0;
	
				 // Second pass to create joinpoints
				compilationUnit(astRoot, w);
			}
			

			// Reset variables and do real tree traversal with printing
			mode = OUTPUT_FILE;
			// Reset variables
			currLine = 1;
			nodeLinePretend = -1;
			currLinePretend = -1;
			usedPretendCounter = false;
			lineCountDiff = 0;

			 // Third pass to actually print out the tree
			compilationUnit(astRoot, w);
		}
		else
		{
			throw new NotSupportedException("astRoot must be a COMPILATION_UNIT node.");
		}
	}

	/// <summary>
	/// Prints the text of the specified AST node and any hidden tokens that exist AFTER the node.
	/// </summary>
	/// <param name="w">The output destination.</param>
	/// <param name="node">The AST node to print.</param>
    private void Print(TextWriter w, ASTNode node)
    {    
    	Print(w, node, null, null);
    }

	/// <summary>
	/// Prints the text of the specified AST node and any hidden tokens that exist 
	/// AFTER the node.
	/// </summary>
	/// <param name="w">The output destination.</param>
	/// <param name="node">The AST node to print.</param>
    private void Print(TextWriter w, ASTNode node, string text, string textAfter)
    {
		// Only print out during the second pass... during first pass, we're only counting lines
    
    	// Fix newlines			// changed by Boriana
    	int num;
		int i;

    	if (nodeLinePretend < 0)
    	{
			num = #node.Line - currLine;
			for (i=0; i<num; i++)
			{
				if(mode == OUTPUT_FILE) 
					w.Write("" + "\n");
				currLine++;
			}
		}
		else			// Attributes and inserted text use this
		{
			num = nodeLinePretend - currLinePretend;

			for (i=0; i<num; i++)
			{
				if(mode == OUTPUT_FILE)
					w.Write("" + "\n");
				currLinePretend++;
				usedPretendCounter = true;
			}	
		}
					
		if(mode == OUTPUT_FILE) {
		
/*			// 1) ------------------ Write out the node's "before" text
			CustomHiddenStreamToken tok = (CustomHiddenStreamToken) node.getHiddenBefore();
			if ((tok == null) && (text != null))
				w.Write(text);
			else
			{
				string temp;
	
				while ( tok != null )
				{
					// Changed by Boriana so it doesn't print out the newline characters after the nodes
					temp = tok.getText();
					if(!temp.EndsWith("\n"))
						w.Write(temp);
					tok = (CustomHiddenStreamToken) tok.getHiddenBefore();
				}
			}
*/		
		
			// 2) ------------------ Write out the node's text		
			if (text == null)
			{
				w.Write(node.getText());
			}
			else
			{
				w.Write(text);
			}
//			w.Write("Line: " + node.Line);
    	
			// 3) ------------------ Write out the node's "after" text
			CustomHiddenStreamToken tok = (CustomHiddenStreamToken) node.getHiddenAfter();
			if ((tok == null) && (textAfter != null))
				w.Write(textAfter);
			else
			{
				string temp;
	
				while ( tok != null )
				{
					// Changed by Boriana so it doesn't print out the newline characters after the nodes
					temp = tok.getText();
					if(!temp.EndsWith("\n"))
						w.Write(temp);
					tok = (CustomHiddenStreamToken) tok.getHiddenAfter();
				}
			}
	    
	    }	// End if(mode == OUTPUT_FILE)
    }

	/// <summary>
	/// Prints the text of the specified AST node and any hidden tokens that exist 
	/// AFTER the node.
	/// </summary>
	/// <param name="w">The output destination.</param>
	/// <param name="node">The AST node to print.</param>
    private void PrintWithBeforeCheck(TextWriter w, ASTNode node)
    {
    	PrintWithBeforeCheck(w, node, null, null);
    }

	/// <summary>
	/// Prints the text of the specified AST node and any hidden tokens that exist 
	/// AFTER the node.
	/// </summary>
	/// <param name="w">The output destination.</param>
	/// <param name="node">The AST node to print.</param>
    private void PrintWithBeforeCheck(TextWriter w, ASTNode node, string text, string textAfter)
    {
    	// Changed by Boriana -- no longer does before checking.
    	Print(w, node, text, textAfter);
    }

	/// <summary>
	/// Prints any hidden tokens that may exist AFTER the specified AST node.
	/// </summary>
	/// <param name="w">The output destination.</param>
	/// <param name="node">The node of the AST to which the hidden tokens are linked.</param>
    private void PrintHiddenAfterTokens(TextWriter w, ASTNode node)
    {
    	CustomHiddenStreamToken tok = (CustomHiddenStreamToken) node.getHiddenAfter();
		while ( tok != null )
		{
	    	w.Write(tok.getText());
	     	tok = (CustomHiddenStreamToken) tok.getHiddenAfter();
	    }
    }

	/// <summary>
	/// Prints any hidden tokens that may exist BEFORE the specified AST node.
	/// </summary>
	/// <param name="w">The output destination.</param>
	/// <param name="node">The node of the AST to which the hidden tokens are linked.</param>
    private void PrintHiddenBeforeTokens(TextWriter w, ASTNode node)
    {
    	CustomHiddenStreamToken firstTok = null;
    	CustomHiddenStreamToken tok      = (CustomHiddenStreamToken) node.getHiddenBefore();

    	while ( tok != null )
    	{
    		firstTok = tok;
    		tok      = (CustomHiddenStreamToken) tok.getHiddenBefore();
    	}
    	
    	tok = firstTok;
		while ( tok != null )
		{
	    	w.Write(tok.getText());
	     	tok = (CustomHiddenStreamToken) tok.getHiddenAfter();
	    }
    }

	private bool NotExcluded(CodeMaskEnums codeMask, CodeMaskEnums construct)
	{
		return ((codeMask & construct) != 0 );
	}


	// __________________________________________________________
	// Returns in string form the type as specified by the "typeof" 
	//	node given as a parameter
	// __________________________________________________________
	private string typeAsString(ASTNode t)
	{
		string result = null;
		ASTNode Child = (ASTNode) t.getFirstChild();
		
		if (Child.Type == DOT)
		{
			while(Child.Type == DOT)
			{
				result += Child.getFirstChild().getText() + ".";
				Child = (ASTNode) Child.getFirstChild().getNextSibling();
			}
			result += Child.getText();
		}
		else
		{
			result = #t.getFirstChild().getText();
		}

		// Add on any stars or brackets, i.e. *'s or []'s
		if ((Child = (ASTNode) #t.GetFirstChildOfType(STARS).getFirstChild()) != null)
		{
			while(Child != null)
			{
				result += "*";
				Child = (ASTNode) Child.getNextSibling();
			}
		}
		if ((Child = (ASTNode) #t.GetFirstChildOfType(ARRAY_RANKS).getFirstChild()) != null)
		{
			while(Child != null)
			{
				result += "[]";
				Child = (ASTNode) Child.getNextSibling();
			}
		}
		
		return (result);
	}


//_______________________________________________________________
// Searches through the tree rooted at the node it receives and 
// returns the list of Joinpoints it generates within the tree
// which can be joinpoint of any type except EXECUTE (because
// it does not have enough knowledge to determine whether the 
// given node is the first/last in a method, or whether the given
// node is even a full statement or part of a statement).
//_______________________________________________________________
public void FindJPs(ASTNode root, ASTNode methodNode, ArrayList jps, Hashtable SymbolTable)
{

	// 1) --------------CHECK IF IT'S A CALL JOINPOINT --------------------
	if(#root.Type == OBJ_CREATE_EXPR)
	{
		Wcs.Aop.Joinpoints.WcsJoinpoint wjp = new Wcs.Aop.Joinpoints.WcsJoinpoint(#methodNode, #root, SymbolTable, tablesList, className);
		wjp.joinpointType = Phx.Aop.Joinpoints.JoinpointType.CALL;
		jps.Add(wjp);
	}
	else if(#root.Type == EXPR_STMT)
	{
		if(#root.getFirstChild().Type == INVOCATION_EXPR)
		{
			Wcs.Aop.Joinpoints.WcsJoinpoint wjp = new Wcs.Aop.Joinpoints.WcsJoinpoint(#methodNode, #root, SymbolTable, tablesList, className);
			wjp.joinpointType = Phx.Aop.Joinpoints.JoinpointType.CALL;
			jps.Add(wjp);
		}
		else if(#root.getFirstChild().Type == ASSIGN)
		{
			// (1) Assigning a value to a member access variable generates a "SET" type:
			// a.val = 10;
			if(#root.getFirstChild().getFirstChild().Type == MEMBER_ACCESS_EXPR)
			{
				Wcs.Aop.Joinpoints.WcsJoinpoint wjp = new Wcs.Aop.Joinpoints.WcsJoinpoint(#methodNode, (ASTNode)#root.getFirstChild().getFirstChild(), SymbolTable, tablesList, className);
				wjp.joinpointType = Phx.Aop.Joinpoints.JoinpointType.SET;
				jps.Add(wjp);
			}
			// (2) Assigning member access variable values to the variable generates "GET" type(s):
			// a.val = b.val (+ c.val);
			if(#root.getFirstChild().getFirstChild().getNextSibling().Type != IDENTIFIER 
				&& 
				#root.getFirstChild().getFirstChild().getNextSibling().Type != INT_LITERAL)
			{
				// If the node has children, traverse them to make sure the children in the expression are not member access variables
				ArrayList nodes = new ArrayList();
				((ASTNode)#root.getFirstChild().getFirstChild().getNextSibling()).FindTypesInTree(MEMBER_ACCESS_EXPR, nodes);
				// Traverse list of all nodes who matched the type to gather their joinpoints
				for(int i=0; i < nodes.Count; i++)
				{
					Wcs.Aop.Joinpoints.WcsJoinpoint wjp = new Wcs.Aop.Joinpoints.WcsJoinpoint(#methodNode, (ASTNode)nodes[i], SymbolTable, tablesList, className);
					wjp.joinpointType = Phx.Aop.Joinpoints.JoinpointType.GET;
					jps.Add(wjp);
				}
			}
		}	        
	} // END  if(Type==EXPR_STMT)
	
	// Apply method recursively to children of the tree
	ASTNode child = (ASTNode)#root.getFirstChild();
    if (child != null)
    {
        FindJPs(child, methodNode, jps, SymbolTable);

        ASTNode sibling = (ASTNode)child.getNextSibling();
        while (sibling != null)
        {
            FindJPs(sibling, methodNode, jps, SymbolTable);
            sibling = (ASTNode)sibling.getNextSibling();
        }
    }
} // END FindJPs


}

//=============================================================================
// Start of RULES
//=============================================================================

compilationUnit [TextWriter w]
	:	#(	COMPILATION_UNIT
			justPreprocessorDirectives[w, currTable]
			usingDirectives[w, currTable]
			globalAttributes[w, currTable]
			namespaceMemberDeclarations[w, currTable]
		)
	;
	
usingDirectives [TextWriter w, Hashtable SymbolTable]
	:	#(	USING_DIRECTIVES 
			(	preprocessorDirective[w, CodeMaskEnums.UsingDirectives, SymbolTable]
			|	usingDirective[w]
			)*
		)
	;
	
usingDirective [TextWriter w]
	:	#(	ndir:USING_NAMESPACE_DIRECTIVE 			{ PrintWithBeforeCheck(w, #ndir, null, " "); } 
			qualifiedIdentifier[w] 					{ if(mode == OUTPUT_FILE) w.Write(";"); }
		)
	|	#(	adir:USING_ALIAS_DIRECTIVE     			{ PrintWithBeforeCheck(w, #adir, null, " "); } 
			identifier[w] 							{ if(mode == OUTPUT_FILE) w.Write(" = "); } 
			qualifiedIdentifier[w] 					{ if(mode == OUTPUT_FILE) w.Write(";"); }
		)
	;

namespaceMemberDeclarations [TextWriter w, Hashtable SymbolTable]
	:	(	namespaceMemberDeclaration[w, SymbolTable]
		|	preprocessorDirective[w, CodeMaskEnums.NamespaceMemberDeclarations, SymbolTable]
		)*
	;
	
namespaceMemberDeclaration [TextWriter w, Hashtable SymbolTable]
	:	namespaceDeclaration[w, SymbolTable]
	|	typeDeclaration[w, SymbolTable]
	;
	
typeDeclaration [TextWriter w, Hashtable SymbolTable]
	:	(	classDeclaration[w, SymbolTable]
		|	structDeclaration[w, SymbolTable]
		|	interfaceDeclaration[w, SymbolTable]
		|	enumDeclaration[w, SymbolTable]
		|	delegateDeclaration[w, SymbolTable]
		)
	;

namespaceDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( 	ns:NAMESPACE { PrintWithBeforeCheck(w, #ns, null, " "); }
			qualifiedIdentifier[w] namespaceBody[w, SymbolTable] 
		)
	;
	
namespaceBody [TextWriter w, Hashtable SymbolTable]
	:	#(	ns:NAMESPACE_BODY { Print(w, #ns); } usingDirectives[w, SymbolTable] namespaceMemberDeclarations[w, SymbolTable] 
			cy:CLOSE_CURLY    { Print(w, #cy); }
		)
	;
	
modifiers [TextWriter w]
	:	#( MODIFIERS ( modifier[w] )* )
	;

modifier [TextWriter w]
	:	(	ab:ABSTRACT		{ PrintWithBeforeCheck(w, #ab, null, " "); }
		|	nw:NEW			{ PrintWithBeforeCheck(w, #nw, null, " "); }
		|	ov:OVERRIDE		{ PrintWithBeforeCheck(w, #ov, null, " "); }
		|	pu:PUBLIC		{ PrintWithBeforeCheck(w, #pu, null, " "); }
		|	pr:PROTECTED	{ PrintWithBeforeCheck(w, #pr, null, " "); }
		|	ir:INTERNAL		{ PrintWithBeforeCheck(w, #ir, null, " "); }
		|	pv:PRIVATE		{ PrintWithBeforeCheck(w, #pv, null, " "); }
		|	sl:SEALED		{ PrintWithBeforeCheck(w, #sl, null, " "); }
		|	st:STATIC		{ PrintWithBeforeCheck(w, #st, null, " "); }
		|	vi:VIRTUAL		{ PrintWithBeforeCheck(w, #vi, null, " "); }
		|	ex:EXTERN		{ PrintWithBeforeCheck(w, #ex, null, " "); }
		|	ro:READONLY		{ PrintWithBeforeCheck(w, #ro, null, " "); }
		|	un:UNSAFE		{ PrintWithBeforeCheck(w, #un, null, " "); }
		|	vo:VOLATILE		{ PrintWithBeforeCheck(w, #vo, null, " "); }
		)
	;


typeName [TextWriter w]
	:	predefinedType[w]
	|	qualifiedIdentifier[w]
	;
	
classTypeName [TextWriter w]
	:	qualifiedIdentifier[w]
	|	obj:OBJECT			{ Print(w, #obj); }
	|	str:STRING			{ Print(w, #str); }
	;
	
identifier [TextWriter w]
	:	id:IDENTIFIER		{ Print(w, #id); }
	;
	
qualifiedIdentifier [TextWriter w]
	:	(	identifier[w]
		|	#( dt:DOT identifier[w]	{ Print(w, #dt); } qualifiedIdentifier[w] )
		)
	;

	
//
// A.2.2 Types
//

type [TextWriter w]
	:	#(	TYPE 
			( 	qualifiedIdentifier[w]
			| 	predefinedType[w]
			|	v:VOID { Print(w, #v); }
			)
			pointerSpecifier[w]
			rankSpecifiers[w]
			{ if(mode == OUTPUT_FILE) w.Write(" "); }
		)
	;
	
pointerSpecifier [TextWriter w]
	:	#(	STARS 
			(	s:STAR { Print(w, #s); }
			)* 
		)
	;
	
classType [TextWriter w]
	:	qualifiedIdentifier[w]
	|	obj:OBJECT			{ Print(w, #obj); }
	|	str:STRING			{ Print(w, #str); }
	;
	
interfaceType [TextWriter w]
	:	qualifiedIdentifier[w]
	;

delegateType [TextWriter w]
	:	qualifiedIdentifier[w]		// typeName
	;


unmanagedType [TextWriter w]
	:	qualifiedIdentifier[w]		// typeName
	;
	
//
// A.2.3 Variables
//

variableReference [TextWriter w]
	:	expression[w]
	;

//
// A.2.4 Expressions
//
	
argumentList [TextWriter w]
	:	#(	ARG_LIST argument[w] ( { if(mode == OUTPUT_FILE) w.Write(", "); } argument[w] )*
		)
	;
	
argument [TextWriter w]
	:	expression[w]
	|	#( r:REF { Print(w, #r, null, " "); } variableReference[w] )
	|	#( o:OUT { Print(w, #o, null, " "); } variableReference[w] )
	;

constantExpression [TextWriter w]
	:	expression[w]
	;
	
booleanExpression [TextWriter w]
	:	expression[w]
	;
	
expressionList [TextWriter w]
	:	#(	EXPR_LIST expression[w] ( { if(mode == OUTPUT_FILE) w.Write(", "); } expression[w] )*
		)
	;

expression [TextWriter w]
	:	#( EXPR expr[w] )
	;
	
expr [TextWriter w]
		// assignmentExpression
		//
	:	#( asg:ASSIGN         expr[w] { Print(w, #asg); } expr[w] )
	|	#( plg:PLUS_ASSIGN    expr[w] { Print(w, #plg); } expr[w] )
	|	#( mig:MINUS_ASSIGN   expr[w] { Print(w, #mig); } expr[w] )
	|	#( stg:STAR_ASSIGN    expr[w] { Print(w, #stg); } expr[w] )
	|	#( dig:DIV_ASSIGN     expr[w] { Print(w, #dig); } expr[w] )
	|	#( mog:MOD_ASSIGN     expr[w] { Print(w, #mog); } expr[w] )
	|	#( ang:BIN_AND_ASSIGN expr[w] { Print(w, #ang); } expr[w] )
	|	#( org:BIN_OR_ASSIGN  expr[w] { Print(w, #org); } expr[w] )
	|	#( xog:BIN_XOR_ASSIGN expr[w] { Print(w, #xog); } expr[w] )
	|	#( slg:SHIFTL_ASSIGN  expr[w] { Print(w, #slg); } expr[w] )
	|	#( srg:SHIFTR_ASSIGN  expr[w] { Print(w, #srg); } expr[w] )
		// conditionalExpression
		//
	|	#( q:QUESTION       expr[w] { Print(w, #q, null, " ? "); } expr[w] { if(mode == OUTPUT_FILE) w.Write(" : "); } expr[w] )
	
		// conditional-XXX-Expressions
		//
	|	#( lor:LOG_OR         expr[w] { Print(w, #lor, null, " "); } expr[w] )
	|	#( lan:LOG_AND        expr[w] { Print(w, #lan, null, " "); } expr[w] )
		
		// bitwise-XXX-Expressions
		//
	|	#( bor:BIN_OR         expr[w] { Print(w, #bor, null, " "); } expr[w] )
	|	#( box:BIN_XOR        expr[w] { Print(w, #box, null, " "); } expr[w] )
	|	#( boa:BIN_AND        expr[w] { Print(w, #boa, null, " "); } expr[w] )

		// equalityExpression
		//
	|	#( eql:EQUAL          expr[w] { Print(w, #eql, null, " "); } expr[w] )
	|	#( neq:NOT_EQUAL      expr[w] { Print(w, #neq, null, " "); } expr[w] )

		// relationalExpression
		//
	|	#( r1:LTHAN          expr[w] { Print(w, #r1, null, " "); } expr[w] )
	|	#( r2:GTHAN          expr[w] { Print(w, #r2, null, " "); } expr[w] )
	|	#( r3:LTE            expr[w] { Print(w, #r3, null, " "); } expr[w] )
	|	#( r4:GTE            expr[w] { Print(w, #r4, null, " "); } expr[w] )
	|	#( r5:IS             expr[w] { if(mode == OUTPUT_FILE) w.Write(" "); Print(w, #r5, null, " "); } type[w] )
	|	#( r6:AS             expr[w] { if(mode == OUTPUT_FILE) w.Write(" "); Print(w, #r6, null, " "); } type[w] )
	
		// shiftExpression
		//
	|	#( sl:SHIFTL         expr[w] { Print(w, #sl, null, " "); } expr[w] )
	|	#( sr:SHIFTR         expr[w] { Print(w, #sr, null, " "); } expr[w] )

		// additiveExpression
		//
	|	#( p:PLUS           expr[w] { Print(w, #p); } expr[w] )
	|	#( m:MINUS          expr[w] { Print(w, #m); } expr[w] )
		// multiplicativeExpression
		//
	|	#( m1:STAR           expr[w] { Print(w, #m1); } expr[w] )
	|	#( m2:DIV            expr[w] { Print(w, #m2); } expr[w] )
	|	#( m3:MOD            expr[w] { Print(w, #m3); } expr[w] )
	|	unaryExpression[w]
	;	

unaryExpression [TextWriter w]
	:	#( c:CAST_EXPR 			{ Print(w, #c); } type[w] { if(mode == OUTPUT_FILE) w.Write(") "); } expr[w] )
	|	#( i:INC              	{ Print(w, #i); } expr[w] )
	|	#( d:DEC              	{ Print(w, #d); } expr[w] )
	|	#( up:UNARY_PLUS       	{ Print(w, #up); } expr[w] )
	|	#( um:UNARY_MINUS      	{ Print(w, #um); } expr[w] )
	|	#( l:LOG_NOT            { Print(w, #l); } expr[w] )
	|	#( b:BIN_NOT			{ Print(w, #b); } expr[w] )
	|	#( p:PTR_INDIRECTION_EXPR { Print(w, #p); } expr[w] )
	|	#( a:ADDRESS_OF_EXPR      { Print(w, #a); } expr[w] )
	|	primaryExpression[w]
	;

primaryExpression [TextWriter w]
	:	#(	// invocationExpression ::= primaryExpression OPEN_PAREN ( argumentList )? CLOSE_PAREN 
			i:INVOCATION_EXPR			primaryExpression[w] { Print(w, #i); } 
										(argumentList[w])?   { if(mode == OUTPUT_FILE) w.Write(")"); } 
		)
	|	#(	// elementAccess		::= primaryNoArrayCreationExpression OPEN_BRACK expressionList CLOSE_BRACK
			e:ELEMENT_ACCESS_EXPR		primaryExpression[w] { Print(w, #e); } 
										expressionList[w]    { if(mode == OUTPUT_FILE) w.Write("]"); }
		)
	|	#(	// pointerElementAccess ::= primaryNoArrayCreationExpression OPEN_BRACK expression     CLOSE_BRACK
			p:PTR_ELEMENT_ACCESS_EXPR	primaryExpression[w] { Print(w, #p); } 
										expressionList[w]    { if(mode == OUTPUT_FILE) w.Write("]"); }
		)
	|	#(	// memberAccess		    ::= primaryExpression DOT identifier
			m:MEMBER_ACCESS_EXPR      									
			(	type[w]
			|	primaryExpression[w]
			)
			{ Print(w, #m); } identifier[w]
		)
	|	// pointerMemberAccess
		#( d:DEREF                   primaryExpression[w] { Print(w, #d); } identifier[w] )
	|	// postIncrementExpression
		#( pi:POST_INC_EXPR          primaryExpression[w] { Print(w, #pi); } )
	|	// postDecrementExpression
		#( pd:POST_DEC_EXPR          primaryExpression[w] { Print(w, #pd); } )
	|	basicPrimaryExpression[w]
	;

basicPrimaryExpression [TextWriter w]
	:	literal[w]
	|	identifier[w]										// simpleName
	|	// parenthesizedExpression
		//
		#( p:PAREN_EXPR { Print(w, #p); } expr[w] { if(mode == OUTPUT_FILE) w.Write(")"); } )
	|	tx:THIS 		{ Print(w, #tx); }
	|	#(	bx:BASE		{ Print(w, #bx); if(mode == OUTPUT_FILE) w.Write("."); } 
			(	identifier[w]
			|	expressionList[w]
			)
		)
	|	newExpression[w]
	|	// typeofExpression
		//
		#( t:TYPEOF 	{ Print(w, #t); if(mode == OUTPUT_FILE) w.Write("("); } type[w]          { if(mode == OUTPUT_FILE) w.Write(")"); } )
	|	#( s:SIZEOF    	{ Print(w, #s); if(mode == OUTPUT_FILE) w.Write("("); } unmanagedType[w] { if(mode == OUTPUT_FILE) w.Write(")"); } )
	|	#( c:CHECKED   	{ Print(w, #c); if(mode == OUTPUT_FILE) w.Write("("); } expression[w]    { if(mode == OUTPUT_FILE) w.Write(")"); } )
	|	#( u:UNCHECKED 	{ Print(w, #u); if(mode == OUTPUT_FILE) w.Write("("); } expression[w]    { if(mode == OUTPUT_FILE) w.Write(")"); } )
	;

newExpression [TextWriter w]
		// objectCreationExpression	  
		//
	:	#( ob:OBJ_CREATE_EXPR { Print(w, #ob, null, " "); } type[w] { if(mode == OUTPUT_FILE) w.Write("("); } (argumentList[w])? { if(mode == OUTPUT_FILE) w.Write(")"); } )
		// delegateCreationExpression	  
		//
	|	#( dl:DLG_CREATE_EXPR { Print(w, #dl, null, " "); } type[w] { if(mode == OUTPUT_FILE) w.Write("("); } argumentList[w] { if(mode == OUTPUT_FILE) w.Write(")"); } )
		// arrayCreationExpression
		//
	|	#( 	ar:ARRAY_CREATE_EXPR { Print(w, #ar, null, " "); }
			type[w]										// nonArrayType ( rankSpecifiers )?
			(	{ if(mode == OUTPUT_FILE) w.Write("[] "); } arrayInitializer[w]
			|	{ if(mode == OUTPUT_FILE) w.Write("[");   } expressionList[w] { if(mode == OUTPUT_FILE) w.Write("]"); } 
				rankSpecifiers[w] ( arrayInitializer[w] )? 
			)
		)
	;

literal [TextWriter w]
	:	tr:TRUE				{ Print(w, #tr); }
	|	fa:FALSE			{ Print(w, #fa); }
	|	il:INT_LITERAL		{ Print(w, #il); }
	|	ui:UINT_LITERAL		{ Print(w, #ui); }
	|	ll:LONG_LITERAL		{ Print(w, #ll); }
	|	ul:ULONG_LITERAL	{ Print(w, #ul); }
	|	dc:DECIMAL_LITERAL	{ Print(w, #dc); }
	|	fl:FLOAT_LITERAL	{ Print(w, #fl); }
	|	dl:DOUBLE_LITERAL	{ Print(w, #dl); }
	|	cl:CHAR_LITERAL		{ Print(w, #cl); }
	|	sl:STRING_LITERAL	{ Print(w, #sl); }
	|	nl:NULL				{ Print(w, #nl); }
	;

predefinedType [TextWriter w]
	:	bl:BOOL				{ Print(w, #bl); } 
	|	by:BYTE				{ Print(w, #by); }
	|	ch:CHAR				{ Print(w, #ch); }
	|	dc:DECIMAL			{ Print(w, #dc); }
	|	db:DOUBLE			{ Print(w, #db); }
	|	fl:FLOAT			{ Print(w, #fl); }
	|	it:INT				{ Print(w, #it); }
	|	lg:LONG				{ Print(w, #lg); }
	|	ob:OBJECT			{ Print(w, #ob); }
	|	sb:SBYTE			{ Print(w, #sb); }
	|	sh:SHORT			{ Print(w, #sh); }
	|	st:STRING			{ Print(w, #st); }
	|	ui:UINT				{ Print(w, #ui); }
	|	ul:ULONG			{ Print(w, #ul); }
	|	us:USHORT			{ Print(w, #us); }
	;


//
// A.2.5 Statements
//

// First check whether the statement has an attribute attached to it, and if so, output the text associated with
// the attribute before writing out the statement.
statement [TextWriter w, ASTNode self, Hashtable SymbolTable, ASTNode methodNode]
	{
		// __________________________________________________________________
		//			CREATE JOINPOINT(S) FOR THIS STATEMENT
		//					IF IT'S THE APPROPRIATE MODE
		// __________________________________________________________________

		if ((mode == CREATE_JOINPOINTS) && (methodNode != null))
		{
			ASTNode blk = methodNode.GetFirstChildOfType(BLOCK);
			ArrayList jps = new ArrayList();
			FindJPs(self, methodNode, jps, SymbolTable);

			// If nothing else matched, make the joinpoint have the generic type STATEMENT
			if (jps.Count == 0)
			{
				Wcs.Aop.Joinpoints.WcsJoinpoint wjp = new Wcs.Aop.Joinpoints.WcsJoinpoint(#methodNode, #self, SymbolTable, tablesList, className);
				wjp.joinpointType = Phx.Aop.Joinpoints.JoinpointType.STATEMENT;
				jps.Add(wjp);
			}

			// Try to match your joinpoint to the advice you have
			for(int i=0; i < jps.Count; i++)
			{
				foreach (Wcs.Aop.WcsAdvice advice in advices)
				{
				   if (!advice.Pointcut.matches((Wcs.Aop.Joinpoints.WcsJoinpoint)jps[i])
						|| advice.AlreadyApplied((Wcs.Aop.Joinpoints.WcsJoinpoint)jps[i]))       // Prevent redundant weaving
				   {
					  continue;
				   }
				   else
				   {
//	Console.WriteLine("I matched");	
						// An advice matches the current joinpoint!
						// Weave in the advice
						int index = advice.getAttributeIndexNeeded();
//	Console.WriteLine("index = " + index);
						string attrType = ((Wcs.Aop.Joinpoints.WcsJoinpoint)jps[i]).getAttributeType(index);
//	Console.WriteLine("attrType = " + attrType);
						if(!attrType.StartsWith("Advice"))		// Don't weave when it's advice attributes
						{
							List <ASTNode> args = null;
							args = ((Wcs.Aop.Joinpoints.WcsJoinpoint)jps[i]).getAttributeArgs(index);
//	if(args != null) {
//		IEnumerator ArgEnum = args.GetEnumerator();
//	    while(ArgEnum.MoveNext())
//	    {
//			Console.WriteLine("arg = " + ((ASTNode)ArgEnum.Current).getText());
//		}
//	}
//	else {
//		Console.WriteLine("no args");
//	}
							#blk.AddAsChildBeforeNode(advice.Method_Decl_To_Call(attrType, args), #self);
							advice.AddAppliedJoinpoint(((Wcs.Aop.Joinpoints.WcsJoinpoint)jps[i]));
						}
				   }
				}

//				Console.WriteLine((Wcs.Aop.Joinpoints.WcsJoinpoint)jps[i]);
			}

		}


		// __________________________________________________________________
		//			CREATE ATTRIBUTE CHILDREN IF IT'S THE APPROPRIATE 
		//					MODE AND THERE ARE ATTRIBUTES
		// __________________________________________________________________			
		if ( (#self != null) && (#self.HasChildOfType(ATTRIBUTE_SECTION)))
		{
			if(mode == CALCULATE_SPACING)
			{
				// Make sure you find a node of the current statement that has a Line attribute attached to it
				ASTNode temp = (ASTNode) #self;
				while (#temp.Line == -1)
				{
					#temp = (ASTNode) #temp.getFirstChild();
				}			

				int newLineNum = #temp.Line + lineCountDiff;
				ASTNode curr = #self.GetFirstChildOfType(ATTRIBUTE_SECTION);
				#curr.GetFirstChildOfType(ATTRIBUTE).GetFirstChildOfType(POSITIONAL_ARGLIST).GetFirstChildOfType(POSITIONAL_ARG).GetFirstChildOfType(ATTRIB_ARGUMENT_EXPR).GetFirstChildOfType(EXPR).GetFirstChildOfType(INT_LITERAL).setText("" + newLineNum);

				while( (#curr.getNextSibling() != null) && (#curr.getNextSibling().Type == ATTRIBUTE_SECTION) )
				{
					#curr = (ASTNode) #curr.getNextSibling();
					#curr.GetFirstChildOfType(ATTRIBUTE).GetFirstChildOfType(POSITIONAL_ARGLIST).GetFirstChildOfType(POSITIONAL_ARG).GetFirstChildOfType(ATTRIB_ARGUMENT_EXPR).GetFirstChildOfType(EXPR).GetFirstChildOfType(INT_LITERAL).setText("" + newLineNum);
				}
				
			}
		}
	}	
	:	#(	l:LABEL_STMT identifier[w] { Print(w, #l, null, " "); } statement[w, null, SymbolTable, methodNode]
		)
	|	localVariableDeclaration[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(";"); }
	|	localConstantDeclaration[w] { if(mode == OUTPUT_FILE) w.Write(";"); }
	|	embeddedStatement[w, SymbolTable]
	|	preprocessorDirective[w, CodeMaskEnums.Statements, SymbolTable]
	;
	
	
embeddedStatement [TextWriter w, Hashtable SymbolTable]
	:	block[w, SymbolTable]
	|	emp:EMPTY_STMT	{ Print(w, #emp); }
	|	#( sxp:EXPR_STMT statementExpression[w] { Print(w, #sxp); } )
	|	#(	ifStmt:IF { Print(w, #ifStmt); }
			{ if(mode == OUTPUT_FILE) w.Write("("); } expression[w] { if(mode == OUTPUT_FILE) w.Write(") "); } embeddedStatement[w, SymbolTable]
			(	#(	elseStmt:ELSE { Print(w, #elseStmt, null, " "); }
					embeddedStatement[w, SymbolTable]
				)
			)? 
		)
	|	#(	switchStmt:SWITCH { Print(w, #switchStmt); }
			{ if(mode == OUTPUT_FILE) w.Write("("); } e:expression[w] { Print(w, #e, null, ") "); }
			#(	oc1:OPEN_CURLY { Print(w, #oc1); }
				( switchSection[w, SymbolTable] )* 
				cc1:CLOSE_CURLY { Print(w, #cc1); }
			) 
		)
	|	#( 	forStmt:FOR { Print(w, #forStmt); if(mode == OUTPUT_FILE) w.Write("("); }
			#( FOR_INIT ( ( localVariableDeclaration[w, SymbolTable] | ( statementExpression[w] )+ ) )? )
			{ if(mode == OUTPUT_FILE) w.Write("; "); }
			#( FOR_COND ( booleanExpression[w] )? )
			{ if(mode == OUTPUT_FILE) w.Write("; "); }
			#( FOR_ITER ( ( statementExpression[w] )+  )? )
			{ if(mode == OUTPUT_FILE) w.Write(") "); }
			embeddedStatement[w, SymbolTable]
		)
	|	#(	whileStmt:WHILE { Print(w, #whileStmt); if(mode == OUTPUT_FILE) w.Write("("); }
			booleanExpression[w] { if(mode == OUTPUT_FILE) w.Write(") "); } embeddedStatement[w, SymbolTable]
		)
	|	#(	doStmt:DO { Print(w, #doStmt, null, " "); }
			embeddedStatement[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(" while ("); } booleanExpression[w] { if(mode == OUTPUT_FILE) w.Write(");"); }
		)
	|	#(	foreachStmt:FOREACH { Print(w, #foreachStmt); if(mode == OUTPUT_FILE) w.Write("("); }
			localVariableDeclaration[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(" in "); } expression[w] 
			{ if(mode == OUTPUT_FILE) w.Write(") "); }
			embeddedStatement[w, SymbolTable]
		)
	|	brk:BREAK		{ Print(w, #brk); if(mode == OUTPUT_FILE) w.Write(";"); }
	|	ctn:CONTINUE	{ Print(w, #ctn); if(mode == OUTPUT_FILE) w.Write(";"); }
	|	#(	gto:GOTO 	{ Print(w, #gto, null, " "); }
			(	identifier[w] 
			|	cse:CASE 	{ Print(w, #cse, null, " "); } constantExpression[w]
			|	dfl:DEFAULT	{ Print(w, #dfl); }
			)
			{ w. Write(";"); }
		)
	|	#( rtn:RETURN { Print(w, #rtn, null, " "); } ( expression[w] )? { if(mode == OUTPUT_FILE) w.Write(";"); } )
	|	#( trw:THROW  { Print(w, #trw, null, " "); } ( expression[w] )? { if(mode == OUTPUT_FILE) w.Write(";"); } )
	|	tryStatement[w, SymbolTable]
	|	#(	checkedStmt:CHECKED 	{ Print(w, #checkedStmt, null, " "); }   block[w, SymbolTable]
		)
	|	#(	uncheckedStmt:UNCHECKED	{ Print(w, #uncheckedStmt, null, " "); } block[w, SymbolTable]
		)
	|	#(	lockStmt:LOCK { Print(w, #lockStmt); }
			{ if(mode == OUTPUT_FILE) w.Write("("); } expression[w] { if(mode == OUTPUT_FILE) w.Write(") "); } embeddedStatement[w, SymbolTable]
		)
	|	#(	usingStmt:USING { Print(w, #usingStmt); }
			{ if(mode == OUTPUT_FILE) w.Write("("); } resourceAcquisition[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(") "); } embeddedStatement[w, SymbolTable]
		)
	|	#(	unsafeStmt:UNSAFE { Print(w, #unsafeStmt, null, " "); }
			block[w, SymbolTable]
		)
	|	// fixedStatement
		#(	fixedStmt:FIXED { Print(w, #fixedStmt); }
			{ if(mode == OUTPUT_FILE) w.Write("("); } type[w] 
			fixedPointerDeclarator[w] ( { if(mode == OUTPUT_FILE) w.Write(", "); } fixedPointerDeclarator[w] )*
			{ if(mode == OUTPUT_FILE) w.Write(") "); }
			embeddedStatement[w, SymbolTable]
		)
	;
	
body[TextWriter w, Hashtable SymbolTable]
	:	block[w, SymbolTable]
	|	empty:EMPTY_STMT	{ Print(w, #empty); }
	;



methodBody [TextWriter w, Hashtable SymbolTable, ASTNode methodDecl]
	: #( blk:BLOCK { Print(w, #blk, null, " "); }
		{		
			// Create a new *local* SymbolTable to pass to the statements inside the method block
			// This way variables declared locally with override the global vars
			Hashtable localTable = new Hashtable();
		
			if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
			{
				// All all elements from the global symbol table to your local one
				IDictionaryEnumerator Enumerator = SymbolTable.GetEnumerator();
				while (Enumerator.MoveNext())
				{
					localTable.Add(Enumerator.Key, Enumerator.Value);
				}
			}
			
		}
		( stmt:statement[w, #stmt, localTable, #methodDecl]				// Create joinpoints in this statement rule
		)* 
		cly:CLOSE_CURLY { Print(w, #cly); }
	 )
	|	empty:EMPTY_STMT	{ Print(w, #empty); }
	;



block [TextWriter w, Hashtable SymbolTable]
	:	#(	blk:BLOCK { Print(w, #blk, null, " "); }
			{
				// Create a new *local* SymbolTable to pass to the statements inside the method block
				// This way variables declared locally with override the global vars
				Hashtable localTable = new Hashtable();

				if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
				{
					// Add all elements from the global symbol table to your local one
					IDictionaryEnumerator Enumerator = SymbolTable.GetEnumerator();
					while (Enumerator.MoveNext())
					{
						localTable.Add(Enumerator.Key, Enumerator.Value);
					}
				}
			}
			( 
				stmt:statement[w, #stmt, localTable, null]
			)*
			cly:CLOSE_CURLY { Print(w, #cly); }
		)
	;
	
statementList[TextWriter w, Hashtable SymbolTable, ASTNode methodDecl]
	:	#( STMT_LIST ( stmt:statement[w, #stmt, SymbolTable, methodDecl] )+ )
	;
	
localVariableDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	LOCVAR_DECLS t:type[w] localVariableDeclarator[w, SymbolTable, typeAsString(#t)]
			(	
				{ if(mode == OUTPUT_FILE) w.Write(", "); }
				localVariableDeclarator[w, SymbolTable, #t.getFirstChild().getText()]
			)*
		)
	;
	
localVariableDeclarator [TextWriter w, Hashtable SymbolTable, String type]
	:	#(	VAR_DECLARATOR id:identifier [w]
			{
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
				{
					// Add the variable you've detected to the symbol table: (var_name, var_type)
					// If it already exists in the symboltable, remove it and replace it with the local value!
					if(SymbolTable.ContainsKey(#id.getText()))
						SymbolTable.Remove(#id.getText());
					SymbolTable.Add(#id.getText(), type);
				}
			}
			(	{ if(mode == OUTPUT_FILE) w.Write(" = "); }
				#(	LOCVAR_INIT
					(	expression[w]
					|	arrayInitializer[w]
					)
				) 
			)? 
		)
	;
	
localConstantDeclaration [TextWriter w]
	:	#(	c:LOCAL_CONST { Print(w, #c, null, " "); } type[w] constantDeclarator[w]
			(	{ if(mode == OUTPUT_FILE) w.Write(", "); }
				constantDeclarator[w] 
			)*
		)
	;
	
constantDeclarator [TextWriter w]
	:	#(	c:CONST_DECLARATOR identifier[w] { Print(w, #c, null, " "); } constantExpression[w]
		)
	;
	
statementExpression [TextWriter w]
	:	expr[w]
	;
	
switchSection[TextWriter w, Hashtable SymbolTable]
	:	#( SWITCH_SECTION switchLabels[w] statementList[w, SymbolTable, null] )
	;
	
switchLabels[TextWriter w]
	:	#(	SWITCH_LABELS 
			(	(	#( cse:CASE { Print(w, #cse, null, " "); } expression[w] )
				|	dfl:DEFAULT	{ Print(w, #dfl, null, " "); }
				)
				{ if(mode == OUTPUT_FILE) w.Write(": "); }
			)+ 
		)
	;
	
tryStatement [TextWriter w, Hashtable SymbolTable]
	:	#(	t:TRY { Print(w, #t, null, " "); }
			block[w, SymbolTable]
			(	finallyClause[w, SymbolTable]
			|	catchClauses[w, SymbolTable] ( finallyClause[w, SymbolTable] )?
			)
		)
	;
	
catchClauses [TextWriter w, Hashtable SymbolTable]
	:	(	
			{
				StringBuilder sb = new StringBuilder();
				StringWriter  sw = new StringWriter(sb);
			}
			#(	c:CATCH { Print(w, #c, null, " "); }
				block[sw, SymbolTable]
				(	{ if(mode == OUTPUT_FILE) w.Write("("); }
					(	type[w]
					|	localVariableDeclaration[w, SymbolTable]
					)
					{ if(mode == OUTPUT_FILE) w.Write(") "); }
				)? 
				{
					if(mode == OUTPUT_FILE) w.Write(sb.ToString());
					sw.Close();
				}
			)
		)+
	;
	
finallyClause [TextWriter w, Hashtable SymbolTable]
	:	#(	f:FINALLY { Print(w, #f, null, " "); }
			block[w, SymbolTable]
		)
	;
	
resourceAcquisition[TextWriter w, Hashtable SymbolTable]
	:	localVariableDeclaration[w, SymbolTable]
	|	expression[w]
	;
	
//	
// A.2.6 Classes
//

classDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( 	cl:CLASS attributes[w, SymbolTable] modifiers[w] { PrintWithBeforeCheck(w, #cl, null, " "); } id:identifier[w]
		{ 	className = #id.getText();							// store the name of the current class we're parsing

			if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
			{
				SymbolTable = new Hashtable();					// New class so start a brand new symbol table
				SymbolTable.Add("this", #id.getText());			// Insert the value of the "this" pointer as refering to the current class in the local table
			}
		}
			#(	CLASS_BASE
				{
					bool needToPrintColon = true;
					bool needToPrintComma = false;
				}
				(	{
						if ( needToPrintColon )
						{
							if(mode == OUTPUT_FILE) w.Write(" : ");
							needToPrintColon = false;
						}
						if ( needToPrintComma )
							if(mode == OUTPUT_FILE) w.Write(", ");
					}
					type[w]
					{ needToPrintComma = true; }
				)* 
			) 
			#(	tb:TYPE_BODY { Print(w, #tb); } classMemberDeclarations[w, SymbolTable] cc:CLOSE_CURLY 
				{
					if (mode == POPULATE_TABLE_AND_FIND_ADVICE)
					{
						tablesList.addSymbolTable(#id.getText(), SymbolTable);
					}
					Print(w, #cc); 
				} 
			)
		)
	;
	
classMemberDeclarations [TextWriter w, Hashtable SymbolTable]
	:	#(	MEMBER_LIST
			(	classMemberDeclaration [w, SymbolTable]
			|	preprocessorDirective[w, CodeMaskEnums.ClassMemberDeclarations, SymbolTable]
			)*
		)
	;
	
classMemberDeclaration [TextWriter w, Hashtable SymbolTable]
	:	(	destructorDeclaration[w, SymbolTable]
		|	typeMemberDeclaration[w, SymbolTable] 
		)
	;
	
typeMemberDeclaration [TextWriter w, Hashtable SymbolTable]
	:
		(	constantDeclaration[w, SymbolTable]
		|	eventDeclaration[w, SymbolTable]
		|	constructorDeclaration[w, SymbolTable]
		|	staticConstructorDeclaration[w, SymbolTable]
		|	propertyDeclaration[w, SymbolTable]
		|	methodDeclaration[w, SymbolTable] 
		|	indexerDeclaration[w, SymbolTable]
		|	fieldDeclaration[w, SymbolTable]
		|	operatorDeclaration[w, SymbolTable]
		|	typeDeclaration[w, SymbolTable]
		)
	;
	
constantDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	c:CONST attributes[w, SymbolTable] modifiers[w] { Print(w, #c, null, " "); } type[w]
			( constantDeclarator[w] )+
			{ if(mode == OUTPUT_FILE) w.Write(";"); }
		)
	;
	
fieldDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	FIELD_DECL attributes[w, SymbolTable] modifiers[w] t:type[w]
			( variableDeclarator[w, SymbolTable, typeAsString(#t) ] )
			( {if(mode == OUTPUT_FILE) w.Write(", ");} variableDeclarator[w, SymbolTable, #t.getFirstChild().getText()] )*
			{ if(mode == OUTPUT_FILE) w.Write(";"); }
		)
	;
	
variableDeclarator [TextWriter w, Hashtable SymbolTable, String type]
	:	#(	VAR_DECLARATOR id:identifier[w]
			{
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
				{
					SymbolTable.Add(#id.getText(), type);
				}
			}
	
			({
				if(mode == OUTPUT_FILE) w.Write(" = "); 
			 }  
			 variableInitializer[w] 
			)? 
		)
	;
	
variableInitializer [TextWriter w]
	:	#(	VAR_INIT
			(	expression[w]
			|	arrayInitializer[w]
			|	stackallocInitializer[w]
			)
		)
	;
		
methodDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	md:METHOD_DECL
			{ nodeLinePretend = #md.Line; currLinePretend = currLine; } 
			attribs:attributes[w, SymbolTable] 
			{ nodeLinePretend = -1; currLinePretend = -1;
				if(usedPretendCounter) {
					currLine = #md.Line;
					usedPretendCounter = false;}
			}
			modifiers[w] type[w] qi:qualifiedIdentifier[w]
			{
				StringBuilder sb = new StringBuilder();
				StringWriter  sw = new StringWriter(sb);
			}
			{
				// Store this method declaration in the symboltable of this class
				// If it already exists in the symboltable, remove it and replace it with the local value!
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE)
				{
					string id = #qi.getText();			// This is the method's key into the table
					// Create a "Within" WcsJoinpointSignature for this method declaration
					Wcs.Aop.Joinpoints.WcsJoinpointSignature within = new Wcs.Aop.Joinpoints.WcsJoinpointSignature(true, SymbolTable, className);
                    within.node = #md;
                    within.methodAttributes = #md.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE_SECTIONS);

					if(SymbolTable.ContainsKey(id))
						SymbolTable.Remove(id);
					SymbolTable.Add(id, within);
				}

				// Create a new *local* SymbolTable to pass to those class properties for which the local
				// tables should not persist past the lifetime of the property
				Hashtable localTable = new Hashtable();
			
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
				{
					// All all elements from the global symbol table to your local one
					IDictionaryEnumerator Enumerator = SymbolTable.GetEnumerator();
					while (Enumerator.MoveNext())
					{
						localTable.Add(Enumerator.Key, Enumerator.Value);
					}
				}
			}

			{ if(mode == OUTPUT_FILE) w.Write("("); } ( formalParameterList[w, localTable] )?
			{ if(mode == OUTPUT_FILE) w.Write(")"); }
						
			methodBody[sw, localTable, #md]		// Give a reference to the METHOD_DECL node, so that the methodBody can use it when creating its joinpoints
			{

			  if(mode == POPULATE_TABLE_AND_FIND_ADVICE)
			  {

				// __________________________________________________________________
				//					GATHER ADVICE IN THE FIRST PASS
				// __________________________________________________________________
				if(((ASTNode)#attribs).GetFirstChildOfType(ATTRIBUTE_SECTION) != null)
				{
				  if(((ASTNode)#attribs).GetFirstChildOfType(ATTRIBUTE_SECTION).GetFirstChildOfType(ATTRIBUTE).GetFirstChildOfType(IDENTIFIER) != null)
					if(((ASTNode)#attribs).GetFirstChildOfType(ATTRIBUTE_SECTION).GetFirstChildOfType(ATTRIBUTE).GetFirstChildOfType(IDENTIFIER).getText().StartsWith("Advice"))
					{
//						Console.WriteLine("I found some advice!! :-)");
						
						// Add the advice to the list of advice
						// Advice format:    [Advice(AdviceType.before, "call(* *.*(..))")]

						Phx.Aop.AdviceType adType;						
						ASTNode argList = ((ASTNode)#attribs).GetFirstChildOfType(ATTRIBUTE_SECTION).GetFirstChildOfType(ATTRIBUTE).GetFirstChildOfType(POSITIONAL_ARGLIST);
						string varType = #argList.GetFirstChildOfType(POSITIONAL_ARG).GetFirstChildOfType(ATTRIB_ARGUMENT_EXPR).GetFirstChildOfType(EXPR).GetFirstChildOfType(MEMBER_ACCESS_EXPR).getFirstChild().getNextSibling().getText();
						if(varType.Equals("before"))
							adType = Phx.Aop.AdviceType.before;
						else if(varType.Equals("after"))
							adType = Phx.Aop.AdviceType.after;
						else
							adType = Phx.Aop.AdviceType.around;
						
						string pointcutStr = ((ASTNode)#argList.GetFirstChildOfType(POSITIONAL_ARG).getNextSibling()).GetFirstChildOfType(ATTRIB_ARGUMENT_EXPR).GetFirstChildOfType(EXPR).getFirstChild().getText();
						Phx.Aop.Attributes.AdviceAttribute myAdviceAttribute = new Phx.Aop.Attributes.AdviceAttribute(adType, pointcutStr);
						advices.Add( new Wcs.Aop.WcsAdvice(myAdviceAttribute, #md, className) );
						
						
//						Console.WriteLine("This is the advice I found: ");
//						Console.WriteLine("Type = " + varType);
//						Console.WriteLine("pointcut string = " + pointcutStr);
					}
				}
			  } // End if(mode == POPULATE_TABLE_AND_FIND_ADVICE)
				
			}
			{
				if(mode == OUTPUT_FILE) w.Write(sb.ToString());
				sw.Close();
			}
		)
	;
	
memberName [TextWriter w]
	:	qualifiedIdentifier[w]					// interfaceType^ DOT! identifier
	;


	
formalParameterList [TextWriter w, Hashtable SymbolTable]
	:	#(	FORMAL_PARAMETER_LIST 
			(	fixedParameters[w, SymbolTable] ( { if(mode == OUTPUT_FILE) w.Write(", "); } parameterArray[w, SymbolTable] )?
			|	parameterArray[w, SymbolTable]
			)
		)
	;
	
fixedParameters [TextWriter w, Hashtable SymbolTable]
	:	fixedParameter[w, SymbolTable] ( { if(mode == OUTPUT_FILE) w.Write(", "); } fixedParameter[w, SymbolTable] )*
	;
	
fixedParameter [TextWriter w, Hashtable SymbolTable]
	:	#( 	PARAMETER_FIXED attributes[w, SymbolTable] 
			{
				StringBuilder sb = new StringBuilder();
				StringWriter  sw = new StringWriter(sb);
			}
			t:type[sw] id:identifier[sw]
			{
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
				{
					// Add the variable you've detected to the symbol table: (var_name, var_type)
					// If it already exists in the symboltable, remove it and replace it with the local value!
					if(SymbolTable.ContainsKey(#id.getText()))
						SymbolTable.Remove(#id.getText());
					SymbolTable.Add(#id.getText(), typeAsString(#t));
/*					// Print out the symbol table
					IDictionaryEnumerator Enumerator = SymbolTable.GetEnumerator();
					Console.WriteLine("----------------------- TABLE: ------------------------");
					while (Enumerator.MoveNext())
					{
						Console.WriteLine("With fixed parameter contains " + Enumerator.Value + " " + Enumerator.Key);
					}
*/					
				}
			}
			( parameterModifier[w] )?
			{
				if(mode == OUTPUT_FILE) w.Write(sb.ToString());
				sw.Close();
			}
		)
	;
	
parameterModifier [TextWriter w]
	:	r:REF		{ Print(w, #r, null, " "); } 
	|	o:OUT		{ Print(w, #o, null, " "); } 
	;
	
parameterArray [TextWriter w, Hashtable SymbolTable]
	:	#(	parray:PARAMS attributes[w, SymbolTable] { Print(w, #parray, null, " "); } type[w] identifier[w]
		)
	;
	
propertyDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	PROPERTY_DECL attributes[w, SymbolTable] modifiers[w] type[w] qualifiedIdentifier[w] 
			{ if(mode == OUTPUT_FILE) w.Write("{"); }
			accessorDeclarations[w, SymbolTable] curl:CLOSE_CURLY { Print(w, #curl); } 
		)
	;
	
accessorDeclarations [TextWriter w, Hashtable SymbolTable]
	:	(	getAccessorDeclaration[w, SymbolTable] ( setAccessorDeclaration[w, SymbolTable] )?
		|	setAccessorDeclaration[w, SymbolTable] ( getAccessorDeclaration[w, SymbolTable] )?
		)
	;
	
getAccessorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( "get" attributes[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write("get"); } accessorBody[w, SymbolTable] )
	;
	
setAccessorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( "set" attributes[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write("set"); } accessorBody[w, SymbolTable] )
	;
	
accessorBody [TextWriter w, Hashtable SymbolTable]
	:	block[w, SymbolTable]
	|	e:EMPTY_STMT	{ Print(w, #e, null, " "); } 
	;
	
eventDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	evt:EVENT attributes[w, SymbolTable] modifiers[w] { Print(w, #evt, null, " "); } t:type[w]
			(	qualifiedIdentifier[w] ocly:OPEN_CURLY { Print(w, #ocly, null, "{"); }
				eventAccessorDeclarations[w, SymbolTable] cly:CLOSE_CURLY { Print(w, #cly); } 
			|	variableDeclarator[w, SymbolTable, #t.getFirstChild().getText()] ( { if(mode == OUTPUT_FILE) w.Write(", "); } variableDeclarator[w, SymbolTable, #t.getFirstChild().getText()] )* { if(mode == OUTPUT_FILE) w.Write("; "); }
			)
		)
	;
	
eventAccessorDeclarations [TextWriter w, Hashtable SymbolTable]
	:	addAccessorDeclaration[w, SymbolTable] removeAccessorDeclaration[w, SymbolTable]
	|	removeAccessorDeclaration[w, SymbolTable] addAccessorDeclaration[w, SymbolTable]
	;
	
addAccessorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( "add"    attributes[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write("add"); } block[w, SymbolTable] )
	;
	
removeAccessorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( "remove" attributes[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write("remove"); } block[w, SymbolTable] )
	;

indexerDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	INDEXER_DECL attributes[w, SymbolTable] modifiers[w]
			type[w] ( interfaceType[w] { if(mode == OUTPUT_FILE) w.Write("."); } )? t:THIS { Print(w, #t); if(mode == OUTPUT_FILE) w.Write("["); }
			formalParameterList[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write("]"); } {if(mode == OUTPUT_FILE) w.Write("{"); } accessorDeclarations[w, SymbolTable] 
			cly:CLOSE_CURLY { Print(w, #cly); }
		)
	;
	
operatorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	(	#(	UNARY_OP_DECL attributes[w, SymbolTable] modifiers[w]
				type[w] { if(mode == OUTPUT_FILE) w.Write("operator "); } overloadableUnaryOperator[w]
				{ if(mode == OUTPUT_FILE) w.Write("("); } formalParameterList[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(")"); } 
	 			operatorBody[w, SymbolTable]
		 	)
		|	#(	BINARY_OP_DECL attributes[w, SymbolTable] modifiers[w]
				type[w] { if(mode == OUTPUT_FILE) w.Write("operator "); } overloadableBinaryOperator[w]
				{ if(mode == OUTPUT_FILE) w.Write("("); } formalParameterList[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(")"); }
		 		operatorBody[w, SymbolTable]
			 )
		|	#(	CONV_OP_DECL attributes[w, SymbolTable] modifiers[w]
				(	i:IMPLICIT { Print(w, #i, null, " "); if(mode == OUTPUT_FILE) w.Write("operator "); } type[w]
					{ if(mode == OUTPUT_FILE) w.Write("("); } formalParameterList[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(")"); }
					operatorBody[w, SymbolTable]
				|	e:EXPLICIT { Print(w, #e, null, " "); if(mode == OUTPUT_FILE) w.Write("operator "); } type[w] 
					{ if(mode == OUTPUT_FILE) w.Write("("); } formalParameterList[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(")"); }
					operatorBody[w, SymbolTable]
				)
			)
		)
	;
	
overloadableUnaryOperator [TextWriter w]
	:	up:UNARY_PLUS	{ Print(w, #up); }
	|	um:UNARY_MINUS	{ Print(w, #um); }
	|	ln:LOG_NOT		{ Print(w, #ln); }
	|	bn:BIN_NOT		{ Print(w, #bn); }
	|	ic:INC			{ Print(w, #ic); }
	|	dc:DEC			{ Print(w, #dc); }
	|	tr:TRUE			{ Print(w, #tr); }
	|	fa:FALSE		{ Print(w, #fa); }
	;
	
overloadableBinaryOperator [TextWriter w]
	:	pl:PLUS			{ Print(w, #pl); }
	|	ms:MINUS		{ Print(w, #ms); }
	|	st:STAR			{ Print(w, #st); }
	|	dv:DIV 			{ Print(w, #dv); }
	|	md:MOD 			{ Print(w, #md); }
	|	ba:BIN_AND 		{ Print(w, #ba); }
	|	bo:BIN_OR 		{ Print(w, #bo); }
	|	bx:BIN_XOR 		{ Print(w, #bx); }
	|	sl:SHIFTL 		{ Print(w, #sl); }
	|	sr:SHIFTR 		{ Print(w, #sr); }
	|	eq:EQUAL		{ Print(w, #eq); }
	|	nq:NOT_EQUAL 	{ Print(w, #nq); }
	|	gt:GTHAN		{ Print(w, #gt); }
	|	lt:LTHAN 		{ Print(w, #lt); }
	|	ge:GTE 			{ Print(w, #ge); }
	|	le:LTE 			{ Print(w, #le); }
	;
	
operatorBody [TextWriter w, Hashtable SymbolTable]
	:	body[w, SymbolTable]
	;

constructorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	cd:CTOR_DECL 
			{ nodeLinePretend = #cd.Line; currLinePretend = currLine; } 
			attributes[w, SymbolTable] 
			{ nodeLinePretend = -1; currLinePretend = -1;
				if(usedPretendCounter) {
					currLine = #cd.Line;
					usedPretendCounter = false;}
			}			

			modifiers[w] ident:identifier[w]
			{
				StringBuilder sb = new StringBuilder();
				StringWriter  sw = new StringWriter(sb);
			}
			
			{
				// Store this method declaration in the symboltable of this class
				// If it already exists in the symboltable, remove it and replace it with the local value!
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE)
				{
					string id = #ident.getText();			// This is the method's key into the table
					// Create a "Within" WcsJoinpointSignature for this method declaration
					Wcs.Aop.Joinpoints.WcsJoinpointSignature within = new Wcs.Aop.Joinpoints.WcsJoinpointSignature(true, SymbolTable, className);
                    within.node = #cd;
                    within.methodAttributes = #cd.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE_SECTIONS);

					if(SymbolTable.ContainsKey(id))
						SymbolTable.Remove(id);
					SymbolTable.Add(id, within);
				}

				// Create a new *local* SymbolTable to pass to those class properties for which the local
				// tables should not persist past the lifetime of the property
				Hashtable localTable = new Hashtable();
			
				if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
				{
					// All all elements from the global symbol table to your local one
					IDictionaryEnumerator Enumerator = SymbolTable.GetEnumerator();
					while (Enumerator.MoveNext())
					{
						localTable.Add(Enumerator.Key, Enumerator.Value);
					}
				}
			}			
			
			methodBody[sw, localTable, #cd]
			{ if(mode == OUTPUT_FILE) w.Write("("); } ( formalParameterList[w, localTable] )? { if(mode == OUTPUT_FILE) w.Write(") "); }
			( constructorInitializer[w] )? 
			{
				if(mode == OUTPUT_FILE) w.Write(sb.ToString());
				sw.Close();
			}
		)
	;
	
constructorInitializer [TextWriter w]
	:	{ if(mode == OUTPUT_FILE) w.Write(": "); }
		(	#( b:BASE { Print(w, #b); if(mode == OUTPUT_FILE) w.Write("("); } ( argumentList[w] )? )
		|	#( t:THIS { Print(w, #t); if(mode == OUTPUT_FILE) w.Write("("); } ( argumentList[w] )? )
		)
		{ if(mode == OUTPUT_FILE) w.Write(")"); }
	;


staticConstructorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	STATIC_CTOR_DECL attributes[w, SymbolTable] modifiers[w] identifier[w] { if(mode == OUTPUT_FILE) w.Write("()"); }
			staticConstructorBody[w, SymbolTable]
		)
	;
	
staticConstructorBody [TextWriter w, Hashtable SymbolTable]
	:	body[w, SymbolTable]
	;
	
destructorDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( 	DTOR_DECL attributes[w, SymbolTable] modifiers[w] { if(mode == OUTPUT_FILE) w.Write("~"); } identifier[w] { if(mode == OUTPUT_FILE) w.Write("()"); }
			destructorBody[w, SymbolTable]
		)
	;
	
destructorBody [TextWriter w, Hashtable SymbolTable]
	:	body[w, SymbolTable]
	;

	
//
// A.2.7 Structs
//

structDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( 	st:STRUCT attributes[w, SymbolTable] modifiers[w] { PrintWithBeforeCheck(w, #st, null, " "); } identifier[w]
			#(	STRUCT_BASE 
				{
					bool needToPrintColon = true;
					bool needToPrintComma = false;
				}
				(	{
						if ( needToPrintColon )
						{
							if(mode == OUTPUT_FILE) w.Write(" : ");
							needToPrintColon = false;
						}
						if ( needToPrintComma )
							if(mode == OUTPUT_FILE) w.Write(", ");
					}
					type[w]
					{ needToPrintComma = true; }
				)* 
			)
			#(	tb:TYPE_BODY { Print(w, #tb); }

				{	// Create a new *local* SymbolTable to pass to the statements inside the method block
					// This way variables declared locally with override the global vars
					Hashtable localTable = new Hashtable();
	
					if (mode == POPULATE_TABLE_AND_FIND_ADVICE || mode == CREATE_JOINPOINTS)
					{
						// Add all elements from the global symbol table to your local one
						IDictionaryEnumerator Enumerator = SymbolTable.GetEnumerator();
						while (Enumerator.MoveNext())
						{
							localTable.Add(Enumerator.Key, Enumerator.Value);
						}
					}
				}
				
				structMemberDeclarations[w, localTable] cc:CLOSE_CURLY { Print(w, #cc); } 
			  )
		)
	;
	
structMemberDeclarations [TextWriter w, Hashtable SymbolTable]
	:	#(	MEMBER_LIST
			(	structMemberDeclaration[w, SymbolTable]
			|	preprocessorDirective[w, CodeMaskEnums.StructMemberDeclarations, SymbolTable]
			)*
		)
	;
	
structMemberDeclaration [TextWriter w, Hashtable SymbolTable]
	:	typeMemberDeclaration[w, SymbolTable]
	;

	
//
// A.2.8 Arrays
//

rankSpecifiers [TextWriter w]
	:	#(	ARRAY_RANKS
			( 	rankSpecifier[w]
			)*
		)
	;
	
rankSpecifier  [TextWriter w]
	:	#(	r:ARRAY_RANK { Print(w, #r); }
			(	c:COMMA { Print(w, #c); if(mode == OUTPUT_FILE) w.Write(","); }
			)* 
			{ if(mode == OUTPUT_FILE) w.Write("]"); }
		)
	;
	
arrayInitializer [TextWriter w]
	:	#(	ai:ARRAY_INIT { Print(w, #ai); } ( variableInitializerList[w] )? cc:CLOSE_CURLY { Print(w, #cc); } )
	;
	
variableInitializerList [TextWriter w]
	:	#( VAR_INIT_LIST variableInitializer[w] ( { if(mode == OUTPUT_FILE) w.Write(", "); } variableInitializer[w] )* )
	;


// 
// A.2.9 Interfaces
//

interfaceDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	iface:INTERFACE attributes[w, SymbolTable] modifiers[w] { PrintWithBeforeCheck(w, #iface, null, " "); } identifier[w]
			#(	INTERFACE_BASE 
				{
					bool needToPrintColon = true;
					bool needToPrintComma = false;
				}
				(	{
						if ( needToPrintColon )
						{
							if(mode == OUTPUT_FILE) w.Write(" : ");
							needToPrintColon = false;
						}
						if ( needToPrintComma )
							if(mode == OUTPUT_FILE) w.Write(", ");
					}
					type[w]
					{ needToPrintComma = true; }
				)* 
			)
			interfaceBody[w, SymbolTable]
		)
	;
	
interfaceBody [TextWriter w, Hashtable SymbolTable]
	:	#(	tb:TYPE_BODY { Print(w, #tb); } interfaceMemberDeclarations[w, SymbolTable] cc:CLOSE_CURLY { Print(w, #cc); } )
	;
	
interfaceMemberDeclarations [TextWriter w, Hashtable SymbolTable]
	:	#(	MEMBER_LIST
			(	interfaceMemberDeclaration[w, SymbolTable]
			|	preprocessorDirective[w, CodeMaskEnums.InterfaceMemberDeclarations, SymbolTable]
			)*
		)
	;
	
interfaceMemberDeclaration [TextWriter w, Hashtable SymbolTable]
	:	(	interfaceMethodDeclaration[w, SymbolTable]
		|	interfacePropertyDeclaration[w, SymbolTable]
		|	interfaceEventDeclaration[w, SymbolTable]
		|	interfaceIndexerDeclaration[w, SymbolTable]
		)
	;
	
interfaceMethodDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	METHOD_DECL attributes[w, SymbolTable] modifiers[w] type[w] qualifiedIdentifier[w]
			{
				StringBuilder sb = new StringBuilder();
				StringWriter  sw = new StringWriter(sb);
			}
			e:EMPTY_STMT { Print(sw, #e); }
			{ if(mode == OUTPUT_FILE) w.Write("("); } ( formalParameterList[w, SymbolTable] )?
			{
				if(mode == OUTPUT_FILE) w.Write(") ");
				if(mode == OUTPUT_FILE) w.Write(sb.ToString());
				sw.Close();
			}
		)
	;
	
interfacePropertyDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	PROPERTY_DECL attributes[w, SymbolTable] modifiers[w] type[w] identifier[w]
			{ if(mode == OUTPUT_FILE) w.Write("{"); } accessorDeclarations[w, SymbolTable] cc:CLOSE_CURLY { Print(w, #cc); }
		)
	;
	
interfaceEventDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	evt:EVENT attributes[w, SymbolTable] modifiers[w] { Print(w, #evt, null, " "); }
			t:type[w] variableDeclarator[w, SymbolTable, #t.getFirstChild().getText()]		 { if(mode == OUTPUT_FILE) w.Write(";");  }
		)
	;
	
interfaceIndexerDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	INDEXER_DECL attributes[w, SymbolTable] modifiers[w] type[w] t:THIS { Print(w, #t); if(mode == OUTPUT_FILE) w.Write("["); } 
			formalParameterList[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write("]"); }
			{ if(mode == OUTPUT_FILE) w.Write("{"); } accessorDeclarations[w, SymbolTable] cc:CLOSE_CURLY { Print(w, #cc); }
		)
	;

	
//
//	A.2.10 Enums
//

enumDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#( 	en:ENUM attributes[w, SymbolTable] modifiers[w] { PrintWithBeforeCheck(w, #en, null, " "); } identifier[w]
			#( 	ENUM_BASE ( { if(mode == OUTPUT_FILE) w.Write(" : "); } type[w] )? )
			#(	tb:TYPE_BODY { Print(w, #tb); }
				#(	MEMBER_LIST 
					( enumMemberDeclaration[w, SymbolTable] { if(mode == OUTPUT_FILE) w.Write(","); } )* 
				) 
				cc:CLOSE_CURLY { Print(w, #cc); }
			)
		)
	;
	
enumMemberDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	id:IDENTIFIER attributes[w, SymbolTable] { Print(w, #id, null, " "); } ( { if(mode == OUTPUT_FILE) w.Write(" = "); } constantExpression[w] )?
		)
 	;


//
// A.2.11 Delegates
//

delegateDeclaration [TextWriter w, Hashtable SymbolTable]
	:	#(	dlg:DELEGATE attributes[w, SymbolTable] modifiers[w] { PrintWithBeforeCheck(w, #dlg, null, " "); }
			type[w] identifier[w] { if(mode == OUTPUT_FILE) w.Write("("); } ( formalParameterList[w, SymbolTable] )? { if(mode == OUTPUT_FILE) w.Write(");"); }
		)
	;
	

//
// A.2.12 Attributes
//

globalAttributes [TextWriter w, Hashtable SymbolTable]
	:	#(	GLOBAL_ATTRIBUTE_SECTIONS 
			(	globalAttributeSection[w]
			|	preprocessorDirective[w, CodeMaskEnums.GlobalAttributes, SymbolTable]
			)*
		)
	;
	
globalAttributeSection [TextWriter w]
	:	#(	sect:GLOBAL_ATTRIBUTE_SECTION { PrintWithBeforeCheck(w, #sect); if(mode == OUTPUT_FILE) w.Write("assembly: "); } 
			( attribute[w] )+  { if(mode == OUTPUT_FILE) w.Write("]"); }
		)
	;

attributes [TextWriter w, Hashtable SymbolTable]
	:	#(	ATTRIBUTE_SECTIONS 
			(	attributeSection[w]
			|	preprocessorDirective[w, CodeMaskEnums.Attributes, SymbolTable]
			)*
		)
	;
	
attributeSection [TextWriter w]
	:	#(	sect:ATTRIBUTE_SECTION { PrintWithBeforeCheck(w, #sect); } ( attributeTarget[w] )?
			( attribute[w] )+  { if(mode == OUTPUT_FILE) w.Write("]"); }
		{ if(nodeLinePretend > 0) 
		  {
			if(mode == OUTPUT_FILE)
				w.Write("\n");
			lineCountDiff++;
		  }	
		}
		)
	;
	
attributeTarget[TextWriter w]
	:	(	fv:"field"			{ Print(w, #fv); }
		|	ev:EVENT			{ Print(w, #ev); }
		|	mv:"method"			{ Print(w, #mv); }
		|	mo:"module"			{ Print(w, #mo); }
		|	pa:"param"			{ Print(w, #pa); }
		|	pr:"property"		{ Print(w, #pr); }
		|	re:RETURN			{ Print(w, #re); }
		|	ty:"type"			{ Print(w, #ty); }
		)
		{ if(mode == OUTPUT_FILE) w.Write(": "); }
	;

attribute [TextWriter w]
	:	#( ATTRIBUTE typeName[w] attributeArguments[w] ) 
	;
	
attributeArguments [TextWriter w]
	:	{ if(mode == OUTPUT_FILE) w.Write("("); }  ( positionalArgumentList[w] )? ( namedArgumentList[w] )? { if(mode == OUTPUT_FILE) w.Write(") "); } 
	;
	
positionalArgumentList [TextWriter w]
	:	#(	POSITIONAL_ARGLIST positionalArgument[w]
			( { if(mode == OUTPUT_FILE) w.Write(", "); } positionalArgument[w] )* 
		)
	;
	
positionalArgument [TextWriter w]
	:	#( POSITIONAL_ARG attributeArgumentExpression[w] )
	;
	
namedArgumentList [TextWriter w]
	:	#(	NAMED_ARGLIST namedArgument[w]
			( { if(mode == OUTPUT_FILE) w.Write(", "); } namedArgument[w] )* 
		)
	;
	
namedArgument [TextWriter w]
	:	#( NAMED_ARG identifier[w] { if(mode == OUTPUT_FILE) w.Write(" = "); } attributeArgumentExpression[w] )
	;
	
attributeArgumentExpression [TextWriter w]
	:	#( ATTRIB_ARGUMENT_EXPR expression[w] )
	;

//
// A.3 Grammar extensions for unsafe code
// 

fixedPointerDeclarator [TextWriter w]
	:	#( PTR_DECLARATOR identifier[w] { if(mode == OUTPUT_FILE) w.Write(" = "); } fixedPointerInitializer[w] )
	;
	
fixedPointerInitializer [TextWriter w]
	:	#(	PTR_INIT
			(	b:BIN_AND { Print(w, #b); } variableReference[w]
			|	expression[w]
			)
		)
	;	
	
stackallocInitializer [TextWriter w]
	:	#( s:STACKALLOC { Print(w, #s, null, " "); } unmanagedType[w] { if(mode == OUTPUT_FILE) w.Write("["); } expression[w] { if(mode == OUTPUT_FILE) w.Write("]"); } )
	;




//======================================
// Preprocessor Directives
//======================================

justPreprocessorDirectives [TextWriter w, Hashtable SymbolTable]
	:	#(	PP_DIRECTIVES 
			(	preprocessorDirective[w, CodeMaskEnums.PreprocessorDirectivesOnly, SymbolTable] 
			)* 
		)
	;
	
preprocessorDirective [TextWriter w, CodeMaskEnums codeMask, Hashtable SymbolTable]
	:	#( d1:PP_DEFINE   { Print(w, #d1, null, " "); } d2:PP_IDENT { Print(w, #d2); } )
	|	#( u1:PP_UNDEFINE { Print(w, #u1, null, " "); } u2:PP_IDENT { Print(w, #u2); } )
	|	#(	l1:PP_LINE    { Print(w, #l1, null, " "); }
			(	l2:DEFAULT   { Print(w, #l2); }
			|	l3:PP_NUMBER { Print(w, #l3, null, " "); } ( l4:PP_FILENAME { Print(w, #l4); } )?
			)
		)
	|	#(	e1:PP_ERROR   { Print(w, #e1); } )
	|	#(	w1:PP_WARNING { Print(w, #w1); } )
	|	regionDirective[w, codeMask, SymbolTable]
	|	conditionalDirective[w, codeMask, SymbolTable]
	;
	
// The following are changed by Boriana to not print out the ppMessages, since I don't use those anymore
regionDirective [TextWriter w, CodeMaskEnums codeMask, Hashtable SymbolTable]
	:	#(	r1:PP_REGION { Print(w, #r1, null, " "); } directiveBlock[w, codeMask, SymbolTable]
			#( r2:PP_ENDREGION { Print(w, #r2, null, " "); } )
		)
	;

conditionalDirective [TextWriter w, CodeMaskEnums codeMask, Hashtable SymbolTable]
	:	#(	c1:PP_COND_IF        { Print(w, #c1, null, " "); } preprocessExpression[w] directiveBlock[w, codeMask, SymbolTable]
			( #( c2:PP_COND_ELIF { Print(w, #c2, null, " "); } preprocessExpression[w] directiveBlock[w, codeMask, SymbolTable] ) )*
			( #( c3:PP_COND_ELSE { Print(w, #c3, null, " "); }                         directiveBlock[w, codeMask, SymbolTable] ) )?
			c4:PP_COND_ENDIF     { Print(w, #c4); }
		)
	;

directiveBlock [TextWriter w, CodeMaskEnums codeMask, Hashtable SymbolTable]
	:	#(	PP_BLOCK
			(	{ NotExcluded(codeMask, CodeMaskEnums.UsingDirectives) }?				usingDirective[w]
			|	{ NotExcluded(codeMask, CodeMaskEnums.GlobalAttributes) }?				globalAttributeSection[w]
			|	{ NotExcluded(codeMask, CodeMaskEnums.Attributes) }?					attributeSection[w]
			|	{ NotExcluded(codeMask, CodeMaskEnums.NamespaceMemberDeclarations) }?	namespaceMemberDeclaration[w, SymbolTable]
			|	{ NotExcluded(codeMask, CodeMaskEnums.ClassMemberDeclarations) }?		classMemberDeclaration[w, SymbolTable]
			|	{ NotExcluded(codeMask, CodeMaskEnums.StructMemberDeclarations) }?		structMemberDeclaration[w, SymbolTable]
			|	{ NotExcluded(codeMask, CodeMaskEnums.InterfaceMemberDeclarations) }?	interfaceMemberDeclaration[w, SymbolTable]
			|	{ NotExcluded(codeMask, CodeMaskEnums.Statements) }?					stmt:statement[w, #stmt, SymbolTable, null]
			|	preprocessorDirective[w, codeMask, SymbolTable]
			)*
		)
	;
	

preprocessExpression [TextWriter w]
	:	#( PP_EXPR preprocessExpr[w] )
	;

preprocessExpr [TextWriter w]
	:	#( o:LOG_OR    preprocessExpr[w] { Print(w, #o); } preprocessExpr[w] )
	|	#( a:LOG_AND   preprocessExpr[w] { Print(w, #a); } preprocessExpr[w] )
	|	#( e:EQUAL     preprocessExpr[w] { Print(w, #e); } preprocessExpr[w] )
	|	#( n:NOT_EQUAL preprocessExpr[w] { Print(w, #n); } preprocessExpr[w] )
	|	preprocessPrimaryExpression[w]
	;
	
preprocessPrimaryExpression [TextWriter w]
	:	i:PP_IDENT		{ Print(w, #i);  } 
	|	tr:TRUE			{ Print(w, #tr); }
	|	fa:FALSE		{ Print(w, #fa); }
	|	#( l:LOG_NOT 	{ Print(w, #l);  } preprocessPrimaryExpression[w] )
	|	#( p:PAREN_EXPR { Print(w, #p);  } preprocessExpr[w] { if(mode == OUTPUT_FILE) w.Write(") "); } )
	;
