//-----------------------------------------------------------------------------
//
// Copyright (C) Boriana Ditcheva.  All Rights Reserved.
//
// $Header: /Wicca/Wicca.wcs/WcsJoinpoint.cs 16    5/27/06 10:22p Bhd2105 $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Wicca.wcs/WcsJoinpoint.cs $
 * 
 * 16    5/27/06 10:22p Bhd2105
 * 
 * 15    5/24/06 1:53a Eaddy
 * o Breakpoint weaver now supports [Arg]
 * 
 * 14    5/19/06 5:17p Bhd2105
 * Add support for statement attributes that are at the method level
 * (those that the programmer has put there, not ones that the compiler
 * lifts up and is thus aware of).
 * 
 * 13    5/19/06 12:22a Bhd2105
 * Now when the joinpoint is creating its wcsSig, if it detects that the
 * joinpoint type is CALL, it replaces its wcsSig with the signature of
 * the method declaration corresponding to this method call, which it
 * finds in the tablesList (now the tablesList contains not only variables
 * but also method declarations).
 * 
 * 
 * 11    5/16/06 4:21p Bhd2105
 * Implement getDataForAttribute() by calling the SimpleJoinpoint's method
 * 
 * 10    5/14/06 4:39p Bhd2105
 * the ToString() method is now back to its original definition. The
 * commented out stuff should be used for debugging as it is very detailed
 * 
 * 9     5/11/06 11:33p Bhd2105
 * Added the getAttributeArgs() and getAttributeType() methods which
 * simply call the respective methods on the joinpoints signature and
 * within members
 * 
 * 8     5/11/06 2:41p Bhd2105
 * 
 * 7     5/11/06 5:43a Bhd2105
 * Implemented function to get the arguments from the attributes of this
 * joinpiont
 * 
 * 6     5/10/06 9:00p Bhd2105
 * I've updated the within sourceLocation to be the same as the wcsSig
 * source location. It's the same as the statement location, not the
 * method location
 * 
 * 5     5/08/06 4:18a Bhd2105
 * 
 * 4     5/08/06 12:55a Bhd2105
 * The joinpoint now passes the className to its joinpointSignatures.
 * Also, if the joinpoint has type EXECUTION, the argTypes of the method
 * it's in are saved as the argTypes the jointpoint also.
 * 
 * 3     5/03/06 1:45a Bhd2105
 * 
 * 2     4/27/06 3:26p Bhd2105
 */

using System;
using System.Text;
using System.Collections;				// For the hashtable used for a symbol table
using System.Collections.Generic;



namespace Wcs.Aop.Joinpoints
{
    public class WcsJoinpoint : Phx.Aop.Joinpoints.IJoinpoint
    {
        // --------- VARIABLES -----------------
        private Kunle.CSharpParser.ASTNode parentNode = null;
        private Kunle.CSharpParser.ASTNode currNode = null;
        private WcsJoinpointSignature wcsSig = null;
        private WcsJoinpointSignature within = null;
        public Phx.Aop.Joinpoints.JoinpointType joinpointType = Phx.Aop.Joinpoints.JoinpointType.UNINITIALIZED;
        private Hashtable currTable = null;
        private Wicca.wcs.TablesList tablesList = null;
        private string className = "";

        // Constructors
        public WcsJoinpoint(Kunle.CSharpParser.ASTNode p, Kunle.CSharpParser.ASTNode c, Hashtable table, Wicca.wcs.TablesList tables, string cname)
        {
            parentNode = p;
            currNode = c;
            currTable = table;
            tablesList = tables;
            className = cname;
        }


        public Phx.Aop.Joinpoints.JoinpointType JoinpointType
        {
            get
            {
                if (joinpointType != Phx.Aop.Joinpoints.JoinpointType.UNINITIALIZED)
                {
                    return joinpointType;
                }

                return joinpointType;
            }

        }


        public Phx.Aop.Joinpoints.IJoinpointSignature Signature
        {
            get
            {
                if (wcsSig == null)
                {
                    wcsSig = new WcsJoinpointSignature(false, joinpointType, currTable, tablesList, className);
                    wcsSig.node = currNode;

                    // If this is an EXECUTION joinpoint, set its argTypes to the method's argTypes in the within joinpoint:
                    if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.EXECUTION)
                    {
                        if (((WcsJoinpointSignature)Within).ArgTypes != null)
                        {
                            int numArgs = ((WcsJoinpointSignature)Within).sig.argTypes.Length;
                            wcsSig.sig.argTypes = new StringBuilder[numArgs];
                            for (int i = 0; i < numArgs; i++)
                                wcsSig.sig.argTypes[i] = within.sig.argTypes[i];
                        }
                    }
                    else if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.CALL)
                    {
                        // This is a call. Find the methodDecl of this method call and set the joinpoint signature
                        // to be that of the methodDecl joinpoint stored
                        WcsJoinpointSignature methodSig = tablesList.findMethodDecl(wcsSig.DeclaringType.ToString(), wcsSig.MemberName.ToString());

                        if (methodSig != null)
                        {
                            methodSig.joinpointType = joinpointType;
                            wcsSig = methodSig;
                            wcsSig.location = new WcsSourceLocation();
                            wcsSig.location.node = currNode;
                        }
                    }

                    // Stores a list of the statement-level attributes for this node
                    wcsSig.allAttributes = parentNode.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE_SECTIONS);
                }

                return wcsSig;
            }
        }

        public Phx.Aop.Joinpoints.IJoinpointSignature Within
        {
            get
            {
                if (within == null)
                {
                    within = new WcsJoinpointSignature(true, joinpointType, currTable, tablesList, className);
                    within.node = parentNode;
                    // The within sourceLocation is the same as the wcsSig source location.
                    // It's the same as the statement location, not the method location
                    within.location = new WcsSourceLocation();
                    within.location.node = currNode;
                }

                return within;
            }
        }

        public StringBuilder Target
        {
            get
            {
                if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.EXECUTION)
                {
                    return This;
                }
                else
                {
                    return Signature.DeclaringType;
                }
            }
        }

        public StringBuilder This
        {
            get
            {
                return within.DeclaringType;
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(JoinpointType.ToString());
            sb.Append(' ');
            sb.Append(((WcsJoinpointSignature)Signature).debugToString());
//            sb.Append(Signature.ToString());
//            sb.Append(" in ");
            sb.Append("\n\tin ");
            sb.Append(((WcsJoinpointSignature)Within).debugToString());
//            sb.Append(Within.ToString());
            sb.Append(' ');
            sb.Append(Signature.SourceLocation.ToString());

            return sb.ToString();
        }

        public object getDataForAttribute(Type attributeType, object attributeVal)
        {
            return Phx.Aop.Joinpoints.SimpleJoinpoint.getDataForAttribute(attributeType, attributeVal, this);
        }

        public List<Kunle.CSharpParser.ASTNode> getAttributeArgs(int attrIndex)
        {
            if (attrIndex == -1)
                return (null);
            List<Kunle.CSharpParser.ASTNode> attrArgs = ((WcsJoinpointSignature)Signature).getAttributeArgs(attrIndex);
            if (attrArgs == null)
                attrArgs = ((WcsJoinpointSignature)Within).getAttributeArgs(attrIndex);

            return (attrArgs);
        }
        public string getAttributeType(int attrIndex)
        {
            if (attrIndex == -1)
                attrIndex = 0;          // By default
            string attrType = ((WcsJoinpointSignature)Signature).getAttributeType(attrIndex);
            if (attrType.Equals(""))
                attrType = ((WcsJoinpointSignature)Within).getAttributeType(attrIndex);

            return (attrType);
        }


    }
}
