//-----------------------------------------------------------------------------
//
// Copyright (C) Boriana Ditcheva.  All Rights Reserved.
//
// $Header: /Wicca/Wicca.wcs/WcsJoinpointSignature.cs 24    5/27/06 10:22p Bhd2105 $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Wicca.wcs/WcsJoinpointSignature.cs $
 * 
 * 24    5/27/06 10:22p Bhd2105
 * 
 * 23    5/26/06 6:24p Bhd2105
 * 
 * 22    5/24/06 1:34a Eaddy
 * o Added ArgNames to IJoinpointSignature so that we can
 * pass context args to the advice method
 * 
 * 21    5/19/06 5:20p Bhd2105
 * 1) Add support for statement annotations that are at the method-level,
 * but are not attached to the statement (because they are directly put
 * there by the programmer, as opposed to being "lifted" by the compiler,
 * so the compiler is not aware of them unless it checks the annotations
 * attached to the method). 
 * 2) Implemented support for constructor calls as CALL joinpoints in the
 * methods.
 * 3) Add support for finding out the type of a node that is casted.
 * 
 * 20    5/19/06 12:23a Bhd2105
 * Now contains methodAttributes and statementAttributes separately.
 * 
 * 
 * 18    5/16/06 4:23p Bhd2105
 * Attribute related methods have been modified to work with statement
 * annotations that do not have the "false" flag as one of the arguments
 * 
 * 17    5/14/06 4:42p Bhd2105
 * Cleaned up code and erased the code that was old and commented out.
 * 
 * 
 * 15    5/11/06 11:34p Bhd2105
 * Modified functions to fix bugs. They seem to work correctly now.
 * 
 * 14    5/11/06 2:41p Bhd2105
 * 
 * 13    5/11/06 5:43a Bhd2105
 * 
 * 12    5/10/06 9:01p Bhd2105
 * AttributeTypes is now implemented, and Attributes is implemented
 * locally as a list of ASTNodes.
 * 
 * 
 * 11    5/10/06 11:43a Eaddy
 * o Fixed compile errors (now uses IEnumerator and IEnumerable ifaces)
 * 
 * 10    5/09/06 9:44p Bhd2105
 * 
 * 9     5/09/06 1:07a Bhd2105
 * Added typeAsString method, and finished implementing attributes.
 * ToString() now includes attribute printouts for debugging.  Added more
 * special cases to the findType() method.
 * 
 * 8     5/08/06 4:19a Bhd2105
 * Starting on the attributes part
 * 
 * 7     5/08/06 12:53a Bhd2105
 * Implemented ArgType. Fixed several bugs with DeclaringType and the
 * findType() function. Execution joinpoints now are given the argTypes of
 * the parent method. Have also added declaringType and memberName for SET
 * joinpoints.
 * 
 * 6     5/04/06 4:59p Eaddy
 * o Changed from IEnumerable<object> to IEnumerator
 * 
 * 5     5/04/06 4:39p Eaddy
 * o Property names now make more sense
 * o Added properties for accessing attributes and args
 * as an object collection
 * 
 * 4     5/04/06 3:56p Eaddy
 * o Updated to reflect changes in IJoinpoint
 * 
 * 3     5/04/06 2:40p Bhd2105
 * 
 * 2     5/03/06 1:46a Bhd2105
 * Have added the findType() method which returns the type of the given
 * variable. It searches in the local symbol table first, and symbol
 * tables of other classes (that have been parsed with this one) if
 * necessary.
 * 
 */

using System.Text;
using System.Collections;				// For the hashtable used for a symbol table
using System;
using System.Collections.Generic;



namespace Wcs.Aop.Joinpoints
{
    public class WcsJoinpointSignature : Phx.Aop.Joinpoints.IJoinpointSignature
    {
        // ----------------- VARIABLES --------------
        public Phx.Aop.Joinpoints.SimpleJoinpointSignature sig = new Phx.Aop.Joinpoints.SimpleJoinpointSignature();
        public Kunle.CSharpParser.ASTNode node = null;
        private Hashtable symbolTable = null;
        private Wicca.wcs.TablesList tablesList = null;
        public Phx.Aop.Joinpoints.JoinpointType joinpointType = Phx.Aop.Joinpoints.JoinpointType.UNINITIALIZED;
        private List<Kunle.CSharpParser.ASTNode> attributes = null;
        public Kunle.CSharpParser.ASTNode methodAttributes = null;
        public Kunle.CSharpParser.ASTNode allAttributes = null;      // Search through these for statement-level attributes which might be attached above at the method level
        public WcsSourceLocation location = null;
        private bool isWithin;
        private string className = null;

        private const int DECLARING_TYPE_MODE = 1;					            // Describe the mode you can search a type in
        private const int NOT_DECLARING_TYPE_MODE = 2;


        // Constructors
        public WcsJoinpointSignature(bool isWithin, Phx.Aop.Joinpoints.JoinpointType jpType, Hashtable table, Wicca.wcs.TablesList tables, string cName)
        {
            this.isWithin = isWithin;
            joinpointType = jpType;
            symbolTable = table;
            tablesList = tables;
            className = cName;
        }
        // This constructor is used when creating Within Joinpoints in the first pass of the classes
        // When the joinpoint type of the statements and the full list of tables aren't known
        //      *NOTE: this constructor should only be used to create WITHIN joinpoints, because
        //      only they don't use the fields not provided here
        public WcsJoinpointSignature(bool isWithin, Hashtable table, string cName)
        {
            this.isWithin = isWithin;
            symbolTable = table;
            className = cName;
        }


        public IEnumerable<string> AttributeTypes
        {
            get
            {
                if (isWithin)
                {
                    if (sig.AttributeTypes != null)
                        return sig.AttributeTypes;

                    // Only count method-level annotations for the within signature
                    if (methodAttributes != null)
                        if (methodAttributes.getFirstChild() != null)
                        {
                            int numChildren = methodAttributes.NumChildren();
                            List<string> temp = new List<string>(numChildren);
                            Kunle.CSharpParser.ASTNode attr = (Kunle.CSharpParser.ASTNode)methodAttributes.getFirstChild();
                            while (attr != null)
                            {
                                string id = attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText();
                                // Only add this attribute if it's not a statement attribute
                                if (!id.StartsWith("Statement"))
                                {
                                    temp.Add(id);
                                }
                                else
                                    numChildren--;

                                attr = (Kunle.CSharpParser.ASTNode)attr.getNextSibling();
                            }

                            // If there were any statement attributes, this part is so you can adjust the size of the attributes.
                            List<string> correctNumTemp = new List<string>(numChildren);
                            IEnumerator enumTemp = temp.GetEnumerator();
                            for (int i = 0; i < numChildren; i++)
                            {
                                enumTemp.MoveNext();
                                correctNumTemp.Add((string)enumTemp.Current);
                            }

                            sig.AttributeTypes = new List<string>(numChildren);
                            sig.AttributeTypes = correctNumTemp;
                        }

                    return (sig.AttributeTypes);
                }
                else
                {
                 if (allAttributes != null)
                    {
                        int numChildren = allAttributes.NumChildren();
						List<string> temp = new List<string>(numChildren);
                        Kunle.CSharpParser.ASTNode attr = (Kunle.CSharpParser.ASTNode) allAttributes.getFirstChild();
                        while (attr != null)
                        {
                            string id = attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText();
                            // Only add this attribute if it's a statement attribute, and it actually refers to the current
                            // statement, i.e. it has the same line number!
                            if (id.StartsWith("Statement")
                                &&
                                uint.Parse( attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARGLIST).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARG).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIB_ARGUMENT_EXPR).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR).
//                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INT_LITERAL).getText()) == (SourceLocation.Line))
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INT_LITERAL).getText()) == (node.Line))
                            {
                                id = ((Kunle.CSharpParser.ASTNode)attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARGLIST).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARG).getNextSibling()).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIB_ARGUMENT_EXPR).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.STRING_LITERAL).getText();

                                if(id.IndexOf("\"") >= 0)
                                    temp.Add(id.Substring(id.IndexOf("\"")+1, id.LastIndexOf("\"")-1));
                                else
                                    temp.Add(id);
                            }
                            else
                                numChildren--;

                            attr = (Kunle.CSharpParser.ASTNode)attr.getNextSibling();
                        }

                        // If there were any statement attributes, this part is so you can adjust the size of the attributes.
                        List<string> correctNumTemp = new List<string>(numChildren);
                        IEnumerator enumTemp = temp.GetEnumerator();
                        for (int i = 0; i < numChildren; i++ )
                        {
                            enumTemp.MoveNext();
                            correctNumTemp.Add((string)enumTemp.Current);
                        }

                        sig.AttributeTypes = new List<string>(numChildren);
                        sig.AttributeTypes = correctNumTemp;
                    }

                    return (sig.AttributeTypes);
                }
            }
        }



        public IEnumerator Attributes
        {
            get
            {
                if (isWithin)
                {
                    if (attributes != null)
                        return attributes.GetEnumerator();

                    // Only count method-level annotations for the within signature
                    if (methodAttributes != null)
                        if (methodAttributes.getFirstChild() != null)
                        {
                            int numChildren = methodAttributes.NumChildren();
                            Kunle.CSharpParser.ASTNode attr = (Kunle.CSharpParser.ASTNode)methodAttributes.getFirstChild();
                            List<Kunle.CSharpParser.ASTNode> temp = new List<Kunle.CSharpParser.ASTNode>(numChildren);
                            while (attr != null)
                            {
                                string id = attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText();
                                // Only add this attribute if it's not a statement attribute
                                if (!id.StartsWith("Statement"))
                                {
                                    temp.Add(attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE));
                                }
                                else
                                    numChildren--;

                                attr = (Kunle.CSharpParser.ASTNode)attr.getNextSibling();
                            }

                            // If there were any statement attributes, this part is so you can adjust the size of the attributes.
                            attributes = new List<Kunle.CSharpParser.ASTNode>(numChildren);
                            IEnumerator enumTemp = temp.GetEnumerator();
                            for (int i = 0; i < numChildren; i++)
                            {
                                enumTemp.MoveNext();
                                attributes.Add((Kunle.CSharpParser.ASTNode)enumTemp.Current);
                            }
                            return attributes.GetEnumerator();

                        }
                }
                else
                {
                    // Only count statement-level annotations for the joinpoint signature
                    if (allAttributes != null)
                    {
                        int numChildren = allAttributes.NumChildren();
                        List<Kunle.CSharpParser.ASTNode> temp = new List<Kunle.CSharpParser.ASTNode>(numChildren);
                        Kunle.CSharpParser.ASTNode attr = (Kunle.CSharpParser.ASTNode)allAttributes.getFirstChild();
                        while (attr != null)
                        {
                            string id = attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText();
                            // Only add this attribute if it's a statement attribute, and it actually refers to the current
                            // statement, i.e. it has the same line number!
                            if (id.StartsWith("Statement")
                                &&
                                uint.Parse(attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARGLIST).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARG).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIB_ARGUMENT_EXPR).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR).
//                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INT_LITERAL).getText()) == (SourceLocation.Line))
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INT_LITERAL).getText()) == (node.Line))
                            {
                                temp.Add(attr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIBUTE));
                            }
                            else
                                numChildren--;

                            attr = (Kunle.CSharpParser.ASTNode)attr.getNextSibling();
                        }

                        // If there were any statement attributes, this part is so you can adjust the size of the attributes.
                        List<Kunle.CSharpParser.ASTNode> correctNumTemp = new List<Kunle.CSharpParser.ASTNode>(numChildren);
                        IEnumerator enumTemp = temp.GetEnumerator();
                        for (int i = 0; i < numChildren; i++)
                        {
                            enumTemp.MoveNext();
                            correctNumTemp.Add((Kunle.CSharpParser.ASTNode)enumTemp.Current);
                        }

                        attributes = new List<Kunle.CSharpParser.ASTNode>(numChildren);
                        attributes = correctNumTemp;
                        return attributes.GetEnumerator();
                    }

                }

                return (null);
            }
        }



        public StringBuilder Access
        {
            get
            {
                if (sig.access != null && sig.access.Length > 0)
                    return sig.access;

                else if (isWithin)
                {
                    // This is the class the joinpoint is within. Check to see its access type
                    sig.SetAccess(node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.MODIFIERS).getText());
                    return sig.access;
                }
                else
                {
                    sig.SetAccess("public");
                    return sig.access;
                }
            }
        }

        public StringBuilder ReturnType
        {
            get
            {
                if (sig.returnType != null)
                    return sig.returnType;
                else if (isWithin)
                {
                    if (node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE) != null)
                    {
                        sig.SetReturnType(node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE).getText());
                    }
                    else
                    {
                        // Must be a constructor, i.e. has no return type
                        sig.SetReturnType("void");
                    }

                    return sig.returnType;
                }
                else
                {
                    sig.SetReturnType("void");
                    return sig.returnType;
                }
            }
        }


        public StringBuilder DeclaringType
        {
            get
            {
                if (sig.declaringType != null)
                {
                    return sig.declaringType;
                }

                // for the Within joinpointSignature, the declaring type is simply the className.
                if (isWithin)
                {
                    sig.SetDeclaringType(className);
                }
                else
                {
                    if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.CALL)
                    {
                        string declType = null;
                        // ---------------- CASE 1: WE'RE DEALING WITH A CONSTRUCTOR CALL ------------------
                        if (node.Type == Kunle.CSharpParser.CSharpParser.OBJ_CREATE_EXPR)
                        {
                            declType = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText();
                        }
                        // ---------------- CASE 2: WE'RE DEALING WITH A METHOD INVOCATION ------------------
                        else if (node.Type == Kunle.CSharpParser.CSharpParser.EXPR_STMT)
                        {
                            Kunle.CSharpParser.ASTNode invocationExpr = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INVOCATION_EXPR);
                            if (invocationExpr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.MEMBER_ACCESS_EXPR) == null)
                            {
                                // This is a simple call of a method in the class. So the declaring type is simply the name of the class
                                declType = className;
                            }
                            else
                            {
                                Kunle.CSharpParser.ASTNode exprMember = (Kunle.CSharpParser.ASTNode)invocationExpr.
                                                                        GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.MEMBER_ACCESS_EXPR).getFirstChild();

                                // Search for the declaring type in the symbol table, if you can't find it, simply write out the text
                                declType = findType(exprMember, className, DECLARING_TYPE_MODE);
                            }
                        }
                        sig.SetDeclaringType(declType);
                    }
                    else if ((joinpointType == Phx.Aop.Joinpoints.JoinpointType.SET) || (joinpointType == Phx.Aop.Joinpoints.JoinpointType.GET))
                    {
                        string declType = findType((Kunle.CSharpParser.ASTNode)node.getFirstChild(), className, DECLARING_TYPE_MODE);
                        sig.SetDeclaringType(declType);
                    }
                    else
                    {
                        sig.SetDeclaringType("null");			// IS THIS RIGHT?????????
                    }
                }

                return sig.declaringType;
            }
        }



        public StringBuilder MemberName
        {
            get
            {
                if (sig.memberName != null)
                {
                    return sig.memberName;
                }

                else if (isWithin)
                {
                    // The MemberName is the method name or "new" in the case of a constructor
                    // ---------------- CASE 1: WE'RE DEALING WITH A CONSTRUCTOR CALL ------------------
                    if (node.Type == Kunle.CSharpParser.CSharpParser.CTOR_DECL)
                    {
                        sig.SetMemberName("new");
                    }
                    // ---------------- CASE 2: WE'RE DEALING WITH A METHOD INVOCATION ------------------
                    else if (node.Type == Kunle.CSharpParser.CSharpParser.METHOD_DECL)
                    {
                        sig.SetMemberName(node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText());
                    }
                    return sig.memberName;
                }

                else
                {
                    if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.CALL)
                    {
                        // ---------------- CASE 1: WE'RE DEALING WITH A CONSTRUCTOR CALL ------------------
                        if (node.Type == Kunle.CSharpParser.CSharpParser.OBJ_CREATE_EXPR)
                        {
                            sig.SetMemberName("new");
                        }
                        // ---------------- CASE 2: WE'RE DEALING WITH A METHOD INVOCATION ------------------
                        else if (node.Type == Kunle.CSharpParser.CSharpParser.EXPR_STMT)
                        {

                            Kunle.CSharpParser.ASTNode invocationExpr = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INVOCATION_EXPR);
                            if (invocationExpr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.MEMBER_ACCESS_EXPR) == null)
                            {
                                // The member name is simply the identifier child -- just a method in the current class
                                sig.SetMemberName(invocationExpr.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER).getText());
                            }
                            else
                            {
                                Kunle.CSharpParser.ASTNode mem = (Kunle.CSharpParser.ASTNode)node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INVOCATION_EXPR).
                                                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.MEMBER_ACCESS_EXPR).getFirstChild();
                                // Find the last child, i.e. the memberName
                                while (mem.getNextSibling() != null)
                                {
                                    mem = (Kunle.CSharpParser.ASTNode)mem.getNextSibling();
                                }

                                sig.SetMemberName(mem.getText());
                            }
                        }
                    }
                    else if ((joinpointType == Phx.Aop.Joinpoints.JoinpointType.SET) || (joinpointType == Phx.Aop.Joinpoints.JoinpointType.GET))
                    {
                        Kunle.CSharpParser.ASTNode mem = (Kunle.CSharpParser.ASTNode)node.getFirstChild().getNextSibling();

                        // Find the last child, i.e. the memberName
                        while (mem.getNextSibling() != null)
                        {
                            mem = (Kunle.CSharpParser.ASTNode)mem.getNextSibling();
                        }

                        sig.SetMemberName(mem.getText());
                    }
                    else
                    {
                        // IS THIS RIGHT??????? *********
                        sig.SetMemberName("null");
                    }

                    return sig.memberName;
                }
            }
        }



        public IEnumerable<StringBuilder> ArgTypes
        {
            get
            {
                if (sig.argTypes != null)
                {
                    return sig.argTypes;
                }

                else if (isWithin)
                {
                    if (node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.FORMAL_PARAMETER_LIST) != null)
                    {
                        // The method does have parameters
                        Kunle.CSharpParser.ASTNode argList = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.FORMAL_PARAMETER_LIST);

                        if (argList != null)
                        {
                            sig.argTypes = new StringBuilder[argList.NumChildren()];
                            int i = 0;

                            Kunle.CSharpParser.ASTNode arg = argList.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.PARAMETER_FIXED);
                            sig.argTypes[i] = new StringBuilder(typeAsString(arg.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE)));
                            i++;
                            while (arg.getNextSibling() != null)
                            {
                                arg = (Kunle.CSharpParser.ASTNode)arg.getNextSibling();
                                sig.argTypes[i] = new StringBuilder(typeAsString(arg.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE)));
                                i++;
                            }
                        }
                    }

                    return sig.argTypes;
                }

                else
                {
                    if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.CALL)
                    {
                        Kunle.CSharpParser.ASTNode argList = null;

                        // ---------------- CASE 1: WE'RE DEALING WITH A CONSTRUCTOR CALL ------------------
                        if (node.Type == Kunle.CSharpParser.CSharpParser.OBJ_CREATE_EXPR)
                        {
                            argList = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ARG_LIST);
                        }
                        // ---------------- CASE 2: WE'RE DEALING WITH A METHOD INVOCATION ------------------
                        else if (node.Type == Kunle.CSharpParser.CSharpParser.EXPR_STMT)
                        {
                            argList = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INVOCATION_EXPR).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ARG_LIST);
                        }

                        if (argList != null)
                        {
                            string[] args = new string[argList.NumChildren()];
                            int i = 0;

                            Kunle.CSharpParser.ASTNode arg = argList.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR);
                            args[i] = findType((Kunle.CSharpParser.ASTNode)arg.getFirstChild(), className, NOT_DECLARING_TYPE_MODE);
                            i++;
                            while (arg.getNextSibling() != null)
                            {
                                arg = (Kunle.CSharpParser.ASTNode)arg.getNextSibling();
                                args[i] = findType((Kunle.CSharpParser.ASTNode)arg.getFirstChild(), className, NOT_DECLARING_TYPE_MODE);
                                i++;
                            }

                            // Store the args in your SimpleSignature "sig" variable
                            sig.argTypes = new StringBuilder[args.Length];
                            for (int j = 0; j < args.Length; j++)
                                sig.argTypes[j] = new StringBuilder(args[j]);
                        }
                    }
                    else if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.SET)
                    {
                        string declType = findType((Kunle.CSharpParser.ASTNode)node.getFirstChild(), className, NOT_DECLARING_TYPE_MODE);

                        // Store the args in your SimpleSignature "sig" variable
                        sig.argTypes = new StringBuilder[1];
                        sig.argTypes[0] = new StringBuilder(declType);
                    }

                    return sig.argTypes;
                }
            }

        }

        public IEnumerable<StringBuilder> ArgNames
        {
            get
            {
                if (sig.argNames != null)
                {
                    return sig.argNames;
                }

                else if (isWithin)
                {
                    if (node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.FORMAL_PARAMETER_LIST) != null)
                    {
                        // The method does have parameters
                        Kunle.CSharpParser.ASTNode argList = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.FORMAL_PARAMETER_LIST);

                        if (argList != null)
                        {
                            sig.argNames = new StringBuilder[argList.NumChildren()];
                            int i = 0;

                            Kunle.CSharpParser.ASTNode arg = argList.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.PARAMETER_FIXED);
                            sig.argNames[i] = new StringBuilder(arg.getText());
                            i++;
                            while (arg.getNextSibling() != null)
                            {
                                arg = (Kunle.CSharpParser.ASTNode)arg.getNextSibling();
                                sig.argNames[i] = new StringBuilder(arg.getText());
                                i++;
                            }
                        }
                    }

                    return sig.argNames;
                }

                else
                {
                    if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.CALL)
                    {
                        Kunle.CSharpParser.ASTNode argList = null;

                        // ---------------- CASE 1: WE'RE DEALING WITH A CONSTRUCTOR CALL ------------------
                        if (node.Type == Kunle.CSharpParser.CSharpParser.OBJ_CREATE_EXPR)
                        {
                            argList = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ARG_LIST);
                        }
                        // ---------------- CASE 2: WE'RE DEALING WITH A METHOD INVOCATION ------------------
                        else if (node.Type == Kunle.CSharpParser.CSharpParser.EXPR_STMT)
                        {
                            argList = node.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.INVOCATION_EXPR).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ARG_LIST);
                        }

                        if (argList != null)
                        {
                            string[] args = new string[argList.NumChildren()];
                            int i = 0;

                            Kunle.CSharpParser.ASTNode arg = argList.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR);
                            args[i] = arg.getFirstChild().getText();
                            i++;
                            while (arg.getNextSibling() != null)
                            {
                                arg = (Kunle.CSharpParser.ASTNode)arg.getNextSibling();
                                args[i] = arg.getFirstChild().getText();
                                i++;
                            }

                            // Store the args in your SimpleSignature "sig" variable
                            sig.argNames = new StringBuilder[args.Length];
                            for (int j = 0; j < args.Length; j++)
                                sig.argNames[j] = new StringBuilder(args[j]);
                        }
                    }
                    else if (joinpointType == Phx.Aop.Joinpoints.JoinpointType.SET)
                    {
                        string declType = findType((Kunle.CSharpParser.ASTNode)node.getFirstChild(), className, NOT_DECLARING_TYPE_MODE);

                        // Store the args in your SimpleSignature "sig" variable
                        sig.argNames = new StringBuilder[1];
                        sig.argNames[0] = new StringBuilder(declType);
                    }

                    return sig.argNames;
                }
            }
        }


        public IEnumerator Args
        {
            get { return null; } // Not implemented
        }


        public Phx.Aop.Joinpoints.ISourceLocation SourceLocation
        {
            get
            {
                if (location == null && !isWithin)
                {
                    location = new WcsSourceLocation();
                    location.node = node;
                }

                return location;
            }
        }



        public override string ToString()
        {
            return (DeclaringType + "." + MemberName);
        }


        public string debugToString()
        {
            // ******************** Detailed ToString Only for debugging **********************            
            string sigDescription = DeclaringType + "." + MemberName;
            if (ArgTypes != null)
            {
                sigDescription += " with arg type(s) ";
                if (ArgTypes != null)
                {
                    bool first = true;
                    foreach (StringBuilder argType in ArgTypes)
                    {
                        if (!first)
                            sigDescription += ", ";

                        sigDescription += argType;
                        first = false;
                    }
                }
                else
                {
                    sigDescription += "(null)";
                }
            }

            // Need to make sure length is zero, because the method can initialize its attributes, but
            // then realize that all of them are statement attributes, in which case the size simply remains 0
            if (AttributeTypes != null)
            {
                sigDescription += "\n\tand with attribute(s) ";
                bool first = true;

                foreach (string attrObj in AttributeTypes)
                {
                    if (!first)
                        sigDescription += ", ";

                    sigDescription += "[" + attrObj + "]";
                    first = false;
                }

            }

            return sigDescription;

        }





        // ======================================================================
        // 						HELPER FUNCTIONS
        // ======================================================================

        private bool found = true;				// Used by the findType function to indicate whether the current var
        private string findType(Kunle.CSharpParser.ASTNode n, string className, int mode)
        {
            string type = null, varName = null;
            varName = n.getText();
            found = true;							// Reset each time

            if (n.Type == Kunle.CSharpParser.CSharpParser.STRING_LITERAL)
            {
                return ("string");
            }
            else if (n.Type == Kunle.CSharpParser.CSharpParser.INT_LITERAL)
            {
                return ("int");
            }
            // This signifies a cast
            else if (n.Type == Kunle.CSharpParser.CSharpParser.PAREN_EXPR &&
           n.getFirstChild().Type == Kunle.CSharpParser.CSharpParser.CAST_EXPR)
            {
                return ( typeAsString(n.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.CAST_EXPR).
                    GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE)) );
            }
            else if (n.Type == Kunle.CSharpParser.CSharpParser.MEMBER_ACCESS_EXPR)
            {
                // This is a member access expression
                n = (Kunle.CSharpParser.ASTNode)n.getFirstChild();
                type = findType(n, className, mode);
                if (found)
                    type = findType((Kunle.CSharpParser.ASTNode)n.getNextSibling(), type, mode);
                else
                    type += "." + findType((Kunle.CSharpParser.ASTNode)n.getNextSibling(), type, mode);
                return (type);
            }
            else if (n.Type == Kunle.CSharpParser.CSharpParser.INVOCATION_EXPR
                || n.Type == Kunle.CSharpParser.CSharpParser.DOT)
            {
                // This is an invocation expression, i.e. a method call
                type = findType((Kunle.CSharpParser.ASTNode)n.getFirstChild(), className, mode);
                return (type);
            }
            else if (n.Type == Kunle.CSharpParser.CSharpParser.ELEMENT_ACCESS_EXPR)
            {
                // This is an array access
                type = findType(n.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.IDENTIFIER), className, mode);
                return (type);
            }
            else if (n.Type == Kunle.CSharpParser.CSharpParser.PLUS
                || n.Type == Kunle.CSharpParser.CSharpParser.MINUS
                || n.Type == Kunle.CSharpParser.CSharpParser.DIV
                || n.Type == Kunle.CSharpParser.CSharpParser.STAR
                || n.Type == Kunle.CSharpParser.CSharpParser.MOD
                || n.Type == Kunle.CSharpParser.CSharpParser.POST_INC_EXPR
                || n.Type == Kunle.CSharpParser.CSharpParser.POST_DEC_EXPR
                || n.Type == Kunle.CSharpParser.CSharpParser.INC
                || n.Type == Kunle.CSharpParser.CSharpParser.DEC)
            {
                // This is a mathematical expression
                type = findType((Kunle.CSharpParser.ASTNode)n.getFirstChild(), className, mode);
                if (!found)
                    type = findType((Kunle.CSharpParser.ASTNode)n.getFirstChild().getNextSibling(), className, mode);
                return (type);
            }
            else if (n.Type == Kunle.CSharpParser.CSharpParser.CAST_EXPR)
            {
                // There is a cast in front of this expression. Return that as the type
                type = typeAsString(n.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.TYPE));
                return (type);
            }


            // Determine the type of the variable stored in 'varName'
            IDictionaryEnumerator Enumerator = symbolTable.GetEnumerator();
            while (Enumerator.MoveNext())
            {
                if (Enumerator.Key.Equals(varName))
                    type = (string)Enumerator.Value;
            }

            // Found the type in the current symbol table, now if there are any more members to traverse,
            // keep looking through the class tables as needed.
            if (type != null)
            {
                // If the mode is DECLARING_TYPE_MODE, only go as far as the second-to-last member, 
                // because the last member is saved as the MemberName
                if (mode == DECLARING_TYPE_MODE)
                {
                    while ((n.getNextSibling() != null) && (n.getNextSibling().getNextSibling() != null))
                    {
                        n = (Kunle.CSharpParser.ASTNode)n.getNextSibling();
                        type = tablesList.findType(type, n.getText());
                    }
                }
                else
                {
                    while (n.getNextSibling() != null)
                    {
                        n = (Kunle.CSharpParser.ASTNode)n.getNextSibling();
                        type = tablesList.findType(type, n.getText());
                    }
                }
            }


            if (type == null)
            {
                found = false;
                // You didn't find it in the current local table, or any other symbol tables, so just spell out the var names 
                // (it's most likely an object from a library like "Console.WriteLine")
                type = n.getText();
                // If the mode is DECLARING_TYPE_MODE, only go as far as the second-to-last member, 
                // because the last member is saved as the MemberName
                if (mode == DECLARING_TYPE_MODE)
                {
                    while ((n.getNextSibling() != null) && (n.getNextSibling().getNextSibling() != null))
                    {
                        n = (Kunle.CSharpParser.ASTNode)n.getNextSibling();
                        type += "." + n.getText();
                    }
                }
                else
                {
                    while (n.getNextSibling() != null)
                    {
                        n = (Kunle.CSharpParser.ASTNode)n.getNextSibling();
                        type += "." + n.getText();
                    }
                }

                return (type);
            }

            return (type);

        }	// End findType()



        // Returns the type as specified by the "typeof" node given as a parameter in string form
        private string typeAsString(Kunle.CSharpParser.ASTNode t)
        {
            string result = null;

            Kunle.CSharpParser.ASTNode Child = (Kunle.CSharpParser.ASTNode)t.getFirstChild();

            if (Child.Type == Kunle.CSharpParser.CSharpParser.DOT)
            {
                while (Child.Type == Kunle.CSharpParser.CSharpParser.DOT)
                {
                    result += Child.getFirstChild().getText() + ".";
                    Child = (Kunle.CSharpParser.ASTNode)Child.getFirstChild().getNextSibling();
                }
                result += Child.getText();
            }
            else
            {
                result = t.getFirstChild().getText();
            }

            // Check if arg type has * or [] after the type
            if ((Child = (Kunle.CSharpParser.ASTNode)t.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.STARS).getFirstChild()) != null)
            {
                while (Child != null)
                {
                    result += "*";
                    Child = (Kunle.CSharpParser.ASTNode)Child.getNextSibling();
                }
            }
            if ((Child = (Kunle.CSharpParser.ASTNode)t.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ARRAY_RANKS).getFirstChild()) != null)
            {
                while (Child != null)
                {
                    result += "[]";
                    Child = (Kunle.CSharpParser.ASTNode)Child.getNextSibling();
                }
            }

            return (result);
        }






        // This function returns as a string the args of the attributes
        // if they match the type given as a parameter.
        // Used to get the attribute args to insert as parameters into the
        // advice method applied to the joinpoint
        public List<Kunle.CSharpParser.ASTNode> getAttributeArgs(int attrIndex)
        {
            List<Kunle.CSharpParser.ASTNode> result = new List<Kunle.CSharpParser.ASTNode>();
            IEnumerator enumAttrs;
            if (Attributes != null)
                enumAttrs = Attributes;
            else
                return (null);

            if (!isWithin)
            {
                int index = 0;
                while (enumAttrs.MoveNext())
                {
                    if (index != attrIndex)
                        index++;
                    else
                    {
                        // Navigate to first arg that will be considered an arg for the advice method, i.e. the third one
                        Kunle.CSharpParser.ASTNode args = (Kunle.CSharpParser.ASTNode)((Kunle.CSharpParser.ASTNode)enumAttrs.Current).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARGLIST).
                            GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARG).getNextSibling().getNextSibling();

                        while (args != null)
                        {
                            result.Add(args.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIB_ARGUMENT_EXPR).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR));
                            args = (Kunle.CSharpParser.ASTNode)args.getNextSibling();
                        }

                        return (result);
                    }
                }
            }
            else if (isWithin)
            {
                int index = 0;

                while (enumAttrs.MoveNext())
                {
                    if (index != attrIndex)
                        index++;
                    else
                    {
                        // Navigate to first arg that will be considered an arg for the advice method, i.e. the fourth one
                        if (((Kunle.CSharpParser.ASTNode)enumAttrs.Current).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARGLIST) != null)
                        {
                            Kunle.CSharpParser.ASTNode args = (Kunle.CSharpParser.ASTNode)((Kunle.CSharpParser.ASTNode)enumAttrs.Current).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARGLIST).
                                GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.POSITIONAL_ARG);

                            while (args != null)
                            {
                                // Enclose individual ars in quotations
                                result.Add(args.GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.ATTRIB_ARGUMENT_EXPR).GetFirstChildOfType(Kunle.CSharpParser.CSharpParser.EXPR));
                                args = (Kunle.CSharpParser.ASTNode)args.getNextSibling();
                            }

                            return (result);
                        }

                        return (null);
                    }
                }
            }

            return (null);
        }


        // Returns the type of the attribute at index attrIndex
        public string getAttributeType(int attrIndex)
        {
            string result = "";
            IEnumerator enumAttrs;
            if (AttributeTypes != null)
                enumAttrs = AttributeTypes.GetEnumerator();
            else
                return "";

            if (!isWithin)
            {
                int index = 0;
                while (enumAttrs.MoveNext())
                {
                    if (index != attrIndex)
                        index++;
                    else
                    {
                        result = (string)enumAttrs.Current;
                        return (result);
                    }
                }
            }
            else if (isWithin)
            {
                int index = 0;

                while (enumAttrs.MoveNext())
                {
                    if (index != attrIndex)
                        index++;
                    else
                    {
                        result = (string)enumAttrs.Current;
                        return (result);
                    }
                }
            }

            return "";
        }

    }

}
