//-----------------------------------------------------------------------------
//
// Copyright (C) Boriana Ditcheva.  All Rights Reserved.
//
// $Header: /Wicca/Wicca.wcs/WcsSymbolTable.cs 2     5/19/06 12:19a Bhd2105 $
//
//	Description: this class is a hashtable of hashtables. Each hashtable entry in the big hashtable
//				represents the symboltable of one classfile that is being parsed.
//				The hashtables are stored with their respective class names as the index.
//				Each symboltable entry has the form <varName><type>
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Wicca.wcs/WcsSymbolTable.cs $
 * 
 * 2     5/19/06 12:19a Bhd2105
 *  Now supports the addition of methodDeclarations in a class's symbol
 * table. The function findMethodDecl() has been added to search for
 * method declarations in a given class's symbol table. It returns the
 * WcsJoinpointSignature of the method declaration which is stored as the
 * Value in the symbol table.
 */


using System;
using System.IO;
using System.Collections;				// For the hashtable used for a symbol table




namespace Wicca.wcs
{	

	public class TablesList
	{
		// 	*************** 	VARIABLES		**************************
		// table of tables for all classes being parsed name
		Hashtable tables;


		// Constructor
		public TablesList()
		{
			tables = new Hashtable();
		}
		
		public void addSymbolTable(string tableName, Hashtable table)
		{
			tables.Add(tableName, table);
		}

		public Hashtable findSymbolTable(string className)
		{
			// Search for the symbol table of the given class
			IDictionaryEnumerator Enumerator = tables.GetEnumerator();
			while (Enumerator.MoveNext())
			{
				if(Enumerator.Key.Equals(className))
					return( (Hashtable)Enumerator.Value );
			}
			
			return null;
		}


		// Returns the type of the variable <varName> from the symbol table of class <className>
		public string findType(string className, string varName)
		{
			IDictionaryEnumerator Enumerator = tables.GetEnumerator();
			while (Enumerator.MoveNext())
			{
				// You've found the symbol table for the given class
				if(Enumerator.Key.Equals(className))
				{
					IDictionaryEnumerator subEnum = ((Hashtable)Enumerator.Value).GetEnumerator();
					while (subEnum.MoveNext())
					{
						// You've found the needed variable in the symbol table of the given class
						// Now return its type
						if(subEnum.Key.Equals(varName))
							return((string)subEnum.Value);
					}
				}
			}
			
			return null;
		}


        // Returns the wcsJoinpoint of the method whose methodDeclaration <methodDecl> 
        // is stored in the symbol table of class <className>
        public Wcs.Aop.Joinpoints.WcsJoinpointSignature findMethodDecl(string className, string methodName)
        {
            IDictionaryEnumerator Enumerator = tables.GetEnumerator();
            while (Enumerator.MoveNext())
            {
                // You've found the symbol table for the given class
                if (Enumerator.Key.Equals(className))
                {
                    IDictionaryEnumerator subEnum = ((Hashtable)Enumerator.Value).GetEnumerator();
                    while (subEnum.MoveNext())
                    {
                        // You've found the needed variable in the symbol table of the given class
                        // Now return its type
                        if (subEnum.Key.Equals(methodName))
                            return ((Wcs.Aop.Joinpoints.WcsJoinpointSignature)subEnum.Value);
                    }
                }
            }

            return (null);
        }



		// For debugging purposes: print out all variables in all symbol tables
		public void printTables()
		{
			IDictionaryEnumerator Enumerator = tables.GetEnumerator();
			while (Enumerator.MoveNext())
			{
				// Print out current table name and then traverse the variables of that table
				Console.WriteLine("----------------------- TABLE: " + Enumerator.Key + "------------------------");
				IDictionaryEnumerator subEnum = ((Hashtable)Enumerator.Value).GetEnumerator();
				while (subEnum.MoveNext())
					Console.WriteLine("contains " + subEnum.Value + " " + subEnum.Key);
			}
		}

	}
	
	
}

