//-----------------------------------------------------------------------------
//
// Copyright (C) Boriana Ditcheva.  All Rights Reserved.
//
// $Header: /Wicca/Wicca.wcs/wcs.cs 10    5/19/06 12:13a Bhd2105 $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/Wicca.wcs/wcs.cs $
 * 
 * 10    5/19/06 12:13a Bhd2105
 * 
 * 9     5/10/06 11:43a Eaddy
 * o Fixed compile errors (now uses IEnumerator and IEnumerable ifaces)
 * 
 * 8     5/08/06 12:57a Bhd2105
 * wcs.cs now has the "weave" flag framework. It also now only parses .cs
 * files and converts them to ".converted.cs" files when they are
 * transformed to have the attributes lifted to method level. When weaving
 * is done, a new file with a ".woven.cs" extension is created to save
 * that version of the file.
 * 
 * 7     5/04/06 2:40p Bhd2105
 * 
 * 6     5/03/06 1:44a Bhd2105
 * Now maintains a list of hashtables for all the classes it is currently
 * parsing. It calls ParseAndWeave() with different modes depending on
 * whether it's populating the tables or parsing the files for
 * modification and output.
 * 
 */


// bug(?) in DotGNU 0.6 - "using antlr" will workaround the problem.
#if __CSCC__
using antlr;
#endif


using System;
using System.Text;
using System.IO;
using System.CodeDom.Compiler;
using System.Collections;				// For the hashtable used for a symbol table
using System.Collections.Generic;		// for the List of advices



namespace Wicca.wcs
{
    using BaseAST = antlr.BaseAST;
    using CommonAST = antlr.CommonAST;
    using ASTFactory = antlr.ASTFactory;
    using RecognitionException = antlr.RecognitionException;
    using AST = antlr.collections.AST;
    using TokenStream = antlr.TokenStream;
    using TokenStreamSelector = antlr.TokenStreamSelector;
    using TokenStreamHiddenTokenFilter = antlr.TokenStreamHiddenTokenFilter;



    public class WcsProvider : CodeDomProvider
    {

        // ************* VARIABLES ******************
        private CodeDomProvider myProvider = null;
        private static TablesList myTables = new TablesList();
        private static List<Wcs.Aop.WcsAdvice> advices = new List<Wcs.Aop.WcsAdvice>();        // Contains all the advice you'll be matching on
        const int POPULATE_TABLE_AND_FIND_ADVICE = 0;
        const int OUTPUT_FILE = 1;
        private static bool weave;



        public WcsProvider(bool weave_in)
        {
            myProvider = new Microsoft.CSharp.CSharpCodeProvider();
            weave = weave_in;
        }

        [ObsoleteAttribute("Callers should not use the ICodeCompiler interface and should instead use the methods directly on the CodeDomProvider class. Those inheriting from CodeDomProvider must still implement this interface, and should exclude this warning or also obsolete this method.")]
        public override ICodeCompiler CreateCompiler()
        {
            return null;
        }

        [ObsoleteAttribute("Callers should not use the ICodeGenerator interface and should instead use the methods directly on the CodeDomProvider class. Those inheriting from CodeDomProvider must still implement this interface, and should exclude this warning or also obsolete this method.")]
        public override ICodeGenerator CreateGenerator()
        {
            return null;
        }


        // ==========================================================================
        //					COMPILE THE CONTENTS OF THE GIVEN FILES
        // ==========================================================================		
        public override CompilerResults CompileAssemblyFromFile(CompilerParameters options, string[] fileNames)
        {
            string[] newFileNames = new string[fileNames.Length];
            string basename = null;
            string fileOut = null;
            int end = 0;


            // For each file in fileNames[] parse the file, and if it's a .wcs file, make it a legal c# file, 
            // change the extension of the new file from .wcs to .cs, and give the new .cs names to the c# parser to 
            // parse. Otherwise, the file is a regular .cs, so just pass it on to the parser as is.

            // In this run through the files, parse them and populate the symbol tables from all the files
            for (int i = 0; i < fileNames.Length; i++)
            {
                ParseAndWeave(fileNames[i], POPULATE_TABLE_AND_FIND_ADVICE);
            }


            // Now tree walk the files to output the new .cs files (with the legalized c code and weaving if necessary)
            for (int i = 0; i < fileNames.Length; i++)
            {
                ParseAndWeave(fileNames[i], OUTPUT_FILE);
            }

            for (int i = 0; i < fileNames.Length; i++)
            {
                end = fileNames[i].LastIndexOf(".cs");
                if (end > 0)
                {
                    basename = fileNames[i].Substring(0, end);
                    //					if (weave)
                    //						fileOut = basename + ".woven.cs";
                    //					else
                    fileOut = basename + ".converted.cs";
                    newFileNames[i] = fileOut;
                }
                else
                {
                    // Don't process a file that's not a .cs file
                    //newFileNames[i] = fileNames[i];
                }
            }

            // Print out the variables you've gathered in your file-specific symbol tables
            //myTables.printTables();

            return (myProvider.CompileAssemblyFromFile(options, newFileNames));
        }



        // ===============================================================
        //						HELPER FUNCTIONS
        // ===============================================================
        public static void ParseAndWeave(string fileName, int mode)
        {
            FileInfo f = new FileInfo(fileName);
            parseFile(f.Name, new FileStream(f.FullName, FileMode.Open, FileAccess.Read), mode);
        }



        // Here's where we do the real work...
        // ==========================================================================
        //				THIS FUNCTION TAKES IN THE NAME OF A FILE TO BE PARSED
        //				AND STORES THE RESULTS OF THE PARSED FILE IN A NEW FILE
        //				WITH A ".cs" EXTENSION
        // ==========================================================================		
        public static void parseFile(string f, Stream s, int mode)
        {
            try
            {
                // Define a selector that can switch from the C# codelexer to the C# preprocessor lexer
                TokenStreamSelector selector = new TokenStreamSelector();

                TokenStream lexer;
                // Create a scanner that reads from the input stream passed to us
                Kunle.CSharpParser.CSharpLexer antlrLexer = new Kunle.CSharpParser.CSharpLexer(new StreamReader(s));

                antlrLexer.Selector = selector;
                antlrLexer.setFilename(f);

                // use the special token object class
                antlrLexer.setTokenCreator(new Kunle.CSharpParser.CustomHiddenStreamToken.CustomHiddenStreamTokenCreator());
                antlrLexer.setTabSize(1);
                selector.addInputStream(antlrLexer, "codeLexer");
                selector.select("codeLexer"); // start with main the CSharp code lexer
                lexer = selector;

                // create the stream filter; hide WS and SL_COMMENT
                TokenStreamHiddenTokenFilter filter = new TokenStreamHiddenTokenFilter(lexer);

                // Changed by Boriana
                filter.hide(Kunle.CSharpParser.CSharpTokenTypes.WHITESPACE);
                filter.hide(Kunle.CSharpParser.CSharpTokenTypes.NEWLINE);
                filter.hide(Kunle.CSharpParser.CSharpTokenTypes.ML_COMMENT);
                filter.hide(Kunle.CSharpParser.CSharpTokenTypes.SL_COMMENT);

                // Create a parser that reads from the scanner
                Kunle.CSharpParser.CSharpParser parser = new Kunle.CSharpParser.CSharpParser(filter);

                // create trees that copy hidden tokens into tree also
                parser.setASTNodeClass(typeof(Kunle.CSharpParser.ASTNode).FullName);
                parser.setASTFactory(new Kunle.CSharpParser.ASTNodeFactory());
                Kunle.CSharpParser.CSharpParser.initializeASTFactory(parser.getASTFactory());
                parser.setFilename(f);

                // ===========================================================================
                // 		Do two parses of the file:
                //	At the 1st parse, determine the aspects (which have the keyword "Advice") and store
                //	them in objects. At the 2nd parse, weave in the aspects you've gathered at the
                //	join points where they are applicable.
                // ===========================================================================
                // start parsing at the compilationUnit rule
                parser.compilationUnit();

                // Walk the tree
                doTreeAction(f, (Kunle.CSharpParser.ASTNode)parser.getAST(), parser.getTokenNames(), mode);

            }
            catch (System.Exception e)
            {
                Console.Error.WriteLine("parser exception: " + e);
                Console.Error.WriteLine(e.StackTrace); 			// so we can get stack trace		
            }

        }

        public static void doTreeAction(string f, Kunle.CSharpParser.ASTNode t, string[] tokenNames, int mode)
        {
            // *********************** VARIABLES *********************
            StringWriter writer = new StringWriter();
            Kunle.CSharpParser.CSharpPrettyPrinter printer = new Kunle.CSharpParser.CSharpPrettyPrinter();
            printer.setASTFactory(new Kunle.CSharpParser.ASTNodeFactory());
            Kunle.CSharpParser.CSharpParser.initializeASTFactory(printer.getASTFactory());
            string basename;
            //			bool weave = false;		

            if (t == null)
                return;

            if (mode == POPULATE_TABLE_AND_FIND_ADVICE)
            {
                basename = f.Substring(0, f.LastIndexOf("."));

                printer.PopulateSymbolTable_GetAdvice(writer, t, myTables, advices);
            }
            else if (mode == OUTPUT_FILE)
            {

                // Only parse files if 
                //	(1) they have the .wcs extension, in which circumstance you parse the file,
                // (thus legalizing it), and save it with a .cs extension.
                //	(2) the weave flag is on, in which case you weave in the advice methods and
                // save the file with a .woven.cs extension.
                string fileOut;
                int end = f.LastIndexOf(".cs");
                if (end > 0)
                {
                    basename = f.Substring(0, end);
                    fileOut = basename + ".converted.cs";

                    // Temporarily set 'weave' to 'true' to get wcs.cs to compile
                    printer.PrintFile(writer, t, string.Empty, myTables, advices, true);
                    StreamWriter w = File.CreateText(fileOut);
                    w.Write(writer.ToString());
                    w.Close();
                }

            }	// End else-if OUTPUT-FILE
            if (weave)
            {
                // IMPLEMENT
            }
        }

    }

}
