//-----------------------------------------------------------------------------
//
// Copyright (C) Boriana Ditcheva.  All Rights Reserved.
//
// $Header: /Wicca/wcs/wcs.cs 4     5/14/06 4:35p Bhd2105 $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/wcs/wcs.cs $
 * 
 * 4     5/14/06 4:35p Bhd2105
 * Now sets the OutputAssembly compiler param right before compiling the
 * file if the user has specified an /out: option
 * 
 * 3     5/08/06 12:50a Bhd2105
 * Modified to print out error message and exit when user tries to use
 * "/recurse" option. Also checks for and passes on "weave" flag
 * 
 * 2     4/27/06 3:26p Bhd2105
 */
using System;
using System.IO;
using System.CodeDom.Compiler;



namespace Wicca.wcs
{
    public delegate string OnCompilerOptionFound(string options, string optionName, int start, int length);

    public class WcsWeaveCompiler
    {
        // 	*************** 	VARIABLES		**************************
        private WcsProvider provider = null;
        private CompilerParameters _compilerParams = new CompilerParameters();
        private string[] _arySourceFiles;
        private static bool weave = true;								// weave by default
        private static string outputFile = null;




        public static void Main(string[] args)
        {
            // *************** Variables ****************
            string compilerOptions = "";
            string[] sources = new string[args.Length];
            int numSources = 0;
            bool userSetOptimization = false;

            // Get compiler options and source files from args and parse
            if (args.Length > 0)
            {
                // for each directory/file specified on the command line
                for (int i = 0; i < args.Length; i++)
                {
                    if (args[i].StartsWith("/out"))
                    {
                        outputFile = args[i].Substring(5);
                    }
                    else
                        compilerOptions += args[i] + " ";
                    if (!args[i].StartsWith("/"))
                    {
                        // Must be a source file, add it to sources
                        sources[numSources] = args[i];
                        numSources++;		// increment index
                    }
                    else if (args[i].StartsWith("/recurse"))
                    {
                        Console.WriteLine();
                        Console.WriteLine("\tWcs does not support the /recurse option. Compilation cannot be completed.");
                        Console.WriteLine();
                        System.Environment.Exit(0);
                    }
                    else if (args[i].StartsWith("/optimize"))
                    {
                        userSetOptimization = true;
                    }
                    else if (args[i].ToLower().Equals("/weave:no") || args[i].ToLower().Equals("/weave:n"))
                    {
                        weave = false;
                    }
                }

                if (!userSetOptimization)
                {
                    // Add as an option that the compiler shouldn't optimize by default
                    compilerOptions += "/optimize-";
                }
            }

            WcsWeaveCompiler myCompiler = new WcsWeaveCompiler(compilerOptions, sources, numSources);

            // Prune some options
            myCompiler.findCompilerOption("/out:", new OnCompilerOptionFound(myCompiler.removeCompilerOption));
            myCompiler.findCompilerOption("/r:", new OnCompilerOptionFound(myCompiler.extractReference));
            myCompiler.findCompilerOption("/reference:", new OnCompilerOptionFound(myCompiler.extractReference));
            myCompiler.findSources(new OnCompilerOptionFound(myCompiler.removeCompilerOption));
            // Remove extra whitespace
            myCompiler._compilerParams.CompilerOptions = myCompiler._compilerParams.CompilerOptions.Trim();

            myCompiler.createWcsCompiler();
            myCompiler.compile();
        }







        // Constructor
        public WcsWeaveCompiler(string compilerOptions, string[] sources, int numSources)
        {
            // Initialize compiler options
            //			_compilerParams.IncludeDebugInformation = debugBuild;
            _compilerParams.CompilerOptions = compilerOptions;
            _compilerParams.GenerateInMemory = false;
            _compilerParams.GenerateExecutable = true;
            _compilerParams.OutputAssembly = "";
            _arySourceFiles = new string[numSources];
            for (int i = 0; i < numSources; i++)
                _arySourceFiles[i] = sources[i];
        }

        /// Searches for a compiler option.
        public void findCompilerOption(string strOptionName,
                                            OnCompilerOptionFound onCompilerOptionFound)
        {
            int nFind = _compilerParams.CompilerOptions.IndexOf(strOptionName);
            while (nFind > -1 && nFind + 1 < _compilerParams.CompilerOptions.Length)
            {
                int nEnd = _compilerParams.CompilerOptions.IndexOf(" ", nFind + 1);
                if (nEnd == -1)
                {
                    nEnd = _compilerParams.CompilerOptions.Length;
                }
                /*                else
                                {
                                    // You've found an "/out" option, set the appropriate parameters to it
                                    // The 5 I'm using is from the length of "/out:"
                                    _compilerParams.OutputAssembly = _compilerParams.CompilerOptions.Substring(nFind + 5, nEnd - nFind - 5);
                                }
                */
                _compilerParams.CompilerOptions =
                    onCompilerOptionFound(_compilerParams.CompilerOptions,
                                            strOptionName, nFind, nEnd - nFind);

                nFind = _compilerParams.CompilerOptions.IndexOf(strOptionName, nFind);
            }
        }

        /// Looks for source files in the compiler option string.
        public void findSources(OnCompilerOptionFound onCompilerOptionFound)
        {
            int nFind = 0;
            while (nFind > -1 && nFind + 1 < _compilerParams.CompilerOptions.Length)
            {
                int nEnd = _compilerParams.CompilerOptions.IndexOf(" ", nFind + 1);
                if (nEnd == -1)
                {
                    nEnd = _compilerParams.CompilerOptions.Length;
                }

                // Only compiler flags start with '/'. The rest must be source files.
                if (_compilerParams.CompilerOptions[nFind + 1] != '/')
                {
                    _compilerParams.CompilerOptions =
                        onCompilerOptionFound(_compilerParams.CompilerOptions,
                        "", nFind, nEnd - nFind);
                }
                else
                {
                    ++nFind; // Skip over the space
                }

                nFind = _compilerParams.CompilerOptions.IndexOf(" ", nFind);
            }
        }

        /// Implements the OnCompilerOptionFound delegate. Removes the found compiler option.
        public string removeCompilerOption(string options, string optionName, int start, int length)
        {
            if (start - 1 >= 0 &&
                options[start - 1] == ' ')
            {
                --start;
                ++length;
            }

            return options.Remove(start, length);
        }

        /// Implements the OnCompilerOptionFound delegate. Extracts the assembly reference option
        /// and adds it to the list of ReferencedAssemblies.
        public string extractReference(string options, string optionName, int start, int length)
        {
            string assembly = options.Substring(start + optionName.Length,
                                                    length - optionName.Length);
            _compilerParams.ReferencedAssemblies.Add(assembly);

            return removeCompilerOption(options, optionName, start, length);
        }


        // Create a wcs provider and compiler
        public void createWcsCompiler()
        {
            provider = new WcsProvider(weave);
        }


        public void compile()
        {
            CompilerResults results = null;
            try
            {
                if (outputFile != null)
                    _compilerParams.OutputAssembly = outputFile;
                results = provider.CompileAssemblyFromFile(_compilerParams, _arySourceFiles);
            }
            catch (IOException ex)
            {
                Console.WriteLine("Compiler had problems accessing file: " + ex);
            }
            catch (Exception ex)
            {
                Console.WriteLine("Compiler threw an exception: " + ex);
            }


            if (results == null)
            {
                Console.WriteLine("Compile failed.");
            }
            else if (results.Errors.Count > 0)
            {
                foreach (CompilerError error in results.Errors)
                {
                    Console.WriteLine(error.ToString());
                }
            }
            else if (results.Output.Count > 0)
            {
                foreach (string strOutput in results.Output)
                {
                    Console.WriteLine(strOutput.ToString());
                }
            }
            else
            {
                Console.WriteLine("Compile succeeded.");
            }

        }
    }



}