//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/wicca/Coordinator.cs 21    5/24/06 9:42p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/wicca/Coordinator.cs $
 * 
 * 21    5/24/06 9:42p Eaddy
 * o Exceptions are errors
 * 
 * 20    5/24/06 2:02p Eaddy
 * o Added Equals() and GetHashCode() so that we can
 * put breakpoint locations in a map
 * 
 * 19    5/24/06 1:53a Eaddy
 * o Breakpoint weaver now supports [Arg]
 * 
 * 18    5/24/06 12:01a Eaddy
 * o Added comments
 * 
 * 17    5/23/06 11:30p Eaddy
 * o Now runs the process to completion so you don't
 * have to keep typing wgo
 * 
 * 16    5/23/06 9:00p Eaddy
 * o More cleanup
 * 
 * 15    5/23/06 7:49p Eaddy
 * o More cleanup
 * 
 * 14    5/23/06 7:41p Eaddy
 * o Cleaning up bpw code
 * 
 * 13    5/12/06 11:22a Asv2105
 * Added checks to prevent errors if the advice method causes the process
 * to die
 * 
 * 12    5/12/06 1:21a Eaddy
 * o Now verifies breakpoint info before setting the breakpoint
 * 
 * 11    5/11/06 5:54p Eaddy
 * o Added exception handlers for exceptions that
 * occur when we call advice methods
 * o Added error handling to advice method dispatch
 * 
 * 10    5/05/06 6:18p Eaddy
 * o Did some renaming to improve clarity
 * o Can now list current advice and enable/disable (although
 * enable/disable doesn't actually work yet)
 * 
 * 9     5/05/06 3:33p Eaddy
 * o Now automatically remaps
 * 
 * 8     5/03/06 8:13p Eaddy
 * o Trivial
 * 
 * 7     5/03/06 12:33a Eaddy
 * o Moved weaving and patching into DynamicWeaver
 * 
 * 6     5/01/06 4:54p Eaddy
 * o Now handles relative paths
 * o Now handles unknown modules (i.e., those generated by System.Xml
 * at runtime)
 * o Fixed diagnostic message
 * 
 * 5     4/28/06 8:01p Eaddy
 * o Trivial change
 * 
 * 4     4/28/06 5:59p Asv2105
 * Added ability to pass integers to process.
 * Changed interface to DebugJoinpoint, BreakpointInfo.
 * Added new AddBreakpoint() method that takes a BreakpointInfo.
 * 
 * 3     4/22/06 12:36p Eaddy
 * o Adding support for using the Debugger API to query for the current
 * metadata and IL
 * 
 * 2     4/21/06 12:34p Eaddy
 * o Now addBreakpoint only supplies the jp and adviceMethod
 * o Started converting CreateDeltas to use the metadata from the Debugger
 * 
 * 1     4/20/06 10:48p Eaddy
 * o Moved files from Wicca.Engine to wicca project
 * 
 * 35    4/20/06 6:49p Eaddy
 * o Fixed bug where weaving doesn't work if arguments are passed
 * to the client program using wrun
 * 
 * 34    4/20/06 5:41p Eaddy
 * o Now gets the verbosity setting from mdbg.  However, to avoid
 * requiring an interface
 * change to mdbg, Wicca will use a default verbosity if the field isn't
 * present (we use
 * Reflection to look for it).  When this is made an official change to
 * mdbg we will stop
 * using Reflection.
 * 
 * 33    4/19/06 1:40a Eaddy
 * o Made verbosity level more fine grained
 * o Now all output is done through a common method
 * 
 * 32    4/19/06 1:16a Eaddy
 * 
 * 31    4/18/06 11:41p Eaddy
 * o Added diagnostics message
 * o No longer tries to patch every time an assembly is loaded.
 * Must explicitly use the "patch" or "weave" commands.
 * o Fixed bug where a "patch" was ignored because the module
 * hadn't been loaded yet
 * 
 * 30    4/12/06 8:15p Eaddy
 * o Can now perform a patch even if the loading is in a funny order
 * 
 * 29    4/12/06 7:37p Eaddy
 * o Removed event mgr stuff that was slowing down startup time
 * 
 * 28    4/12/06 4:02p Eaddy
 * o Changed CreateDeltas to PatchActiveProcess
 * o Changed beginExecution to StartOrResumeExecution
 * o Made access to pendingTasks thread safe
 * 
 * 27    4/11/06 3:39p Asv2105
 * Added ability to call code in debugged process
 * 
 * 26    4/11/06 12:42p Eaddy
 * o Now displays msg before weaving
 * 
 * 25    4/07/06 8:52p Eaddy
 * o Removed unused code
 * 
 * 24    4/07/06 6:41p Eaddy
 * o Now tries to apply the deltas
 * 
 * 23    4/04/06 9:28p Eaddy
 * o Now obtains paths from AssemblyHelper
 * 
 * 22    4/03/06 1:08a Asv2105
 * Moved location of breakpoint hit notification to above advice
 * invokation
 * Changed so breakpoints are only added once
 * 
 * 21    4/03/06 12:52a Asv2105
 * Added a bunch of code for executing code in the client process when a
 * breakpoint is hit that doesn't entirely work
 * 
 * 20    4/03/06 12:24a Eaddy
 * o Changed jp.WithinSignature to jp.Within and made it an
 * IJoinpointSignature.
 * This is needed because we need access to return type, arg types, etc.
 * just like we do for jp.Signature.
 * 
 * 19    4/02/06 5:23p Asv2105
 * Made so advice method is called with dynamic info pulled from
 * SimpleJoinpoint
 * 
 * Added first shot at filling SimpleJoinpoint with useful data
 * 
 * 18    4/02/06 4:34p Asv2105
 * Added ability for BreakpointWeaver to pass MethodInfo to the
 * coordinator
 * 
 * Adjusted ICoordinator, IEventListener interfaces to match reality
 * 
 * Added ability for Coordinator to call static methods when a breakpoint
 * fires and pass off to EventMgr for non-static methods
 * 
 * 17    4/02/06 11:32a Eaddy
 * o Fixed exception
 * 
 * 16    4/02/06 10:53a Eaddy
 * o Fixed compile error
 * 
 * 15    4/02/06 10:51a Eaddy
 * o Starting to pass args
 * 
 * 14    4/01/06 10:43a Eaddy
 * o Now catches COM exceptions
 * 
 * 13    4/01/06 10:25a Eaddy
 * o Started to create some jp context
 * 
 * 12    4/01/06 10:00a Eaddy
 * o Changed formatting back to the way it was orginally
 * 
 * 11    4/01/06 9:53a Eaddy
 * o Brokeout some classes
 * 
 * 10    3/31/06 7:31p Eaddy
 * o Separated CheckDeltas so WiccaPlugin can use it
 * 
 * 9     3/31/06 5:24p Asv2105
 * Added initial EventMgr setup stuff
 * 
 * 8     3/31/06 1:36p Eaddy
 * o Changed bp to breakpoint
 * 
 * 7     3/16/06 1:56a Eaddy
 * o Now creates delta metadata file during binary weaving
 * 
 * 6     3/15/06 4:57p Eaddy
 * 
 * 5     3/15/06 3:05p Eaddy
 * o Added prelim support for specifying weaver type from the command-line
 * 
 * 4     3/15/06 2:36p Eaddy
 * o Now callback value serves as the flag
 * 
 * 3     3/15/06 2:30p Eaddy
 * o Now sets the breakpoint weaver callback in Phx.Morph
 * 
 * 2     3/15/06 12:07p Eaddy
 * o Now weaves using Phx.Morph
 * 
 * 1     3/14/06 5:45p Eaddy
 */

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;

using Dbg = System.Diagnostics.Debug;

using Microsoft.Samples.Tools.Mdbg;
using Microsoft.Samples.Debugging.MdbgEngine;
using Microsoft.Samples.Debugging.CorDebug;
using Microsoft.Samples.Debugging.CorDebug.NativeApi;

using Phx.Aop.Joinpoints;

namespace Wicca
{
    public class Coordinator : MarshalByRefObject
    {
        const string BP_STOP_REASON = "BreakpointHit";
        const string CUSTOM_BP_STOP_REASON = "CustomBreakpointHit";
        const string REMAP_OPPORTUNITY_STOP_REASON = "RemapOpportunityReached";
        const string REMAP_COMPLETE_STOP_REASON = "FunctionRemapComplete";

        /// <summary>
        /// The active process
        /// </summary>
        MDbgProcess m_process = null;

        /// <summary>
        /// Breeakpoints that will be set when the main module is loaded
        /// </summary>
        Stack<BreakpointLineNumberLocation> pendingBreakpoints = null;

        /// <summary>
        /// The advices associted with each breakpoint (advised join point)
        /// </summary>
        Dictionary<BreakpointLineNumberLocation, List<AdviceClosure>> advisedJoinpoints = 
            new Dictionary<BreakpointLineNumberLocation, List<AdviceClosure>>();

        /// <summary>
        /// Advice waiting to be invoked by the Shell thread
        /// </summary>
        Queue<AdviceClosure> pendingAdvice = new Queue<AdviceClosure>();

        /// <summary>
        /// Performs the binary or breakpoint weaving
        /// </summary>
        DynamicWeaver m_dynamicBinaryWeaver = null;

        /// <summary>
        /// The fully-qualified process path
        /// </summary>
        public string ProcessPath
        {
            get
            {
                if (m_process == null)
                    return null;

                string path;

                // Extract command-line args
                int index = m_process.Name.IndexOf(".exe ");
                if (index == -1)
                    path = m_process.Name;
                else
                    path = m_process.Name.Substring(0, index+4);

                // The Debugger expects the path to be full-qualified
                // or filename only
                return Path.GetFullPath(path);
            }
        }

        /// <summary>
        /// The weaver type
        /// </summary>
        public WeaverType WeaverType
        {
            get { return m_dynamicBinaryWeaver.WeaverType; }
            set { m_dynamicBinaryWeaver.WeaverType = value; }
        }

        /// <summary>
        /// The loaded advice methods
        /// </summary>
        public IEnumerable Advices
        {
            [System.Diagnostics.DebuggerStepThrough]
            get { return Phx.Morph.Morpher.Advices; }
        }

        /// <summary>
        /// Used by the CoordinatorFactory to create the singleton Coordinator
        /// </summary>
        internal Coordinator()
        {
            m_dynamicBinaryWeaver = new DynamicWeaver(  SetBreakpointCallback, 
                                                        WeaverType.breakpoint);
        }

        public void Init(int verbosity)
        {
            WiccaPlugin.Shell.OnCommandExecuted += CommandExecutedEventHandler;

            // Ensures Morpher's verbosity is set

            m_dynamicBinaryWeaver.Verbosity = verbosity;
        }

        /// <summary>
        /// Called after each shell command is executed
        /// </summary>
        /// <remarks>
        /// We must trap this event so that we can trap module load events and
        /// other events.  If we don't do this here, we will miss the first few
        /// events.
        /// </remarks>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void CommandExecutedEventHandler(object sender, CommandExecutedEventArgs e)
        {
            // Although the underlying MDbg engine supports multiple processes,
            // We'll only support 1 process from the UI to keep things simple.             
            if (m_process == null && WiccaPlugin.Debugger.Processes.HaveActive)
            {
                m_process = WiccaPlugin.Debugger.Processes.Active;
                Dbg.Assert(m_process != null);

                m_process.PostDebugEvent += PostCallbackEventHandler;

                // Now that we've gotten the module load event we can clear
                // the RawModeType flag that was set by the 'wrun' command
                m_process.RawModeType = MDbgProcess.RawMode.None;
            }
        }

        /// <summary>
        /// Handles module load events and exceptions
        /// </summary>
        /// <remarks>
        /// When a module is loaded we extract its advice and use them to
        /// weave all the other loaded modules
        /// </remarks>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void PostCallbackEventHandler(object sender, CustomPostCallbackEventArgs e)
        {
            Dbg.Assert(sender == m_process);

            string processName = Path.GetFileName(ProcessPath);

            string msg = string.Format("{0} {1}", processName, e.CallbackType);

            if (e.CallbackType == ManagedCallbackType.OnModuleLoad)
            {
                CorModuleEventArgs args = e.CallbackArgs as CorModuleEventArgs;

                if (args.Module.Name == "<unknown>")
                {
                    msg += ": <unknown>";
                    WiccaPlugin.WriteOutput(msg);
                    return;
                }

                string moduleName = Path.GetFileName(args.Module.Name);

                msg += ": " + moduleName;

                WiccaPlugin.WriteOutput(msg);

                bool isMainModule = (moduleName == processName);

                if (isMainModule)
                {
                    AddPendingBreakpoints();
                }

                m_dynamicBinaryWeaver.addModule(args.Module.Name, true, isMainModule);
            }
            else if (e.CallbackType == ManagedCallbackType.OnException ||
                     e.CallbackType == ManagedCallbackType.OnException2)
            {
                CorThreadEventArgs exArgs = e.CallbackArgs as CorThreadEventArgs;
                if (exArgs != null && 
                   exArgs.Thread != null &&
                   exArgs.Thread.CurrentException != null)
                {
                   MDbgValue mv = new MDbgValue(WiccaPlugin.Debugger.Processes.Active,
                      exArgs.Thread.CurrentException);
                   msg += ": " + mv.GetStringValue(1);
                }

                WiccaPlugin.WriteError(msg);
            }
            else
            {
                WiccaPlugin.WriteOutput(msg);
            }
        }

        /// <summary>
        /// Explicitly loads an aspect module
        /// </summary>
        /// <param name="aspectPath"></param>
        public void LoadAspect(string aspectPath)
        {
            m_dynamicBinaryWeaver.addModule(aspectPath, false, false);
        }

        /// <summary>
        /// Patches the currently running process so that its the same
        /// as the specified module
        /// </summary>
        /// <param name="newModulePath"></param>
        public void CreateAndApplyDeltas(string newModulePath)
        {
            m_dynamicBinaryWeaver.createAndApplyDeltas(ProcessPath, newModulePath);
        }

        /// <summary>
        /// 
        /// </summary>
        public void RunToCompletion()
        {
            while (WiccaPlugin.Debugger.Processes.Count > 0)
            {
                WiccaPlugin.Debugger.Processes.Active.Go().WaitOne();
                if (WiccaPlugin.Debugger.Processes.Count == 0)
                    break;

                string stopReason = WiccaPlugin.Debugger.Processes.Active.StopReason.ToString();

                WiccaPlugin.WriteOutput("Stop reason: {0}", 
                    WiccaPlugin.Debugger.Processes.Active.StopReason);

                if (stopReason == REMAP_OPPORTUNITY_STOP_REASON)
                {
                   RemapOpportunityReachedStopReason sr =
                      (RemapOpportunityReachedStopReason) WiccaPlugin.Debugger.Processes.Active.StopReason;
                   int newILOffset = 0;
                   sr.Thread.ActiveFrame.RemapFunction(newILOffset);
                   WiccaPlugin.Debugger.Processes.Active.Threads.RefreshStack();
                   WiccaPlugin.WriteOutput("Remap successful.");
                }
                else if (stopReason == CUSTOM_BP_STOP_REASON)
                {
                   int pendingCount;
                   
                   do
                   {
                       AdviceClosure closure;
                       lock (pendingAdvice)
                       { 
                           pendingCount = pendingAdvice.Count;
                           if (pendingCount <= 0)
                               break;

                           closure = pendingAdvice.Dequeue(); 
                           pendingCount = pendingAdvice.Count;
                       } 

                       InvokeAdvice(closure);

                   } while (pendingCount > 0);
                }
                else if (stopReason == "MDbgError")
                {
                   MDbgErrorStopReason sr =
                      (MDbgErrorStopReason) WiccaPlugin.Debugger.Processes.Active.StopReason;
                   WiccaPlugin.WriteError("Debugger Error: " + sr.ExceptionThrown);
                }
            }
        }

        //-------------------------------------------------
        // Breakpoint weaving and handling
        //-------------------------------------------------

        /// <summary>
        /// Sets a breakpoint for a source file in the current process
        /// </summary>
        /// <remarks>
        /// If the process isn't loaded yet, the breakpoint will be
        /// queued
        /// </remarks>
        /// <param name="filename"></param>
        /// <param name="linenumber"></param>
        public void SetBreakpoint(string fileName, uint line)
        {
            if (!IsValidBreakpoint(fileName, line))
            {
                return;
            }

            SetBreakpoint(new BreakpointLineNumberLocation(fileName, (int) line));
        }

        /// <summary>
        /// Sets a breakpoint for a source file in the current process
        /// </summary>
        /// <remarks>
        /// If the process isn't loaded yet, the breakpoint will be
        /// queued
        /// </remarks>
        /// <param name="info"></param>
        public void SetBreakpoint(BreakpointLineNumberLocation info) 
        {
            if (pendingBreakpoints == null)
                pendingBreakpoints = new Stack<BreakpointLineNumberLocation>();

            pendingBreakpoints.Push(info);

            AddPendingBreakpoints();
        }

        /// <summary>
        /// Called by Phx.Morph when an advice matches a join point
        /// </summary>
        /// <param name="jp"></param>
        /// <param name="adviceMethodToCall"></param>
        public void SetBreakpointCallback(   IJoinpoint jp, 
                                             uint lineToSetBreakpoint,
                                             MethodInfo adviceMethodToCall)
        {
            Dbg.Assert(jp != null);
            Dbg.Assert(jp.Within != null);
            Dbg.Assert(jp.Within.SourceLocation != null);
            Dbg.Assert(jp.Within.SourceLocation.FileName != null);
            Dbg.Assert(adviceMethodToCall != null);

            string fileName = jp.Within.SourceLocation.FileName.ToString();

            // For after advice, jp.Within.SourceLocation.Line may not be the
            // same as lineToSetBreakpoint
            Dbg.Assert(lineToSetBreakpoint >= jp.Within.SourceLocation.Line);
            uint line = lineToSetBreakpoint;

            if (!IsValidBreakpoint(fileName, line))
            {
                return;
            }

            BreakpointLineNumberLocation bp = new BreakpointLineNumberLocation(fileName, 
                                                                              (int) line);

            List<AdviceClosure> closures;
            if (!advisedJoinpoints.TryGetValue(bp, out closures))
            {
                // This is the first time we are setting this particular
                // breakpoint location so create the initial advice closure
                // list and set the breakpoint

                closures = new List<AdviceClosure>(1);
                advisedJoinpoints.Add(bp, closures);
                
                // closures.Add must be called before SetBreakpoint
                closures.Add(new AdviceClosure(jp, adviceMethodToCall));
                SetBreakpoint(bp);
            }
            else
            {
                closures.Add(new AdviceClosure(jp, adviceMethodToCall));
            }
        }

        /// <summary>
        /// Sets breakpoints in a module as soon as it is loaded
        /// </summary>
        void AddPendingBreakpoints()
        {
            if (m_process == null || pendingBreakpoints == null)
            {
                return;
            }

            while (pendingBreakpoints.Count > 0)
            {
                BreakpointLineNumberLocation bpLoc = pendingBreakpoints.Pop();
                MDbgBreakpoint mbp = m_process.Breakpoints.CreateBreakpoint(bpLoc, 
                                                        CustomBreakpointEventHandler);

                Dbg.Assert(mbp.Enabled);
                Dbg.Assert(mbp.IsBound);

                CorFunctionBreakpoint cbp =
                   mbp.CorBreakpoint as CorFunctionBreakpoint;
                Dbg.Assert(cbp != null);

                MDbgFunction func = 
                   WiccaPlugin.Debugger.Processes.Active.Modules.LookupFunction(cbp.Function);
                Dbg.Assert(func != null);
                if (func == null)
                   return;

                MDbgSourcePosition pos = func.GetSourcePositionFromIP(cbp.Offset);

                if (pos.Line == bpLoc.LineNumber)
                {
                    WiccaPlugin.WriteOutput("Custom breakpoint set: {0}, {1}",
                          bpLoc.FileName,
                          bpLoc.LineNumber);
                }
                else
                {
                    // For after advice, we just set the bp line to be the original
                    // jp line + 1.  This line may not actually exist so 
                    // CreateBreakpoint() will use the next available line. When
                    // this happens, we have to update our advisedJoinpoints map
                    // with the new line number.

                    Dbg.Assert(pos.Line > bpLoc.LineNumber);

                    WiccaPlugin.WriteOutput("Custom breakpoint set: {0}, {1} (Orig Line: {2})",
                          bpLoc.FileName,
                          pos.Line,
                          bpLoc.LineNumber);

                    // Get the advices for the old bp location
                    List<AdviceClosure> advices = advisedJoinpoints[bpLoc];
                    
                    // Remove the old bp location
                    advisedJoinpoints.Remove(bpLoc);

                    // Add the new bp location
                    bpLoc = new BreakpointLineNumberLocation(bpLoc.FileName, pos.Line);
                    advisedJoinpoints[bpLoc] = advices;
                }
            }
        }

        /// <summary>
        /// Called by mdbg when a custom breakpoint (i.e., an advised join point) is hit
        /// </summary>
        /// <remarks>
        /// We can't invoke the advice from the thread that called this callback
        /// function because this causes a deadlock.  Instead, we queue the
        /// advice and signal the controller to stop so that the controller thread
        /// can execute the advice.
        /// </remarks>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void CustomBreakpointEventHandler(object sender, CustomBreakpointEventArgs e)
        {
            MDbgProcess process = sender as MDbgProcess;
            Dbg.Assert(process == m_process);

            CorFunctionBreakpoint bp =
               e.BreakpointHitCallbackArgs.Breakpoint as CorFunctionBreakpoint;

            Dbg.Assert(bp != null);
            if (bp == null)
                return;

            MDbgFunction func = WiccaPlugin.Debugger.Processes.Active.Modules.LookupFunction(bp.Function);
            Dbg.Assert(func != null);
            if (func == null)
                return;

            MDbgSourcePosition pos = func.GetSourcePositionFromIP(bp.Offset);

            BreakpointLineNumberLocation bpHit = new BreakpointLineNumberLocation(  Path.GetFileName(pos.Path), 
                                                        pos.Line);

            WiccaPlugin.WriteOutput("Custom breakpoint hit: {0}, {1}",
                  bpHit.FileName,
                  bpHit.LineNumber);

            List<AdviceClosure> advices;

            if (advisedJoinpoints.TryGetValue(bpHit, out advices))
            {
                Dbg.Assert(advices.Count > 0);

                lock (pendingAdvice)
                {
                    foreach (AdviceClosure advice in advices)
                    {
                        advice.jp.SetProcessAndFunction(process, func);
                        pendingAdvice.Enqueue(advice);
                    }
                }
            }
            else
            {
                WiccaPlugin.WriteError("Custom breakpoint hit ({0},{1}) but there is no advice.",
                      Path.GetFileName(pos.Path),
                      pos.Line);
                return;
            }

            e.Controller.Stop(e.BreakpointHitCallbackArgs.Thread, CUSTOM_BP_STOP_REASON);
        }

        /// <summary>
        /// Verifies the breakpoint info
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="line"></param>
        /// <returns>
        /// True if the file is nonempty and the line is greater than 0
        /// </returns>
        bool IsValidBreakpoint(string fileName, uint line)
        {
            if (fileName == null)
            {
                WiccaPlugin.WriteError("Failed to set breakpoint: File name is null.");
                return false;
            }
            else if (fileName == string.Empty ||
                fileName == "(null)")
            {
                WiccaPlugin.WriteError("Failed to set breakpoint: File name is empty.");
                return false;
            }
            else if (line <= 0)
            {
                WiccaPlugin.WriteError("Failed to set breakpoint for file '{0}' and line {1}: " +
                    "Line number must be positive.", fileName, line);
                return false;
            }

            return true;
        }

        /// <summary>
        /// Invokes the advice method
        /// </summary>
        /// <param name="context"></param>
        void InvokeAdvice(AdviceClosure adviceClosure)
        {
            Dbg.Assert(adviceClosure != null);

            MethodInfo mi = adviceClosure.adviceMethodToCall;
            Dbg.Assert(mi != null);
            Dbg.Assert(mi.DeclaringType != null);

            // Step 1:  Convert the MethodInfo into an MDbgFunction we can call

            string adviceMethodName = mi.DeclaringType.FullName + "." + mi.Name;

            MDbgFunction func = 
                WiccaPlugin.Debugger.Processes.Active.ResolveFunctionNameFromScope(adviceMethodName);
            if (func == null)
            {
                WiccaPlugin.WriteError("Advice method not found: '{0}'. Are you missing a reference?", 
                    adviceMethodName);
                return;
            }

            // Step 2:  Create the advice method context parameters

            ParameterInfo[] paramsNeeded = mi.GetParameters();
            Dbg.Assert(paramsNeeded != null);

            List<CorValue> paramsToPass = 
                new List<CorValue>(paramsNeeded.Length > 0 ? paramsNeeded.Length : 1);

            // Bind each context parameter

            for(int i = 0; i < paramsNeeded.Length; ++i)
            {
                object[] paramAttributes = paramsNeeded[i].GetCustomAttributes(true);
                
                bool hasPhxAttribute = false;

                // Find the first Phx.Aop.XXX attribute for this context parameter

                for(int j = 0; j < paramAttributes.Length; ++j)
                {
                    object paramValue = 
                        adviceClosure.jp.getDataForAttribute(paramAttributes[j].GetType(), 
                                                             paramAttributes[j]);

                    if (paramValue == null &&
                        (paramAttributes[j].GetType().Namespace == "Phx.Aop" ||
                        paramAttributes[j].GetType().Namespace == "Phx.Morph"))
                    {
                        // The parameter may have a non-Phx attribute (rare but
                        // possible)
                        continue;
                    }

                    hasPhxAttribute = true;

                    if (paramValue == null)
                    {
                        WiccaPlugin.WriteError("Failed to obtain value for context parameter {0} (Type: {1}) of advice method '{2}'.",
                            i+1, paramAttributes[j].GetType(), adviceMethodName);

                        // Pass null since we can't obtain the correct value
                        paramsToPass.Add(CreateNull());
                    }
                    else if (paramValue.GetType().IsPrimitive)
                    {
                        paramsToPass.Add(CreatePrimitive(paramValue));
                    }
                    else if (paramValue is CorValue)
                    {
                        paramsToPass.Add((CorValue)paramValue);
                    }
                    else if (paramValue is string)
                    {
                        paramsToPass.Add(CreateStringLiteral((string)paramValue));
                    }
                    else
                    {
                        WiccaPlugin.WriteError("Context parameter {0} (Attribute: {1}) of advice method '{2}' has an unsupported type: {3}.",
                           i+1, paramAttributes[j].GetType(), adviceMethodName, paramValue.GetType());

                        // Pass null since we can't obtain the correct value
                        paramsToPass.Add(CreateNull());
                    }

                    break;
                }

                if (paramAttributes.Length == 0 || !hasPhxAttribute)
                {
                    WiccaPlugin.WriteError(
                        "Parameter {0} of advice method '{1}' cannot be passed because it does not have a context parameter attribute. Passing null instead.",
                        i, adviceMethodName);

                    // Pass null since we can't obtain the correct value
                    paramsToPass.Add(CreateNull());
                }
            }

            // Step 3:  Call the advice method using func-eval

            CorEval eval = WiccaPlugin.Debugger.Processes.Active.Threads.Active.CorThread.CreateEval();
            eval.CallFunction(func.CorFunction, paramsToPass.ToArray());

            // Step 4:  Wait for the advice method to complete

            do
            {
                WiccaPlugin.Debugger.Processes.Active.Go().WaitOne();

            } while (WiccaPlugin.Debugger.Processes.Count > 0 && 
                !(WiccaPlugin.Debugger.Processes.Active.StopReason is EvalCompleteStopReason));
        }

        /// <summary>
        /// Creates a CorValue for a primitive object (int, uint, etc.)
        /// </summary>
        /// <param name="val"></param>
        /// <returns></returns>
        CorValue CreatePrimitive(object val) 
        {
            Dbg.Assert(val.GetType().IsPrimitive);

            CorElementType corType;

            TypeCode typeCode = Type.GetTypeCode(val.GetType());

            switch (typeCode)
            {
                case TypeCode.Boolean:
                    corType = CorElementType.ELEMENT_TYPE_BOOLEAN; break;
                case TypeCode.Char:
                    corType = CorElementType.ELEMENT_TYPE_CHAR; break;
                case TypeCode.SByte:
                    corType = CorElementType.ELEMENT_TYPE_I1; break;
                case TypeCode.Byte:
                    corType = CorElementType.ELEMENT_TYPE_U1; break;
                case TypeCode.Int16:
                    corType = CorElementType.ELEMENT_TYPE_I2; break;
                case TypeCode.UInt16:
                    corType = CorElementType.ELEMENT_TYPE_U2; break;
                case TypeCode.Int32:
                    corType = CorElementType.ELEMENT_TYPE_I4; break;
                case TypeCode.UInt32:
                    corType = CorElementType.ELEMENT_TYPE_U4; break;
                case TypeCode.Int64:
                    corType = CorElementType.ELEMENT_TYPE_I8; break;
                case TypeCode.UInt64:
                    corType = CorElementType.ELEMENT_TYPE_U8; break;
                case TypeCode.Single:
                    corType = CorElementType.ELEMENT_TYPE_R4; break;
                case TypeCode.Double:
                    corType = CorElementType.ELEMENT_TYPE_R8; break;
                default:
                    WiccaPlugin.WriteError("Unsupported parameter value type: {0}.",
                        typeCode);
                    return null;
            }

            CorEval eval = 
                WiccaPlugin.Debugger.Processes.Active.Threads.Active.CorThread.CreateEval();
            if (eval == null)
                return null;

            CorValue corVal = eval.CreateValue(corType, null);
            if (corVal == null)
                return null;

            CorGenericValue primitiveVal = corVal.CastToGenericValue();
            if (primitiveVal == null)
                return null;

            primitiveVal.SetValue(val);

            return primitiveVal;
        }

        CorValue CreateStringLiteral(string value) 
        {
            CorEval eval = 
                WiccaPlugin.Debugger.Processes.Active.Threads.Active.CorThread.CreateEval();
            if (eval == null)
                return null;

            eval.NewString(value);

            do
            {
                WiccaPlugin.Debugger.Processes.Active.Go().WaitOne();

            } while (!(WiccaPlugin.Debugger.Processes.Active.StopReason is EvalCompleteStopReason));

            return (WiccaPlugin.Debugger.Processes.Active.StopReason as EvalCompleteStopReason).Eval.Result;
        }

        CorValue CreateNull() 
        {
            CorEval e = WiccaPlugin.Debugger.Processes.Active.Threads.Active.CorThread.CreateEval();
            return e.CreateValueForType(WiccaPlugin.Debugger.Processes.Active.ResolveType("System.Object"));
        }
    }
}
