//-----------------------------------------------------------------------------
//
// Phoenix
// Copyright (C) Microsoft Corporation.  All Rights Reserved.
//
// $Header: /Wicca/wicca/DebugJoinpoint.cs 7     5/24/06 2:02p Eaddy $
//
// Description:
//
// Remarks:
//
//-----------------------------------------------------------------------------
/**
 * $Log: /Wicca/wicca/DebugJoinpoint.cs $
 * 
 * 7     5/24/06 2:02p Eaddy
 * o Added Equals() and GetHashCode() so that we can
 * put breakpoint locations in a map
 * 
 * 6     5/24/06 1:53a Eaddy
 * o Breakpoint weaver now supports [Arg]
 * 
 * 5     5/24/06 1:34a Eaddy
 * o Added ArgNames to IJoinpointSignature so that we can
 * pass context args to the advice method
 * 
 * 4     5/23/06 7:49p Eaddy
 * o More cleanup
 * 
 * 3     5/23/06 7:41p Eaddy
 * o Cleaning up bpw code
 */

using System;

using Dbg = System.Diagnostics.Debug;

using Wicca;
using Phx.Aop.Attributes;
using Microsoft.Samples.Debugging.MdbgEngine;

namespace Phx.Aop.Joinpoints 
{
    public class DebugJoinpoint : SimpleJoinpoint 
    {
        MDbgProcess process;
        MDbgFunction function;

        public DebugJoinpoint(IJoinpoint jp) 
            : base(jp) 
        { }

        public void SetProcessAndFunction(MDbgProcess process, MDbgFunction f)
        {
            this.process = process;
            this.function = f;
        }

        public override object getDataForAttribute(Type attributeType, object attributeVal) 
        {
            Dbg.Assert(process != null && function != null,
                "Must call SetProcessAndFunction() first!");
            Dbg.Assert(attributeVal != null);

            if (attributeVal is ThisAttribute) 
            {
                MDbgValue val = WiccaPlugin.Debugger.Processes.Active.ResolveVariable("this", 
                    process.Threads.Active.CurrentFrame);
                if (val == null) 
                {
                    return null;
                }
                return val.CorValue;
            } 
            else if (attributeVal is ArgAttribute)
            {
                ArgAttribute argAttribute = attributeVal as ArgAttribute;
                Dbg.Assert(argAttribute != null);

                int argIndex = argAttribute.Index;
                
                string argNameToUse = null;

                int argCount = 0;
                foreach(System.Text.StringBuilder argName in this.Signature.ArgNames)
                {
                    if (argIndex == argCount)
                    {
                        argNameToUse = argName.ToString();
                        break;
                    }

                    ++argCount;
                }

                if (argIndex < 0 || argIndex > argCount)
                {
                    WiccaPlugin.WriteError("Invalid advice method parameter '[Arg({0})]'.  Expected [Arg(0)]-[Arg({1})].",
                       argIndex,
                       argCount);
                    return null;
                }

                Dbg.Assert(argNameToUse != null);

                MDbgValue val = WiccaPlugin.Debugger.Processes.Active.ResolveVariable(argNameToUse, 
                    process.Threads.Active.CurrentFrame);
                if (val == null) 
                {
                    return null;
                }
                return val.CorValue;
            }
            else if (attributeVal is TargetAttribute) 
            {
                // NYI
            }
            
            return base.getDataForAttribute(attributeType, attributeVal);
        }

        public override string ToString()
        {
           return base.ToString();
        }
    }
}
