///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2004 Columbia University.  All Rights Reserved.
//
//      $Header: /Common/AssemblyHelper/MethodInfo.cpp 9     5/10/06 8:02p Eaddy $
//
//    File name:	MethodInfo.cpp
//   Created By:	Marc Eaddy
//   Created on:	10/24/04
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Common/AssemblyHelper/MethodInfo.cpp $
 * 
 * 9     5/10/06 8:02p Eaddy
 * o Closer to multiple weaving
 * 
 * 8     5/03/06 8:15p Eaddy
 * o Now allows IMetaDataImport to come from Wicca
 * 
 * 7     4/22/06 12:35p Eaddy
 * o Adding support for using the Debugger API to query for the current
 * metadata and IL
 * 
 * 6     4/18/06 11:43p Eaddy
 * o Changed back to the old emit algorithm
 * o Improved diagnostics
 * o No longer outputs empty dil or dmeta files
 * 
 * 5     4/14/06 5:36p Eaddy
 * o Trying to fix emitting methods
 * 
 * 4     4/12/06 7:39p Eaddy
 * o Fixed bug where type name wasn't retrieved yet
 * 
 * 3     4/11/06 11:12p Eaddy
 * o Dynamic binary updating proof-of-concept works!
 * 
 * 2     4/07/06 8:54p Eaddy
 * o Now supports outputing updated methods to a char buffer
 * in addition to using IMemoryAlloc
 * 
 * 1     3/16/06 1:53a Eaddy
 * 
 * 1     3/16/06 1:49a Eaddy
 * 
 * 16    1/20/06 9:39p Eaddy
 * o Migrated to VS 2005.  Replaced many function calls with their safe
 * version.
 * 
 * 15    12/16/05 8:40p Eaddy
 * o Added isCtor and isCCtor
 * 
 * 14    12/15/05 2:37a Eaddy
 * o Improved statistics gathering and display
 * 
 * 13    12/14/05 2:16a Eaddy
 * o Even if we don't support emitting new types, methods, and fields, 
 * we still need to fix them up.  And we can't assume that they can
 * be iterated in lock step since they may have been reordered.
 * Actually, this only really happens when the build description file
 * doesn't match the executable because the exe was built using
 * VS.
 * 
 * 12    12/10/05 12:40a Eaddy
 * o Now each patch keeps track of its time
 * 
 * 11    12/09/05 9:33p Eaddy
 * o Removed log message
 * o Comments
 * 
 * 10    12/09/05 9:03p Eaddy
 * o Made improvements to try to emit methods but this doesn't work
 * because DefineMethod must be called during ModuleLoadFinished
 * 
 * 9     12/09/05 4:32p Eaddy
 * o No longer depends on methodDef_iterator
 * 
 * 8     12/08/05 10:34p Eaddy
 * o Now really fixed problem with changing a method's local variables
 * 
 * 7     12/08/05 8:29p Eaddy
 * o Now compares signatures correctly
 * o Now supports changing a method's local variables
 * 
 * 6     12/07/05 7:56p Eaddy
 * o Now uses SigCmp
 * o Now handles comparing method tokens as well as strings
 * 
 * 5     11/11/05 5:28p Eaddy
 * o Improved diagnostics
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * Revision 1.4  2005/01/14 23:28:47  me133
 * o Now initialized method name
 *
 * Revision 1.3  2005/01/14 23:03:47  me133
 * o Made some functions static
 * o Move check into separate function
 *
 * Revision 1.2  2004/10/26 04:43:07  me133
 * no message
 *
 * Revision 1.1  2004/10/25 01:03:49  me133
 * o Method IL comparison is now complete and correct
 */

#include "stdafx.h"

#include "MethodInfo.h"			// Me

#include "ILHelper.h"			// opcode_iterator, FindOpCode()

#include "MetadataHelpers.h"	// Token2TypeName()

#include "ILFetcher.h"          // CILFetcher

#include <dbghelp.h>			// ImageNtHeader()

using namespace std;

// HELPER FUNCTIONS

HRESULT MethodILIsEquivalent(	const CMethodInfo& pre, 
								const CMethodInfo& post, 
								const TypeListTokenMappings& listEquivalentTokens);

///////////////////////////////
// METHOD INFO IMPLEMENTATION
///////////////////////////////

CMethodInfo::CMethodInfo()
   : m_bMustDelete(false) // This has to be done before calling clear()!
{ 
	clear();
}

CMethodInfo::~CMethodInfo()
{
   if (m_bMustDelete)
      delete[] m_pbMethodHeader;
}

#pragma warning (disable:4351) // elements of array will be default initialized
CMethodInfo::CMethodInfo(const CMethodInfo& that)
	: m_spiImport(that.m_spiImport)
    , m_pilFetcher(that.m_pilFetcher)
	, m_typeDef(that.m_typeDef)
	, m_methodDef(that.m_methodDef)
	, m_szTypeName()
	, m_szMemberName()
	, m_ulNameLen(that.m_ulNameLen)
	, m_dwFlags(that.m_dwFlags)
	, m_tokSig(that.m_tokSig)
	, m_pbSigBlob(that.m_pbSigBlob)
	, m_ulSigBlob(that.m_ulSigBlob)
	, m_ulCodeRVA(that.m_ulCodeRVA)
	, m_ulImplFlags(that.m_ulImplFlags)
	, m_pbMethodHeader(that.m_pbMethodHeader)
    , m_bMustDelete(false)
	, m_ulMethodSize(that.m_ulMethodSize)
	, m_ulMethodHeaderSize(that.m_ulMethodHeaderSize)
	, m_pbMethodIL(that.m_pbMethodIL)
	, m_ulCodeSize(that.m_ulCodeSize)
	, m_tokLocalVarSig(that.m_tokLocalVarSig)
	, m_pLocalVarSig(that.m_pLocalVarSig)
	, m_ulLocalVarSigSize(that.m_ulLocalVarSigSize)
	, m_moduleID(that.m_moduleID)
	, m_functionID(that.m_functionID)
	, m_bWasModified(that.m_bWasModified)
	, m_dwTicksStart(0)
	, m_dwWeaveTimeEnd(0)
	, m_dwCompileTimeEnd(0)
	, m_dwPatchTimeEnd(0)
	, m_dwWaitForReJitEnd(0)
{
	wcscpy_s(m_szTypeName, that.m_szTypeName);
	wcscpy_s(m_szMemberName, that.m_szMemberName);

    if (that.m_ulMethodSize > 0 && that.m_bMustDelete)
    {
        cloneMethodInfo(that, 0);
        _ASSERTE(m_bMustDelete);
    }
}

#pragma warning (default:4351)

void CMethodInfo::clear()
{
	m_spiImport = 0;
    m_pilFetcher = 0;
	m_typeDef = mdTypeDefNil;
	m_methodDef = mdMethodDefNil;
	*m_szTypeName = L'\0';
	*m_szMemberName = L'\0';
	m_ulNameLen = 0;
	m_dwFlags = 0;
	m_tokSig = mdSignatureNil;
	m_pbSigBlob = 0;
	m_ulSigBlob = 0;
	m_ulCodeRVA = 0;
	m_ulImplFlags = 0;

    if (m_bMustDelete)
    {
      delete[] m_pbMethodHeader;
      m_bMustDelete = false;
    }

	m_pbMethodHeader = 0;
	m_ulMethodSize = 0;
	m_ulMethodHeaderSize = 0;
	m_pbMethodIL = 0;
	m_ulCodeSize = 0;
	m_tokLocalVarSig = mdSignatureNil;
	m_pLocalVarSig = 0;
	m_ulLocalVarSigSize = 0;
	m_moduleID = -1;
	m_functionID = -1;
	m_bWasModified = false;
	m_dwTicksStart = 0;
	m_dwTicksStart = 0;
	m_dwWeaveTimeEnd = 0;
	m_dwCompileTimeEnd = 0;
	m_dwPatchTimeEnd = 0;
	m_dwWaitForReJitEnd = 0;
}

HRESULT CMethodInfo::init(ICorProfilerInfo* piProfInfo, 
                          FunctionID functionID, 
                          const WCHAR* szTypeName, 
                          const CILFetcher* pilFetcher)
{
	HRESULT hr = S_OK;

	if (!piProfInfo) return E_POINTER;

   m_functionID = functionID;

   mdToken methodDef = mdMethodDefNil;
   CComPtr<IMetaDataImport> spiMetaDataImport;

   hr = piProfInfo->GetTokenAndMetaDataFromFunction( functionID, 
									IID_IMetaDataImport,
									(LPUNKNOWN *)&spiMetaDataImport.p, 
									&methodDef );
   if (FAILED(hr)) return hr;
	if (!spiMetaDataImport) return E_UNEXPECTED;

	return init(spiMetaDataImport, methodDef, szTypeName, pilFetcher);
}

HRESULT CMethodInfo::init(IMetaDataImport* piImport, 
                          mdMethodDef methodDef, 
                          const WCHAR* szTypeName, 
                          const CILFetcher* pilFetcher)
{
	HRESULT hr = S_OK;

	// No AddRef/Release necessary
	_ASSERTE(piImport);
	if (!piImport) return E_POINTER;

	clear();

	_ASSERTE(!m_spiImport); // Only call init() once

	m_spiImport = piImport;
	m_methodDef = methodDef;

	if (szTypeName)
	{
		_ASSERTE(*szTypeName);
		wcscpy_s(m_szTypeName, szTypeName);
	}

	// Step 1:	Get some method properties

	hr = piImport->GetMethodProps(	methodDef, 
									&m_typeDef, 
									m_szMemberName, 
									nMAX_ID,
									&m_ulNameLen, 
									&m_dwFlags, 
									&m_pbSigBlob, 
									&m_ulSigBlob, 
									&m_ulCodeRVA, 
									&m_ulImplFlags);
	VERIFY_HR(hr, _T("GetMethodProps() failed."), ILXLOG_ERROR);

    m_pilFetcher = pilFetcher;

	if (!pilFetcher) return S_FALSE;

	// Get the method sig token

	CComPtr<IMetaDataEmit> spiEmit;
	hr = m_spiImport->QueryInterface(IID_IMetaDataEmit, (void**) &spiEmit.p );
	VERIFY_HR_COND(hr, spiEmit, _T("QI for IMetaDataEmit failed."), ILXLOG_ERROR);

	// Note: The signature already exists so this should only look it up.
	// We don't expect a new sig to be emitted.

	hr = spiEmit->GetTokenFromSig(m_pbSigBlob, m_ulSigBlob, &m_tokSig);
	VERIFY_HR2(hr, ILXLOG_ERROR, _T("GetTokenFromSig('%ls.%ls') failed."), 
		getTypeName(), m_szMemberName);

	// PInvoke methods are those defined using dllimport/extern. We ignore
	// them since they aren't managed code and don't have bytecode.
	if (isPInvoke()) return S_FALSE;

	// Abstract methods don't have a method body.
	if (isAbstract()) return S_FALSE;

	if (m_ulCodeRVA == 0)
	{
		ILXLOGMSG2(ILXLOG_INFO, _T("RVA is 0 for method '%ls.%ls'."), 
			getTypeName(), m_szMemberName);
		return S_OK;
	}

	// Step 2:	Get the method header and IL code info

    bool bMustDelete = false;

	// This is expensive!
	hr = pilFetcher->GetMethodPtrInfo(	m_ulCodeRVA,
                                        m_methodDef,
							            &m_ulMethodSize,
							            &m_pbMethodHeader,
							            &m_ulMethodHeaderSize,
							            &m_pbMethodIL,
							            &m_ulCodeSize,
							            &m_tokLocalVarSig,
                                        &bMustDelete);
	VERIFY_HR2(hr, ILXLOG_ERROR, _T("GetMethodPtrInfo('%ls.%ls') failed."), 
		getTypeName(), m_szMemberName);

    if (bMustDelete)
    {
        _ASSERTE(m_bMustDelete == false);
        m_bMustDelete = bMustDelete;
    }

	//ILXLOGMSG4(ILXLOG_SUCCESS, _T("Loaded method '%ls' (Load:0x%08X RVA:0x%08X Hdr:0x%08X)."),
	//	toString(), m_pbLoadAddress, m_ulCodeRVA, m_pbMethodHeader);

	// Step 3:	Get the local variable signatures

	if (m_tokLocalVarSig != mdSignatureNil)
	{
		hr = piImport->GetSigFromToken(	m_tokLocalVarSig,
										&m_pLocalVarSig,
										&m_ulLocalVarSigSize);
		VERIFY_HR(hr, _T("GetSigFromToken() failed."), ILXLOG_ERROR);
	}

	return S_OK;
}

const WCHAR* CMethodInfo::toString() const
{
	// Not thread safe or reentrant safe! The following won't work:
	// wcscmp(lhs.toString(), rhs.toString())!

	static WCHAR szFullName[2048*2];

	wcscpy_s(szFullName, getTypeName());
	wcscat_s(szFullName, L".");
	wcscat_s(szFullName, m_szMemberName);

	return szFullName;
}

const WCHAR* CMethodInfo::toDump() const
{
	static WCHAR szDumpSig[2048];
	Sig2String(m_spiImport, m_pbSigBlob, m_ulSigBlob, szDumpSig, 
		(sizeof(szDumpSig)/sizeof(*szDumpSig))-1);
	return szDumpSig;
}

// Required by STL::map.  We are only concerned with comparing method
// name and signature here, since we're only trying to determine if
// the method is present, not if it has been updated.
bool operator<(const CMethodInfo& lhs, const CMethodInfo& rhs)
{
	int cmp = lhs.m_ulNameLen - rhs.m_ulNameLen;
	if (cmp != 0)
		return cmp < 0;

	cmp = wcscmp(lhs.m_szMemberName, rhs.m_szMemberName);
	if (cmp != 0)
		return cmp < 0;

	return SigCmp(	lhs.m_pbSigBlob, lhs.m_ulSigBlob, lhs.m_spiImport,
					rhs.m_pbSigBlob, rhs.m_ulSigBlob, rhs.m_spiImport, NULL, NULL) < 0;
}

HRESULT CMethodInfo::emit(	IMetaDataEmit* piMetaDataEmit, 
							mdToken tokEnclosingScope, 
							mdToken* ptokEmitted)
{
	HRESULT hr = S_OK;

	// Apparently, this only works when called from ModuleLoadFinished
	hr = piMetaDataEmit->DefineMethod(	tokEnclosingScope,
										m_szMemberName, 
										m_dwFlags,
										m_pbSigBlob, 
										m_ulSigBlob, 
										0, 
										m_ulImplFlags,
										ptokEmitted);
	VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to emit method '%ls'."), toString());

	return hr;
}

HRESULT CMethodInfo::emit(	IMetaDataEmit* piMetaDataEmit, 
							mdToken* ptokEmitted)
{
	HRESULT hr = S_OK;

	// Apparently, this only works when called from ModuleLoadFinished
	hr = piMetaDataEmit->DefineMethod(	m_typeDef,
										m_szMemberName, 
										m_dwFlags,
										m_pbSigBlob, 
										m_ulSigBlob, 
										m_ulCodeRVA, 
										m_ulImplFlags,
										ptokEmitted);
	VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to emit method '%ls'."), toString());

	return hr;
}

HRESULT CMethodInfo::cloneMethodInfo(const CMethodInfo& methodInfoToClone,
                                     IMethodMalloc* piAlloc)
{
	HRESULT hr = S_OK;

	_ASSERTE(wcscmp(getTypeName(), methodInfoToClone.getTypeName()) == 0);
	_ASSERTE(wcscmp(m_szMemberName, methodInfoToClone.m_szMemberName) == 0);
	_ASSERTE(m_ulNameLen == methodInfoToClone.m_ulNameLen);

	// Copy the method header and body
	methodInfoToClone.cloneMethodHeader(piAlloc, &m_pbMethodHeader, &m_bMustDelete);
	VERIFY_COND(m_pbMethodHeader, _T("Alloc() failed."), ILXLOG_ERROR);

	mdSignature tokLocalVarSigNew;

	// Update the method header and code properties

	hr = GetMethodIL(	/*[in]*/  m_pbMethodHeader,
						/*[out]*/ &m_ulMethodSize,
						/*[out]*/ &m_ulMethodHeaderSize,
						/*[out]*/ &m_pbMethodIL,
						/*[out]*/ &m_ulCodeSize,
						/*[out]*/ &tokLocalVarSigNew);

	_ASSERTE(m_ulMethodSize = methodInfoToClone.m_ulMethodSize);
	_ASSERTE(m_ulMethodHeaderSize = methodInfoToClone.m_ulMethodHeaderSize);
	_ASSERTE(m_ulCodeSize == methodInfoToClone.m_ulCodeSize);

	// Copy the flags
	m_dwFlags = methodInfoToClone.m_dwFlags;

	// Copy the impl flags
	m_ulImplFlags = methodInfoToClone.m_ulImplFlags;

	return S_OK;
}

HRESULT CMethodInfo::update(	const CMethodInfo& methodInfoNew, 
								ModuleID moduleID,
								FunctionID functionID,
								IMethodMalloc* piAlloc,
								IMetaDataEmit* piEmit,
								const TypeListTokenMappings& listEquivalentTokens)
{
	HRESULT hr = S_OK;

	_ASSERTE(wcscmp(getTypeName(), methodInfoNew.getTypeName()) == 0);
	_ASSERTE(wcscmp(m_szMemberName, methodInfoNew.m_szMemberName) == 0);
	_ASSERTE(m_ulNameLen == methodInfoNew.m_ulNameLen);

	// Copy the method header and body
    hr = cloneMethodInfo(methodInfoNew, piAlloc);
    VERIFY_HR(hr, _T("Failed to clone method."), ILXLOG_ERROR);

	// Update the rva
	/*
	if (m_ulCodeRVA == 0 && !isDelegate())
	{
		m_ulCodeRVA = 0; //(ULONG) m_pbMethodHeader; //0x00002110; // FIX ME!
		hr = piEmit->SetRVA(m_methodDef, m_ulCodeRVA);
		VERIFY_HR1(hr, ILXLOG_ERROR, _T("Failed to update RVA newly emitted method '%ls'."),
			toString());

		ILXLOGMSG4(ILXLOG_SUCCESS, _T("Updated RVA for newly emitted method '%ls' to 0x%08X 0x%08X 0x%08X."),
			toString(), m_pbMethodHeader, m_pbLoadAddress, m_ulCodeRVA);
	}
	*/

	// Update the method signature. If the signature is new, it should
	// have been emitted previously during a call to MetadataDiff, and
	// it will appear in the token map.
	hr = updateSignature(methodInfoNew.m_tokSig, &m_tokSig,
		&m_pbSigBlob, &m_ulSigBlob, listEquivalentTokens);

	// Update the local var signature. If the signature is new, it should
	// have been emitted previously during a call to MetadataDiff, and
	// it will appear in the token map.
	hr = updateSignature(methodInfoNew.m_tokLocalVarSig, &m_tokLocalVarSig,
		&m_pLocalVarSig, &m_ulLocalVarSigSize, listEquivalentTokens);

	m_moduleID = moduleID;
	m_functionID = functionID;
	m_bWasModified = true;

	return S_OK;
}

HRESULT CMethodInfo::updateSignature(	mdSignature tokSignatureNew,
										mdSignature* ptokSignature,
										PCCOR_SIGNATURE* pcorSigBlob,
										ULONG* pulBlobLen,
										const TypeListTokenMappings& listEquivalentTokens)
{
	HRESULT hr = S_OK;

	_ASSERTE(ptokSignature);

    MapToken(listEquivalentTokens, &tokSignatureNew);

	if (*ptokSignature == tokSignatureNew)
		return S_FALSE;

	*ptokSignature = tokSignatureNew;

	// Update the blob and blob size
	if (*ptokSignature == mdSignatureNil)
	{
		*pcorSigBlob = 0;
		*pulBlobLen = 0;
	}
	else
	{
		hr = m_spiImport->GetSigFromToken(	*ptokSignature,
											pcorSigBlob,
											pulBlobLen);
		VERIFY_HR(hr, _T("GetSigFromToken() failed."), ILXLOG_ERROR);
	}
	
	return S_OK;
}

void CMethodInfo::cloneMethodHeader(	IMethodMalloc* piAlloc,
                                        LPBYTE* ppbMethodHeaderNew,
                                        bool* pbMustDelete) const
{
	HRESULT hr = S_OK;

   _ASSERTE(pbMustDelete);
   _ASSERTE(m_ulMethodSize > 0);

	// Step 1:  Allocate memory for the new method

    if (*pbMustDelete)
    {
        delete[] (*ppbMethodHeaderNew);
    }

    if (piAlloc)
    {
        (*ppbMethodHeaderNew) = (LPBYTE) piAlloc->Alloc(m_ulMethodSize);
        *pbMustDelete = false;
    }
    else
    {
        (*ppbMethodHeaderNew) = (LPBYTE) new unsigned char[m_ulMethodSize];
        *pbMustDelete = true;
    }

    VERIFY_COND_V(*ppbMethodHeaderNew, _T("Alloc() failed."), ILXLOG_ERROR);

	// Step 2:	Copy the method (header and code)
	memcpy(*ppbMethodHeaderNew, m_pbMethodHeader, m_ulMethodSize);
}

LPBYTE CMethodInfo::cloneMethodHeader(	IMethodMalloc* piAlloc,
										const TypeListTokenMappings& listEquivalentTokens,
                                        bool* pbMustDelete) const
{
	HRESULT hr = S_OK;

	_ASSERTE(m_pbMethodHeader);
   _ASSERTE(pbMustDelete);

	// Clone the method

	LPBYTE pbMethodHeaderNew = 0;
    cloneMethodHeader(piAlloc, &pbMethodHeaderNew, pbMustDelete);
    VERIFY_COND_(0, pbMethodHeaderNew, _T("Alloc() failed."), ILXLOG_ERROR);

	mdSignature tokLocalVarSigNew = m_tokLocalVarSig;
    
    MapToken(listEquivalentTokens, &tokLocalVarSigNew);

	if (m_tokLocalVarSig != tokLocalVarSigNew)
	{
		_ASSERTE(m_tokLocalVarSig != mdSignatureNil);

		COR_ILMETHOD_DECODER methodDecoder(
			reinterpret_cast<COR_ILMETHOD*>(pbMethodHeaderNew));
		_ASSERTE(methodDecoder.IsFat());

		// Hammer the new local var sig token
		((IMAGE_COR_ILMETHOD_FAT*) pbMethodHeaderNew)->LocalVarSigTok = 
			tokLocalVarSigNew;

		#ifdef _DEBUG
		COR_ILMETHOD_DECODER methodDecoder2(
			reinterpret_cast<COR_ILMETHOD*>(pbMethodHeaderNew));
		mdSignature tokTest = methodDecoder2.GetLocalVarSigTok() ? 
			methodDecoder2.GetLocalVarSigTok() : mdSignatureNil;
		_ASSERTE(tokTest == tokLocalVarSigNew);
		#endif // _DEBUG
	}

	return pbMethodHeaderNew;
}

LPBYTE CMethodInfo::getCodeStart() const
{
	_ASSERTE(m_typeDef != mdTypeDefNil);
	return const_cast<LPBYTE>(m_pbMethodIL);
}

LPBYTE CMethodInfo::getCodeEnd() const
{
	_ASSERTE(m_typeDef != mdTypeDefNil);
	return getCodeStart() + m_ulCodeSize;
}

bool CMethodInfo::isPInvoke() const
{
	return IsMdPinvokeImpl(m_dwFlags) != 0;
}

bool CMethodInfo::isAbstract() const
{
	return IsMdAbstract(m_dwFlags) != 0;
}

bool CMethodInfo::isCtor() const
{
	return IsMdInstanceInitializerW(m_dwFlags, m_szMemberName);
}

bool CMethodInfo::isCCtor() const
{
	return IsMdClassConstructorW(m_dwFlags, m_szMemberName);
}

const WCHAR* CMethodInfo::getTypeName() const
{
	HRESULT hr = S_OK;

	// For Global methods, typedef is nil. I've noticed
	// this with unmanaged/pinvoke methods.
	//_ASSERTE(m_typeDef != mdTypeDefNil);

	if (m_typeDef == mdTypeDefNil)
	{
		return L"[global]";
	}
	else if (!*m_szTypeName && m_spiImport)
	{
		DWORD dwSize = nMAX_ID-1;
		DWORD dwRetSize = dwSize;

		hr = m_spiImport->GetTypeDefProps(	m_typeDef, 
											m_szTypeName, 
											dwSize,
											&dwRetSize, 
											0, 0 );
		VERIFY_HR1_(L"<ERROR>", hr, ILXLOG_WARNING,
			_T("Failed to obtain containing type for method '%ls'."),
			m_szMemberName);

		_ASSERTE(dwRetSize <= dwSize);
	}

	return m_szTypeName;
}

/////////////////////
// HELPER FUNCTIONS
/////////////////////

HRESULT MethodsAreEquivalent(	const CMethodInfo& pre, 
								const CMethodInfo& post, 
								const TypeListTokenMappings& listEquivalentTokens)
{
	HRESULT hr = S_OK;

	// Ignore m_typeDef differences due to different assemblies
	// Ignore m_ulCodeRVA differences due to different assemblies

	// Do quick comparisons first

	if (pre.m_szMemberName != post.m_szMemberName)
	{
		if (!pre.m_szMemberName || !post.m_szMemberName)
		{
			ILXLOGMSG4(ILXLOG_WARNING, _T("Method names not equal: '%ls.%ls' != '%ls.%ls'."), 
				pre.getTypeName(), pre.m_szMemberName ? pre.m_szMemberName : L"(null)", 
				post.getTypeName(), post.m_szMemberName ? post.m_szMemberName : L"(null)");
			return E_FAIL; // One is null but not the other
		}

		if (wcscmp(pre.m_szMemberName, post.m_szMemberName) != 0)
		{
			ILXLOGMSG4(ILXLOG_WARNING, _T("Method names not equal: '%ls.%ls' != '%ls.%ls'."), 
				pre.getTypeName(), pre.m_szMemberName, post.getTypeName(), post.m_szMemberName);
			return E_FAIL; // Strings are different (method name has changed)
		}
	}

	if (pre.m_ulNameLen != post.m_ulNameLen)
	{
		ILXLOGMSG4(ILXLOG_WARNING, _T("Method names not equal: '%ls.%ls' != '%ls.%ls'."), 
			pre.getTypeName(), pre.m_szMemberName ? pre.m_szMemberName : L"(null)", 
			post.getTypeName(), post.m_szMemberName ? post.m_szMemberName : L"(null)");
		return E_FAIL; // Method names aren't the same
	}

	if (pre.m_dwFlags != post.m_dwFlags)
	{
		ILXLOGMSG4(ILXLOG_INFO, _T("Flags have changed for method '%ls.%ls' (Old:%d New:%d)."), 
			pre.getTypeName(), pre.m_szMemberName, pre.m_dwFlags, post.m_dwFlags);
		return S_FALSE;
	}

	if (SigCmp(	pre.m_pbSigBlob, pre.m_ulSigBlob, pre.m_spiImport,
				post.m_pbSigBlob, post.m_ulSigBlob, post.m_spiImport,
				L"Method", pre.m_szMemberName) != 0)
	{
		return S_FALSE;
	}

	if (pre.m_ulImplFlags != post.m_ulImplFlags)
	{
		ILXLOGMSG4(ILXLOG_INFO, _T("Impl flags have changed for method '%ls.%ls' (Old:%d New:%d)."), 
			pre.getTypeName(), pre.m_szMemberName, pre.m_ulImplFlags, post.m_ulImplFlags);
		return S_FALSE;
	}

	if (pre.m_ulMethodHeaderSize != post.m_ulMethodHeaderSize)
	{
		ILXLOGMSG4(ILXLOG_INFO, _T("Header size has changed for method '%ls.%ls' (Old:%d New:%d)."), 
			pre.getTypeName(), pre.m_szMemberName, pre.m_ulMethodHeaderSize, post.m_ulMethodHeaderSize);
		return S_FALSE;
	}

	if (pre.m_ulCodeSize != post.m_ulCodeSize)
	{
		ILXLOGMSG4(ILXLOG_INFO, _T("Code size has changed for method '%ls.%ls' (Old:%d New:%d)."), 
			pre.getTypeName(), pre.m_szMemberName, pre.m_ulCodeSize, post.m_ulCodeSize);
		return S_FALSE;
	}

	// Compare local variables

	if (!pre.m_pbMethodHeader || !post.m_pbMethodHeader)
	{
		// Unable to obtain method header for some reason
		ILXLOGMSG4(ILXLOG_WARNING, _T("Invalid method header for '%ls.%ls' (Old:0x%08X New:%0x%08X)."), 
			pre.getTypeName(), pre.m_szMemberName, pre.m_pbMethodHeader, post.m_pbMethodHeader);
		return E_FAIL; 
	}

	if (SigCmp(	pre.m_pLocalVarSig, pre.m_ulLocalVarSigSize, pre.m_spiImport,
				post.m_pLocalVarSig, post.m_ulLocalVarSigSize, post.m_spiImport,
				L"Var", pre.m_szMemberName) != 0)
	{
		return S_FALSE;
	}

	// Compare method header

	if (pre.m_pbMethodHeader != post.m_pbMethodHeader)
	{
		if (!pre.m_pbMethodHeader || !post.m_pbMethodHeader)
		{
			ILXLOGMSG4(ILXLOG_WARNING, _T("Invalid method header for '%ls.%ls' (Old:0x%08X New:%0x%08X)."), 
				pre.getTypeName(), pre.m_szMemberName, pre.m_pbMethodHeader, post.m_pbMethodHeader);
			return E_FAIL; // One is null but not the other
		}

		//if (memcmp(pre.m_pbMethodHeader, post.m_pbMethodHeader, pre.m_ulMethodHeaderSize) != 0)
		//{
		//	ILXLOGMSG1(ILXLOG_WARNING, _T("Method headers are different for '%ls'."), 
		//		pre.m_szMemberName);
		//	return E_FAIL; // Method headers should be the same
		//}
	}

	// Compare the method IL
	hr = MethodILIsEquivalent(pre, post, listEquivalentTokens);
	VERIFY_HR2(hr, ILXLOG_ERROR, _T("Failed to compare method '%ls.%ls'."),
		pre.getTypeName(), pre.m_szMemberName);

	if (hr == S_FALSE)
	{
		ILXLOGMSG2(ILXLOG_INFO, _T("Method IL is different for '%ls.%ls'."), 
			pre.getTypeName(), pre.m_szMemberName);
		return S_FALSE;
	}

	// Compare exception handlers

	return S_OK;
}

// The methods might only be different in terms of token references
// (strings vars) which end up getting fixed up later. The method
// bodies may be exactly the same but the actual strings referenced
// are different. This will happen if all you do is change a string
// value in a method.
//
// We must "smartly" compare IL in the face of 1) string refs that
// seem different but are really the same, 2) string refs that
// seem the same but are really different.  We only have to
// perform this special comparison if this list of mapped references
// is not empty. If its empty, a simple binary compare will do.
HRESULT MethodILIsEquivalent(	const CMethodInfo& pre, 
								const CMethodInfo& post, 
								const TypeListTokenMappings& listEquivalentTokens)
{
	if (pre.m_pbMethodIL == post.m_pbMethodIL)
	{
		return E_UNEXPECTED;
	}
	else if (!pre.m_pbMethodIL || !post.m_pbMethodIL)
	{
		return E_FAIL; // One is null but not the other
	}
	else if (listEquivalentTokens.size() == 0)
	{
		// If the string tables are exactly the same then we can just perform
		// a binary comparison
		return memcmp(	pre.m_pbMethodIL, 
						post.m_pbMethodIL, 
						pre.m_ulCodeSize) == 0 ? S_OK : S_FALSE;
	}

	// Compare each method instruction

	opcode_iterator itOPCODEs_PRE(pre);
	opcode_iterator itOPCODEs_POST(post);
	
	for(; 
		!itOPCODEs_PRE.end() && !itOPCODEs_POST.end(); 
		++itOPCODEs_PRE, ++itOPCODEs_POST)
	{
		if (!OpcodesAreEqual(itOPCODEs_PRE, itOPCODEs_POST))
		{
			// Opcodes don't match
			return S_FALSE;
		}

		opcode_format_t operandFormat = g_ILcodes[itOPCODEs_PRE.getOpCode()].operandFormat;

		if (operandFormat == InlineString ||
			operandFormat == InlineMethod)
		{
			// Compare two tokens

			const DWORD* pdwTokAddrPre = (DWORD*) itOPCODEs_PRE.getOpCodeArg();
			const DWORD* pdwTokAddrPost = (DWORD*) itOPCODEs_POST.getOpCodeArg();

			_ASSERTE(pdwTokAddrPre && pdwTokAddrPost);
			if (!(pdwTokAddrPre && pdwTokAddrPost))
				return E_UNEXPECTED; // Return true on error so we ignore the method

			mdToken tokPre = *pdwTokAddrPre;
			mdToken tokPost = *pdwTokAddrPost;

			// Ideally, we could patch up the method right here
            MapToken(listEquivalentTokens, &tokPost);

			if (tokPre != tokPost)
			{
				// String references don't refer to the same string value
				return S_FALSE;
			}
		}
		else if (!InstructionsAreEqual(itOPCODEs_PRE, itOPCODEs_POST))
		{
			// Opcodes are the same but the opcode params are different
			return S_FALSE;
		}

	} // for each instruction

	if (itOPCODEs_PRE.isError() || itOPCODEs_POST.isError())
	{
		// There was an IL parsing error so ignore this method
		return E_FAIL;
	}
	else
	{
		// Returns true if we reached the end of both methods at the same time
		return itOPCODEs_PRE.end() && itOPCODEs_POST.end() ? S_OK : S_FALSE;
	}
}

HRESULT GetMethodNameFromFunctionId(	ICorProfilerInfo* pProfInfo,
										FunctionID functionId,
										wchar_t* wszClass,
										wchar_t* wszMethod,
										DWORD dwSize)
{
	HRESULT hr = S_OK;

	CMethodInfo methodInfo;
	hr = methodInfo.init(pProfInfo, functionId, 0, NULL);
    if (FAILED(hr)) return hr;

	wcsncpy_s(wszClass, dwSize, methodInfo.getTypeName(), dwSize);
	wszClass[dwSize-1] = '\0';

	wcsncpy_s(wszMethod, dwSize, methodInfo.m_szMemberName, dwSize);
	wszMethod[dwSize-1] = '\0';

    return S_OK;
}
